@@ -1,7 +1,7 @@
 /*
  * Nail - a mail user agent derived from Berkeley Mail.
  *
- * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ * Copyright (c) 2000-2004 Gunnar Ritter, Freiburg i. Br., Germany.
  */
 /*
  * Copyright (c) 1980, 1993
@@ -38,41 +38,45 @@
 
 #ifndef lint
 #ifdef	DOSCCS
-static char sccsid[] = "@(#)cmd3.c	2.16 (gritter) 3/30/03";
+static char sccsid[] = "@(#)cmd3.c	2.76 (gritter) 12/3/04";
 #endif
 #endif /* not lint */
 
+#include <float.h>
 #include "rcv.h"
 #include "extern.h"
+#include <unistd.h>
+#include <errno.h>
 
 /*
  * Mail -- a mail program
  *
  * Still more user commands.
  */
-static int	diction __P((const void *, const void *));
-static int	bangexp __P((char **, size_t *));
-static void	make_ref __P((struct message *, struct header *));
-static int	respond_internal __P((int *, int));
-static char	*reedit __P((char *));
-static void	sort __P((char **));
-static int	Respond_internal __P((int [], int));
-static int	forward1 __P((void *, int));
-static RETSIGTYPE	onpipe __P((int));
-static enum okay	delete_shortcut __P((const char *));
-#ifdef	__STDC__
-static int	(*respond_or_Respond(int c))(int *, int);
-#else
-static int	(*respond_or_Respond())();
-#endif
+
+static int bangexp(char **str, size_t *size);
+static void make_ref(struct message *mp, struct header *head);
+static int (*respond_or_Respond(int c))(int *, int);
+static int respond_internal(int *msgvec, int recipient_record);
+static char *reedit(char *subj);
+static char *fwdedit(char *subj);
+static void onpipe(int signo);
+static void asort(char **list);
+static int diction(const void *a, const void *b);
+static int file1(char *name);
+static int shellecho(const char *cp);
+static int Respond_internal(int *msgvec, int recipient_record);
+static int resend1(void *v, int add_resent);
+static void list_shortcuts(void);
+static enum okay delete_shortcut(const char *str);
+static float huge(void);
 
 /*
  * Process a shell escape by saving signals, ignoring signals,
  * and forking a sh -c
  */
-int
-shell(v)
-	void *v;
+int 
+shell(void *v)
 {
 	char *str = v;
 	sighandler_type sigint = safe_signal(SIGINT, SIG_IGN);
@@ -85,9 +89,9 @@ shell(v)
 	if (bangexp(&cmd, &cmdsize) < 0)
 		return 1;
 	if ((shell = value("SHELL")) == NULL)
-		shell = PATH_CSHELL;
-	(void) run_command(shell, 0, -1, -1, "-c", cmd, NULL);
-	(void) safe_signal(SIGINT, sigint);
+		shell = SHELL;
+	run_command(shell, 0, -1, -1, "-c", cmd, NULL);
+	safe_signal(SIGINT, sigint);
 	printf("!\n");
 	free(cmd);
 	return 0;
@@ -97,18 +101,17 @@ shell(v)
  * Fork an interactive shell.
  */
 /*ARGSUSED*/
-int
-dosh(v)
-	void *v;
+int 
+dosh(void *v)
 {
 	sighandler_type sigint = safe_signal(SIGINT, SIG_IGN);
 	char *shell;
 
 	if ((shell = value("SHELL")) == NULL)
-		shell = PATH_CSHELL;
-	(void) run_command(shell, 0, -1, -1, NULL, NULL, NULL);
-	(void) safe_signal(SIGINT, sigint);
-	sputc('\n', stdout);
+		shell = SHELL;
+	run_command(shell, 0, -1, -1, NULL, NULL, NULL);
+	safe_signal(SIGINT, sigint);
+	putchar('\n');
 	return 0;
 }
 
@@ -121,9 +124,7 @@ static char	*lastbang;
 static size_t	lastbangsize;
 
 static int
-bangexp(str, size)
-	char **str;
-	size_t *size;
+bangexp(char **str, size_t *size)
 {
 	char *bangbuf;
 	int changed = 0;
@@ -168,45 +169,43 @@ bangexp(str, size)
 }
 
 /*ARGSUSED*/
-int
-help(v)
-	void *v;
+int 
+help(void *v)
 {
-	const char *helptext = catgets(catd, CATSET, 36,
-"    Nail   Commands\n\
-t <message list>                type messages\n\
-n                               goto and type next message\n\
-e <message list>                edit messages\n\
-f <message list>                give head lines of messages\n\
-d <message list>                delete messages\n\
-s <message list> file           append messages to file\n\
-u <message list>                undelete messages\n\
-R <message list>                reply to message senders\n\
-r <message list>                reply to message senders and all recipients\n\
-pre <message list>              make messages go back to /var/mail\n\
-m <user list>                   mail to specific users\n\
-q                               quit, saving unresolved messages in mbox\n\
-x                               quit, do not remove system mailbox\n\
-h                               print out active message headers\n\
+	const char *helptext =
+"               %s commands\n\
+type <message list>             type messages\n\
+next                            goto and type next message\n\
+from <message list>             give head lines of messages\n\
+headers                         print out active message headers\n\
+delete <message list>           delete messages\n\
+undelete <message list>         undelete messages\n\
+save <message list> folder      append messages to folder and mark as saved\n\
+copy <message list> folder      append messages to folder without marking them\n\
+write <message list> file       append message texts to file, save attachments\n\
+preserve <message list>         keep incoming messages in mailbox even if saved\n\
+Reply <message list>            reply to message senders\n\
+reply <message list>            reply to message senders and all recipients\n\
+mail addresses                  mail to specific recipients\n\
+file folder                     change to another folder\n\
+quit                            quit and apply changes to folder\n\
+xit                             quit and discard changes made to folder\n\
 !                               shell escape\n\
-cd [directory]                  chdir to directory or home if none given\n\
+cd <directory>                  chdir to directory or home if none given\n\
+list                            list names of all available commands\n\
 \n\
-A <message list> consists of integers, ranges of same, or user names separated\n\
-by spaces.  If omitted, Nail uses the last message typed.\n\
-\n\
-A <user list> consists of user names or aliases separated by spaces.\n\
-Aliases are defined in .mailrc in your home directory.\n");
+A <message list> consists of integers, ranges of same, or other criteria\n\
+separated by spaces.  If omitted, %s uses the last message typed.\n";
 
-	fputs(helptext, stdout);
+	fprintf(stdout, helptext, progname, progname);
 	return(0);
 }
 
 /*
  * Change user's working directory.
  */
-int
-schdir(v)
-	void *v;
+int 
+schdir(void *v)
 {
 	char **arglist = v;
 	char *cp;
@@ -223,10 +222,8 @@ schdir(v)
 	return 0;
 }
 
-static void
-make_ref(mp, head)
-struct message *mp;
-struct header *head;
+static void 
+make_ref(struct message *mp, struct header *head)
 {
 	char *oldref, *oldmsgid, *newref;
 	size_t reflen;
@@ -270,13 +267,8 @@ struct header *head;
 	head->h_ref = n;
 }
 
-#ifdef	__STDC__
 static int
 (*respond_or_Respond(int c))(int *, int)
-#else
-static int
-(*respond_or_Respond(c))()
-#endif
 {
 	int opt = 0;
 
@@ -285,44 +277,38 @@ static int
 	return ((opt == 1) ^ (c == 'R')) ? Respond_internal : respond_internal;
 }
 
-int
-respond(v)
-	void *v;
+int 
+respond(void *v)
 {
 	return (respond_or_Respond('r'))((int *)v, 0);
 }
 
-int
-respondall(v)
-	void *v;
+int 
+respondall(void *v)
 {
 	return respond_internal((int *)v, 0);
 }
 
-int
-respondsender(v)
-	void *v;
+int 
+respondsender(void *v)
 {
 	return Respond_internal((int *)v, 0);
 }
 
-int
-followup(v)
-	void *v;
+int 
+followup(void *v)
 {
 	return (respond_or_Respond('r'))((int *)v, 1);
 }
 
-int
-followupall(v)
-	void *v;
+int 
+followupall(void *v)
 {
 	return respond_internal((int *)v, 1);
 }
 
-int
-followupsender(v)
-	void *v;
+int 
+followupsender(void *v)
 {
 	return Respond_internal((int *)v, 1);
 }
@@ -331,15 +317,16 @@ followupsender(v)
  * Reply to a list of messages.  Extract each name from the
  * message header and send them off to mail1()
  */
-static int
-respond_internal(msgvec, recipient_record)
-	int *msgvec;
+static int 
+respond_internal(int *msgvec, int recipient_record)
 {
 	struct message *mp;
 	char *cp, *rcv;
+	enum gfield	gf = value("fullnames") ? GFULL : GSKIN;
 	struct name *np = NULL;
 	struct header head;
 
+	memset(&head, 0, sizeof head);
 	if (msgvec[1] != 0) {
 		printf(catgets(catd, CATSET, 37,
 			"Sorry, can't reply to multiple messages at once\n"));
@@ -348,36 +335,34 @@ respond_internal(msgvec, recipient_record)
 	mp = &message[msgvec[0] - 1];
 	touch(mp);
 	setdot(mp);
-	if ((rcv = skin(hfield("reply-to", mp))) == NULL)
-		if ((rcv = skin(hfield("from", mp))) == NULL)
-			rcv = skin(nameof(mp, 1));
+	if ((rcv = hfield("reply-to", mp)) == NULL)
+		if ((rcv = hfield("from", mp)) == NULL)
+			rcv = nameof(mp, 1);
 	if (rcv != NULL)
-		np = extract(rcv, GTO);
-	if ((cp = skin(hfield("to", mp))) != NULL)
-		np = cat(np, extract(cp, GTO));
+		np = sextract(rcv, GTO|gf);
+	if ((cp = hfield("to", mp)) != NULL)
+		np = cat(np, sextract(cp, GTO|gf));
 	np = elide(np);
 	/*
 	 * Delete my name from the reply list,
 	 * and with it, all my alternate names.
 	 */
 	np = delete_alternates(np);
-	if (np == NIL)
-		np = extract(rcv, GTO);
+	if (np == NULL)
+		np = sextract(rcv, GTO|gf);
 	head.h_to = np;
 	if ((head.h_subject = hfield("subject", mp)) == NULL)
 		head.h_subject = hfield("subj", mp);
 	head.h_subject = reedit(head.h_subject);
-	if ((cp = skin(hfield("cc", mp))) != NULL) {
-		np = elide(extract(cp, GCC));
+	if ((cp = hfield("cc", mp)) != NULL) {
+		np = elide(sextract(cp, GCC|gf));
 		np = delete_alternates(np);
 		head.h_cc = np;
-	} else
-		head.h_cc = NIL;
-	head.h_bcc = NIL;
+	}
 	make_ref(mp, &head);
-	head.h_attach = NULL;
-	head.h_smopts = NIL;
-	mail1(&head, 1, mp, NULL, recipient_record, 0);
+	if (mail1(&head, 1, mp, NULL, recipient_record, 0, 0) == OKAY &&
+			value("markanswered"))
+		mp->m_flag |= MANSWER|MANSWERED;
 	return(0);
 }
 
@@ -386,8 +371,7 @@ respond_internal(msgvec, recipient_record)
  * it does not already.
  */
 static char *
-reedit(subj)
-	char *subj;
+reedit(char *subj)
 {
 	char *newsubj;
 	struct str in, out;
@@ -407,13 +391,104 @@ reedit(subj)
 	return newsubj;
 }
 
+/*
+ * Forward a message to a new recipient, in the sense of RFC 2822.
+ */
+static int
+forward1(char *str, int recipient_record)
+{
+	int	*msgvec, f;
+	char	*recipient;
+	struct message	*mp;
+	struct header	head;
+
+	msgvec = salloc((msgCount + 2) * sizeof *msgvec);
+	if ((recipient = laststring(str, &f, 0)) == NULL) {
+		puts(catgets(catd, CATSET, 47, "No recipient specified."));
+		return 1;
+	}
+	if (!f) {
+		*msgvec = first(0, MMNORM);
+		if (*msgvec == 0) {
+			if (inhook)
+				return 0;
+			printf("No messages to forward.\n");
+			return 1;
+		}
+		msgvec[1] = 0;
+	}
+	if (f && getmsglist(str, msgvec, 0) < 0)
+		return 1;
+	if (*msgvec == 0) {
+		if (inhook)
+			return 0;
+		printf("No applicable messages.\n");
+		return 1;
+	}
+	if (msgvec[1] != 0) {
+		printf("Cannot forward multiple messages at once\n");
+		return 1;
+	}
+	mp = &message[*msgvec - 1];
+	touch(mp);
+	setdot(mp);
+	memset(&head, 0, sizeof head);
+	if ((head.h_to = sextract(recipient,
+			GTO | (value("fullnames") ? GFULL : GSKIN))) == NULL)
+		return 1;
+	if ((head.h_subject = hfield("subject", mp)) == NULL)
+		head.h_subject = hfield("subj", mp);
+	head.h_subject = fwdedit(head.h_subject);
+	mail1(&head, 1, mp, NULL, recipient_record, 1, 0);
+	return 0;
+}
+
+/*
+ * Modify the subject we are replying to to begin with Fwd:.
+ */
+static char *
+fwdedit(char *subj)
+{
+	char *newsubj;
+	struct str	in, out;
+
+	if (subj == NULL || *subj == '\0')
+		return NULL;
+	in.s = subj;
+	in.l = strlen(subj);
+	mime_fromhdr(&in, &out, TD_ISPR|TD_ICONV);
+	newsubj = salloc(strlen(out.s) + 6);
+	strcpy(newsubj, "Fwd: ");
+	strcpy(&newsubj[5], out.s);
+	free(out.s);
+	return newsubj;
+}
+
+/*
+ * The 'forward' command.
+ */
+int
+forwardcmd(void *v)
+{
+	return forward1(v, 0);
+}
+
+/*
+ * Similar to forward, saving the message in a file named after the
+ * first recipient.
+ */
+int
+Forwardcmd(void *v)
+{
+	return forward1(v, 1);
+}
+
 /*
  * Preserve the named messages, so that they will be sent
  * back to the system mailbox.
  */
-int
-preserve(v)
-	void *v;
+int 
+preserve(void *v)
 {
 	int *msgvec = v;
 	struct message *mp;
@@ -430,6 +505,7 @@ preserve(v)
 		mp->m_flag |= MPRESERVE;
 		mp->m_flag &= ~MBOX;
 		setdot(mp);
+		did_print_dot = 1;
 	}
 	return(0);
 }
@@ -437,9 +513,8 @@ preserve(v)
 /*
  * Mark all given messages as unread.
  */
-int
-unread(v)
-	void *v;
+int 
+unread(void *v)
 {
 	int	*msgvec = v;
 	int *ip;
@@ -448,16 +523,33 @@ unread(v)
 		setdot(&message[*ip-1]);
 		dot->m_flag &= ~(MREAD|MTOUCH);
 		dot->m_flag |= MSTATUS;
+		if (mb.mb_type == MB_IMAP || mb.mb_type == MB_CACHE)
+			imap_unread(&message[*ip-1], *ip);
 	}
 	return(0);
 }
 
 /*
- * Print the size of each message.
+ * Mark all given messages as read.
  */
 int
-messize(v)
-	void *v;
+seen(void *v)
+{
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip; ip++) {
+		setdot(&message[*ip-1]);
+		touch(&message[*ip-1]);
+	}
+	return 0;
+}
+
+/*
+ * Print the size of each message.
+ */
+int 
+messize(void *v)
 {
 	int *msgvec = v;
 	struct message *mp;
@@ -466,8 +558,12 @@ messize(v)
 	for (ip = msgvec; *ip != 0; ip++) {
 		mesg = *ip;
 		mp = &message[mesg-1];
-		printf(catgets(catd, CATSET, 40, "%d: %d/%u\n"),
-				mesg, mp->m_lines, (unsigned int)mp->m_size);
+		printf("%d: ", mesg);
+		if (mp->m_xlines > 0)
+			printf("%ld", mp->m_xlines);
+		else
+			putchar(' ');
+		printf("/%lu\n", (unsigned long)mp->m_xsize);
 	}
 	return(0);
 }
@@ -477,9 +573,8 @@ messize(v)
  * by returning an error.
  */
 /*ARGSUSED*/
-int
-rexit(v)
-	void *v;
+int 
+rexit(void *v)
 {
 	if (sourcing)
 		return(1);
@@ -490,8 +585,8 @@ rexit(v)
 static sigjmp_buf	pipejmp;
 
 /*ARGSUSED*/
-static RETSIGTYPE
-onpipe(signo)
+static void 
+onpipe(int signo)
 {
 	siglongjmp(pipejmp, 1);
 }
@@ -500,9 +595,8 @@ onpipe(signo)
  * Set or display a variable value.  Syntax is similar to that
  * of sh.
  */
-int
-set(v)
-	void *v;
+int 
+set(void *v)
 {
 	char **arglist = v;
 	struct var *vp;
@@ -515,17 +609,18 @@ set(v)
 	(void)&cp;
 	(void)&ap;
 	(void)&obuf;
+	(void)&bsdset;
 	if (*arglist == NULL) {
 		for (h = 0, s = 1; h < HSHSIZE; h++)
-			for (vp = variables[h]; vp != NOVAR; vp = vp->v_link)
+			for (vp = variables[h]; vp != NULL; vp = vp->v_link)
 				s++;
 		/*LINTED*/
 		ap = (char **)salloc(s * sizeof *ap);
 		for (h = 0, p = ap; h < HSHSIZE; h++)
-			for (vp = variables[h]; vp != NOVAR; vp = vp->v_link)
+			for (vp = variables[h]; vp != NULL; vp = vp->v_link)
 				*p++ = vp->v_name;
 		*p = NULL;
-		sort(ap);
+		asort(ap);
 		if (is_a_tty[0] && is_a_tty[1] && (cp = value("crt")) != NULL) {
 			if (s > (*cp == '\0' ? screensize() : atoi(cp)) + 3) {
 				cp = get_pager();
@@ -590,9 +685,8 @@ endpipe:
 /*
  * Unset a bunch of variable values.
  */
-int
-unset(v)
-	void *v;
+int 
+unset(void *v)
 {
 	int errs;
 	char **ap;
@@ -606,9 +700,8 @@ unset(v)
 /*
  * Put add users to a group.
  */
-int
-group(v)
-	void *v;
+int 
+group(void *v)
 {
 	char **argv = v;
 	struct grouphead *gh;
@@ -619,15 +712,15 @@ group(v)
 
 	if (*argv == NULL) {
 		for (h = 0, s = 1; h < HSHSIZE; h++)
-			for (gh = groups[h]; gh != NOGRP; gh = gh->g_link)
+			for (gh = groups[h]; gh != NULL; gh = gh->g_link)
 				s++;
 		/*LINTED*/
 		ap = (char **)salloc(s * sizeof *ap);
 		for (h = 0, p = ap; h < HSHSIZE; h++)
-			for (gh = groups[h]; gh != NOGRP; gh = gh->g_link)
+			for (gh = groups[h]; gh != NULL; gh = gh->g_link)
 				*p++ = gh->g_name;
 		*p = NULL;
-		sort(ap);
+		asort(ap);
 		for (p = ap; *p != NULL; p++)
 			printgroup(*p);
 		return(0);
@@ -638,10 +731,10 @@ group(v)
 	}
 	gname = *argv;
 	h = hash(gname);
-	if ((gh = findgroup(gname)) == NOGRP) {
+	if ((gh = findgroup(gname)) == NULL) {
 		gh = (struct grouphead *)scalloc(1, sizeof *gh);
 		gh->g_name = vcopy(gname);
-		gh->g_list = NOGE;
+		gh->g_list = NULL;
 		gh->g_link = groups[h];
 		groups[h] = gh;
 	}
@@ -664,9 +757,8 @@ group(v)
 /*
  * Delete the passed groups.
  */
-int
-ungroup(v)
-	void *v;
+int 
+ungroup(void *v)
 {
 	char **argv = v;
 
@@ -685,9 +777,8 @@ ungroup(v)
  * Sort the passed string vecotor into ascending dictionary
  * order.
  */
-static void
-sort(list)
-	char **list;
+static void 
+asort(char **list)
 {
 	char **ap;
 
@@ -702,9 +793,8 @@ sort(list)
  * Do a dictionary order comparison of the arguments from
  * qsort.
  */
-static int
-diction(a, b)
-	const void *a, *b;
+static int 
+diction(const void *a, const void *b)
 {
 	return(strcmp(*(char **)a, *(char **)b));
 }
@@ -713,59 +803,125 @@ diction(a, b)
  * Change to another file.  With no argument, print information about
  * the current file.
  */
-int
-file(v)
-	void *v;
+int 
+cfile(void *v)
 {
 	char **argv = v;
-	int i;
 
 	if (argv[0] == NULL) {
 		newfileinfo();
 		return 0;
 	}
-	i = setfile(*argv, 0);
+	strncpy(mboxname, expand("&"), sizeof mboxname)[sizeof mboxname-1]='\0';
+	return file1(*argv);
+}
+
+static int 
+file1(char *name)
+{
+	int	i;
+
+	if (inhook) {
+		fprintf(stderr, "Cannot change folder from within a hook.\n");
+		return 1;
+	}
+	i = setfile(name, 0);
 	if (i < 0)
 		return 1;
+	callhook(mailname, 0);
 	if (i > 0 && value("emptystart") == NULL)
 		return 1;
 	announce(value("bsdcompat") != NULL || value("bsdannounce") != NULL);
 	return 0;
 }
 
+static int
+shellecho(const char *cp)
+{
+	int	cflag = 0, n;
+	char	c;
+
+	while (*cp) {
+		if (*cp == '\\') {
+			switch (*++cp) {
+			case '\0':
+				return cflag;
+			case 'a':
+				putchar('\a');
+				break;
+			case 'b':
+				putchar('\b');
+				break;
+			case 'c':
+				cflag = 1;
+				break;
+			case 'f':
+				putchar('\f');
+				break;
+			case 'n':
+				putchar('\n');
+				break;
+			case 'r':
+				putchar('\r');
+				break;
+			case 't':
+				putchar('\t');
+				break;
+			case 'v':
+				putchar('\v');
+				break;
+			default:
+				putchar(*cp&0377);
+				break;
+			case '0':
+				c = 0;
+				n = 3;
+				while (n-- && octalchar(cp[1]&0377)) {
+					c <<= 3;
+					c |= cp[1] - '0';
+					cp++;
+				}
+				putchar(c);
+			}
+		} else
+			putchar(*cp & 0377);
+		cp++;
+	}
+	return cflag;
+}
+
 /*
  * Expand file names like echo
  */
-int
-echo(v)
-	void *v;
+int 
+echo(void *v)
 {
 	char **argv = v;
 	char **ap;
 	char *cp;
+	int cflag = 0;
 
 	for (ap = argv; *ap != NULL; ap++) {
 		cp = *ap;
 		if ((cp = expand(cp)) != NULL) {
 			if (ap != argv)
-				sputc(' ', stdout);
-			printf("%s", cp);
+				putchar(' ');
+			cflag |= shellecho(cp);
 		}
 	}
-	sputc('\n', stdout);
+	if (!cflag)
+		putchar('\n');
 	return 0;
 }
 
-int
-Respond(v)
-	void *v;
+int 
+Respond(void *v)
 {
 	return (respond_or_Respond('R'))((int *)v, 0);
 }
 
-int
-Followup(v)
-	void *v;
+int 
+Followup(void *v)
 {
 	return (respond_or_Respond('R'))((int *)v, 1);
 }
@@ -775,37 +931,35 @@ Followup(v)
  * and not messing around with the To: and Cc: lists as in normal
  * reply.
  */
-static int
-Respond_internal(msgvec, recipient_record)
-	int msgvec[];
+static int 
+Respond_internal(int *msgvec, int recipient_record)
 {
 	struct header head;
 	struct message *mp;
+	enum gfield	gf = value("fullnames") ? GFULL : GSKIN;
 	int *ap;
 	char *cp;
 
-	head.h_to = NIL;
+	memset(&head, 0, sizeof head);
 	for (ap = msgvec; *ap != 0; ap++) {
 		mp = &message[*ap - 1];
 		touch(mp);
 		setdot(mp);
-		if ((cp = skin(hfield("reply-to", mp))) == NULL)
-			if ((cp = skin(hfield("from", mp))) == NULL)
-				cp = skin(nameof(mp, 2));
-		head.h_to = cat(head.h_to, extract(cp, GTO));
+		if ((cp = hfield("reply-to", mp)) == NULL)
+			if ((cp = hfield("from", mp)) == NULL)
+				cp = nameof(mp, 2);
+		head.h_to = cat(head.h_to, sextract(cp, GTO|gf));
 	}
-	if (head.h_to == NIL)
+	if (head.h_to == NULL)
 		return 0;
 	mp = &message[msgvec[0] - 1];
 	if ((head.h_subject = hfield("subject", mp)) == NULL)
 		head.h_subject = hfield("subj", mp);
 	head.h_subject = reedit(head.h_subject);
-	head.h_cc = NIL;
-	head.h_bcc = NIL;
 	make_ref(mp, &head);
-	head.h_attach = NULL;
-	head.h_smopts = NIL;
-	mail1(&head, 1, mp, NULL, recipient_record, 0);
+	if (mail1(&head, 1, mp, NULL, recipient_record, 0, 0) == OKAY &&
+			value("markanswered"))
+		mp->m_flag |= MANSWER|MANSWERED;
 	return 0;
 }
 
@@ -813,9 +967,8 @@ Respond_internal(msgvec, recipient_record)
  * Conditional commands.  These allow one to parameterize one's
  * .mailrc and do some things if sending, others if receiving.
  */
-int
-ifcmd(v)
-	void *v;
+int 
+ifcmd(void *v)
 {
 	char **argv = v;
 	char *cp;
@@ -852,9 +1005,8 @@ ifcmd(v)
  * flip over the conditional flag.
  */
 /*ARGSUSED*/
-int
-elsecmd(v)
-	void *v;
+int 
+elsecmd(void *v)
 {
 
 	switch (cond) {
@@ -888,9 +1040,8 @@ elsecmd(v)
  * End of if statement.  Just set cond back to anything.
  */
 /*ARGSUSED*/
-int
-endifcmd(v)
-	void *v;
+int 
+endifcmd(void *v)
 {
 
 	if (cond == CANY) {
@@ -905,9 +1056,8 @@ endifcmd(v)
 /*
  * Set the list of alternate names.
  */
-int
-alternates(v)
-	void *v;
+int 
+alternates(void *v)
 {
 	char **namelist = v;
 	int c;
@@ -923,10 +1073,10 @@ alternates(v)
 		return(0);
 	}
 	if (altnames != 0)
-		free((char *) altnames);
-	altnames = (char **)scalloc(c, sizeof (char *));
+		free(altnames);
+	altnames = scalloc(c, sizeof (char *));
 	for (ap = namelist, ap2 = altnames; *ap; ap++, ap2++) {
-		cp = (char *)scalloc(strlen(*ap) + 1, sizeof (char));
+		cp = scalloc(strlen(*ap) + 1, sizeof (char));
 		strcpy(cp, *ap);
 		*ap2 = cp;
 	}
@@ -935,20 +1085,19 @@ alternates(v)
 }
 
 /*
- * Do the real work of forwarding.
+ * Do the real work of resending.
  */
-static int
-forward1(v, add_resent)
-void *v;
+static int 
+resend1(void *v, int add_resent)
 {
 	char *name, *str;
 	struct name *to;
-	struct name sn;
+	struct name *sn;
 	int f, *ip, *msgvec;
 
 	str = (char *)v;
 	/*LINTED*/
-	msgvec = (int *)salloc((msgcount + 2) * sizeof *msgvec);
+	msgvec = (int *)salloc((msgCount + 2) * sizeof *msgvec);
 	name = laststring(str, &f, 1);
 	if (name == NULL) {
 		puts(catgets(catd, CATSET, 47, "No recipient specified."));
@@ -957,6 +1106,8 @@ void *v;
 	if (!f) {
 		*msgvec = first(0, MMNORM);
 		if (*msgvec == 0) {
+			if (inhook)
+				return 0;
 			puts(catgets(catd, CATSET, 48,
 					"No applicable messages."));
 			return 1;
@@ -964,35 +1115,37 @@ void *v;
 		msgvec[1] = 0;
 	} else if (getmsglist(str, msgvec, 0) < 0)
 		return 1;
-	sn.n_flink = sn.n_blink = NULL;
-	sn.n_type = GTO;
-	sn.n_name = name;
-	to = usermap(&sn);
-	for (ip = msgvec; *ip && ip - msgvec < msgcount; ip++) {
-		if (forward_msg(&message[*ip - 1], to, add_resent) != 0)
+	if (*msgvec == 0) {
+		if (inhook)
+			return 0;
+		printf("No applicable messages.\n");
+		return 1;
+	}
+	sn = nalloc(name, GTO);
+	to = usermap(sn);
+	for (ip = msgvec; *ip && ip - msgvec < msgCount; ip++) {
+		if (resend_msg(&message[*ip - 1], to, add_resent) != 0)
 			return 1;
 	}
 	return 0;
 }
 
 /*
- * Forward a message list to a third person.
+ * Resend a message list to a third person.
  */
-int
-forwardcmd(v)
-void *v;
+int 
+resendcmd(void *v)
 {
-	return forward1(v, 1);
+	return resend1(v, 1);
 }
 
 /*
- * Forward a message list to a third person without adding headers.
+ * Resend a message list to a third person without adding headers.
  */
-int
-Forwardcmd(v)
-void *v;
+int 
+Resendcmd(void *v)
 {
-	return forward1(v, 0);
+	return resend1(v, 0);
 }
 
 /*
@@ -1000,22 +1153,21 @@ void *v;
  * mail back.
  */
 /*ARGSUSED*/
-int
-newmail(v)
-void *v;
+int 
+newmail(void *v)
 {
-	int omsgcount = msgcount;
-	int val;
+	int val = 1, mdot;
 
-	if ((val = setfile(mailname, 1)) == 0) {
-		newfileinfo();
-		setdot(&message[omsgcount]);
+	if ((mb.mb_type != MB_IMAP || imap_newmail(1)) &&
+			(val = setfile(mailname, 1)) == 0) {
+		mdot = getmdot(1);
+		setdot(&message[mdot - 1]);
 	}
 	return val;
 }
 
-static void
-list_shortcuts()
+static void 
+list_shortcuts(void)
 {
 	struct shortcut *s;
 
@@ -1023,9 +1175,8 @@ list_shortcuts()
 		printf("%s=%s\n", s->sh_short, s->sh_long);
 }
 
-int
-shortcut(v)
-	void *v;
+int 
+shortcut(void *v)
 {
 	char **args = (char **)v;
 	struct shortcut *s;
@@ -1058,8 +1209,7 @@ shortcut(v)
 }
 
 struct shortcut *
-get_shortcut(str)
-	const char *str;
+get_shortcut(const char *str)
 {
 	struct shortcut *s;
 
@@ -1069,9 +1219,8 @@ get_shortcut(str)
 	return s;
 }
 
-static enum okay
-delete_shortcut(str)
-	const char *str;
+static enum okay 
+delete_shortcut(const char *str)
 {
 	struct shortcut *sp, *sq;
 
@@ -1090,9 +1239,8 @@ delete_shortcut(str)
 	return STOP;
 }
 
-int
-unshortcut(v)
-	void *v;
+int 
+unshortcut(void *v)
 {
 	char **args = (char **)v;
 	int errs = 0;
@@ -1112,3 +1260,434 @@ unshortcut(v)
 	}
 	return errs;
 }
+
+struct oldaccount {
+	struct oldaccount	*ac_next;	/* next account in list */
+	char	*ac_name;			/* name of account */
+	char	**ac_vars;			/* variables to set */
+};
+
+static struct oldaccount	*oldaccounts;
+
+struct oldaccount *
+get_oldaccount(const char *name)
+{
+	struct oldaccount	*a;
+
+	for (a = oldaccounts; a; a = a->ac_next)
+		if (a->ac_name && strcmp(name, a->ac_name) == 0)
+			break;
+	return a;
+}
+
+int 
+account(void *v)
+{
+	char	**args = (char **)v;
+	struct oldaccount	*a;
+	char	*cp;
+	int	i, mc;
+	FILE	*fp = stdout;
+
+	if (args[0] == NULL) {
+		if ((fp = Ftemp(&cp, "Ra", "w+", 0600, 1)) == NULL) {
+			perror("tmpfile");
+			return 1;
+		}
+		rm(cp);
+		Ftfree(&cp);
+		mc = listaccounts(fp);
+		for (a = oldaccounts; a; a = a->ac_next)
+			if (a->ac_name) {
+				if (mc++)
+					fputc('\n', fp);
+				fprintf(fp, "%s:\n", a->ac_name);
+				for (i = 0; a->ac_vars[i]; i++)
+					fprintf(fp, "\t%s\n", a->ac_vars[i]);
+			}
+		if (mc)
+			try_pager(fp);
+		Fclose(fp);
+		return 0;
+	}
+	if (args[1] && args[1][0] == '{' && args[1][1] == '\0') {
+		if (args[2] != NULL) {
+			fprintf(stderr, "Syntax is: account <name> {\n");
+			return 1;
+		}
+		if ((a = get_oldaccount(args[0])) != NULL)
+			a->ac_name = NULL;
+		return define1(args[0], 1);
+	}
+	strncpy(mboxname, expand("&"), sizeof mboxname)[sizeof mboxname-1]='\0';
+	if ((a = get_oldaccount(args[0])) == NULL) {
+		if (args[1]) {
+			a = scalloc(1, sizeof *a);
+			a->ac_next = oldaccounts;
+			oldaccounts = a;
+		} else {
+			if ((i = callaccount(args[0])) != CBAD)
+				goto setf;
+			printf("Account %s does not exist.\n", args[0]);
+			return 1;
+		}
+	}
+	if (args[1]) {
+		delaccount(args[0]);
+		a->ac_name = sstrdup(args[0]);
+		for (i = 1; args[i]; i++);
+		a->ac_vars = scalloc(i, sizeof *a->ac_vars);
+		for (i = 0; args[i+1]; i++)
+			a->ac_vars[i] = sstrdup(args[i+1]);
+	} else {
+		unset_allow_undefined = 1;
+		set(a->ac_vars);
+		unset_allow_undefined = 0;
+	setf:	if (!starting)
+			return file1("%");
+	}
+	return 0;
+}
+
+int 
+cflag(void *v)
+{
+	struct message	*m;
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip != 0; ip++) {
+		m = &message[*ip-1];
+		setdot(m);
+		if ((m->m_flag & (MFLAG|MFLAGGED)) == 0)
+			m->m_flag |= MFLAG|MFLAGGED;
+	}
+	return 0;
+}
+
+int 
+cunflag(void *v)
+{
+	struct message	*m;
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip != 0; ip++) {
+		m = &message[*ip-1];
+		setdot(m);
+		if (m->m_flag & (MFLAG|MFLAGGED)) {
+			m->m_flag &= ~(MFLAG|MFLAGGED);
+			m->m_flag |= MUNFLAG;
+		}
+	}
+	return 0;
+}
+
+int 
+canswered(void *v)
+{
+	struct message	*m;
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip != 0; ip++) {
+		m = &message[*ip-1];
+		setdot(m);
+		if ((m->m_flag & (MANSWER|MANSWERED)) == 0)
+			m->m_flag |= MANSWER|MANSWERED;
+	}
+	return 0;
+}
+
+int 
+cunanswered(void *v)
+{
+	struct message	*m;
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip != 0; ip++) {
+		m = &message[*ip-1];
+		setdot(m);
+		if (m->m_flag & (MANSWER|MANSWERED)) {
+			m->m_flag &= ~(MANSWER|MANSWERED);
+			m->m_flag |= MUNANSWER;
+		}
+	}
+	return 0;
+}
+
+int 
+cdraft(void *v)
+{
+	struct message	*m;
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip != 0; ip++) {
+		m = &message[*ip-1];
+		setdot(m);
+		if ((m->m_flag & (MDRAFT|MDRAFTED)) == 0)
+			m->m_flag |= MDRAFT|MDRAFTED;
+	}
+	return 0;
+}
+
+int 
+cundraft(void *v)
+{
+	struct message	*m;
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip != 0; ip++) {
+		m = &message[*ip-1];
+		setdot(m);
+		if (m->m_flag & (MDRAFT|MDRAFTED)) {
+			m->m_flag &= ~(MDRAFT|MDRAFTED);
+			m->m_flag |= MUNDRAFT;
+		}
+	}
+	return 0;
+}
+
+static float 
+huge(void)
+{
+	/*
+	 * This is not perfect, but correct for machines with a 32-bit
+	 * IEEE float, and does at least not produce SIGFPE on the Cray
+	 * Y-MP.
+	 */
+	union {
+		float	f;
+		unsigned long	l;
+	} u;
+
+	u.l = 0xff800000; /* -inf */
+	return u.f;
+}
+
+int 
+ckill(void *v)
+{
+	struct message	*m;
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip != 0; ip++) {
+		m = &message[*ip-1];
+		m->m_flag |= MKILL;
+		m->m_score = huge();
+	}
+	return 0;
+}
+
+int 
+cunkill(void *v)
+{
+	struct message	*m;
+	int	*msgvec = v;
+	int	*ip;
+
+	for (ip = msgvec; *ip != 0; ip++) {
+		m = &message[*ip-1];
+		m->m_flag &= ~MKILL;
+		m->m_score = 0;
+	}
+	return 0;
+}
+
+int 
+cscore(void *v)
+{
+	char	*str = v;
+	char	*sscore, *xp;
+	int	f, *msgvec, *ip;
+	double	nscore;
+	struct message	*m;
+
+	msgvec = salloc((msgCount+2) * sizeof *msgvec);
+	if ((sscore = laststring(str, &f, 0)) == NULL) {
+		fprintf(stderr, "No score given.\n");
+		return 1;
+	}
+	nscore = strtod(sscore, &xp);
+	if (*xp) {
+		fprintf(stderr, "Invalid score: \"%s\"\n", sscore);
+		return 1;
+	}
+	if (nscore > FLT_MAX)
+		nscore = FLT_MAX;
+	else if (nscore < -FLT_MAX)
+		nscore = -FLT_MAX;
+	if (!f) {
+		*msgvec = first(0, MMNORM);
+		if (*msgvec == 0) {
+			if (inhook)
+				return 0;
+			fprintf(stderr, "No messages to score.\n");
+			return 1;
+		}
+		msgvec[1] = 0;
+	} else if (getmsglist(str, msgvec, 0) < 0)
+		return 1;
+	if (*msgvec == 0) {
+		if (inhook)
+			return 0;
+		fprintf(stderr, "No applicable messages.\n");
+		return 1;
+	}
+	for (ip = msgvec; *ip && ip-msgvec < msgCount; ip++) {
+		m = &message[*ip-1];
+		if (m->m_score != huge()) {
+			m->m_score += nscore;
+			if (m->m_score < 0)
+				m->m_flag |= MKILL;
+			else if (m->m_score > 0)
+				m->m_flag &= ~MKILL;
+			if (m->m_score >= 0)
+				setdot(m);
+		}
+	}
+	return 0;
+}
+
+/*ARGSUSED*/
+int 
+cnoop(void *v)
+{
+	switch (mb.mb_type) {
+	case MB_IMAP:
+		imap_noop();
+		break;
+	case MB_POP3:
+		pop3_noop();
+		break;
+	default:
+		break;
+	}
+	return 0;
+}
+
+int 
+cremove(void *v)
+{
+	char	vb[LINESIZE];
+	char	**args = v;
+	char	*name;
+	int	ec = 0;
+
+	if (*args == NULL) {
+		fprintf(stderr, "Syntax is: remove mailbox ...\n");
+		return 1;
+	}
+	do {
+		name = expand(*args);
+		if (strcmp(name, mailname) == 0) {
+			fprintf(stderr,
+				"Cannot remove current mailbox \"%s\".\n",
+				name);
+			ec |= 1;
+			continue;
+		}
+		snprintf(vb, sizeof vb, "Remove \"%s\" (y/n) ? ", name);
+		if (yorn(vb) == 0)
+			continue;
+		switch (which_protocol(name)) {
+		case PROTO_FILE:
+			if (unlink(name) < 0) {	/* do not handle .gz .bz2 */
+				perror(name);
+				ec |= 1;
+			}
+			break;
+		case PROTO_POP3:
+			fprintf(stderr, "Cannot remove POP3 mailbox \"%s\".\n",
+					name);
+			ec |= 1;
+			break;
+		case PROTO_IMAP:
+			if (imap_remove(name) != OKAY)
+				ec |= 1;
+			break;
+		case PROTO_MAILDIR:
+			if (maildir_remove(name) != OKAY)
+				ec |= 1;
+			break;
+		case PROTO_UNKNOWN:
+			fprintf(stderr,
+				"Unknown protocol in \"%s\". Not removed.\n",
+				name);
+			ec |= 1;
+		}
+	} while (*++args);
+	return ec;
+}
+
+int 
+crename(void *v)
+{
+	char	**args = v, *old, *new;
+	enum protocol	oldp, newp;
+	int	ec = 0;
+
+	if (args[0] == NULL || args[1] == NULL || args[2] != NULL) {
+		fprintf(stderr, "Syntax: rename old new\n");
+		return 1;
+	}
+	old = expand(args[0]);
+	oldp = which_protocol(old);
+	new = expand(args[1]);
+	newp = which_protocol(new);
+	if (strcmp(old, mailname) == 0 || strcmp(new, mailname) == 0) {
+		fprintf(stderr, "Cannot rename current mailbox \"%s\".\n", old);
+		return 1;
+	}
+	if ((oldp == PROTO_IMAP || newp == PROTO_IMAP) && oldp != newp) {
+		fprintf(stderr, "Can only rename folders of same type.\n");
+		return 1;
+	}
+	if (newp == PROTO_POP3)
+		goto nopop3;
+	switch (oldp) {
+	case PROTO_FILE:
+		if (link(old, new) < 0) {
+			switch (errno) {
+			case EACCES:
+			case EEXIST:
+			case ENAMETOOLONG:
+			case ENOENT:
+			case ENOSPC:
+			case EXDEV:
+				perror(new);
+				break;
+			default:
+				perror(old);
+			}
+			ec |= 1;
+		} else if (unlink(old) < 0) {
+			perror(old);
+			ec |= 1;
+		}
+		break;
+	case PROTO_MAILDIR:
+		if (rename(old, new) < 0) {
+			perror(old);
+			ec |= 1;
+		}
+		break;
+	case PROTO_POP3:
+	nopop3:	fprintf(stderr, "Cannot rename POP3 mailboxes.\n");
+		ec |= 1;
+		break;
+	case PROTO_IMAP:
+		if (imap_rename(old, new) != OKAY)
+			ec |= 1;
+		break;
+	case PROTO_UNKNOWN:
+		fprintf(stderr, "Unknown protocol in \"%s\" and \"%s\". "
+				"Not renamed.\n", old, new);
+		ec |= 1;
+	}
+	return ec;
+}
