@@ -1,7 +1,7 @@
 /*
  * Nail - a mail user agent derived from Berkeley Mail.
  *
- * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ * Copyright (c) 2000-2004 Gunnar Ritter, Freiburg i. Br., Germany.
  */
 /*
  * Copyright (c) 1980, 1993
@@ -38,7 +38,7 @@
 
 #ifndef lint
 #ifdef	DOSCCS
-static char sccsid[] = "@(#)collect.c	2.14 (gritter) 3/30/03";
+static char sccsid[] = "@(#)collect.c	2.42 (gritter) 12/25/04";
 #endif
 #endif /* not lint */
 
@@ -51,6 +51,8 @@ static char sccsid[] = "@(#)collect.c	2.14 (gritter) 3/30/03";
 
 #include "rcv.h"
 #include "extern.h"
+#include <unistd.h>
+#include <sys/stat.h>
 
 /*
  * Read a message from standard output and return a read file to it
@@ -77,25 +79,28 @@ static	sigjmp_buf	collabort;	/* To end collection with error */
 
 static	sigjmp_buf pipejmp;		/* On broken pipe */
 
-static struct attachment	*read_attachment_data __P((struct attachment *,
-					unsigned));
-static void	insertcommand __P((FILE *, char *));
-static int	exwrite __P((char [], FILE *, int));
-static enum okay	makeheader __P((FILE *, struct header *));
-static void	mesedit __P((int, struct header *));
-static void	mespipe __P((char []));
-static int	forward __P((char [], FILE *, int));
-static void	collhup __P((int));
-static int	include_file __P((FILE *, char *, int *, int *, int));
-static void	collint __P((int));
-static void	collstop __P((int));
-static struct attachment	*append_attachments __P((struct attachment *,
-					char *));
-static RETSIGTYPE	onpipe __P((int));
+static void onpipe(int signo);
+static void insertcommand(FILE *fp, char *cmd);
+static void print_collf(FILE *collf, struct header *hp);
+static int include_file(FILE *fbuf, char *name, int *linecount,
+		int *charcount, int echo);
+static struct attachment *read_attachment_data(struct attachment *ap,
+		unsigned number);
+static struct attachment *append_attachments(struct attachment *attach,
+		char *names);
+static int exwrite(char *name, FILE *fp, int f);
+static enum okay makeheader(FILE *fp, struct header *hp);
+static void mesedit(int c, struct header *hp);
+static void mespipe(char *cmd);
+static int forward(char *ms, FILE *fp, int f);
+static void collstop(int s);
+static void collint(int s);
+static void collhup(int s);
+static int putesc(const char *s, FILE *stream);
 
 /*ARGSUSED*/
-static RETSIGTYPE
-onpipe(signo)
+static void 
+onpipe(int signo)
 {
 	siglongjmp(pipejmp, 1);
 }
@@ -104,9 +109,7 @@ onpipe(signo)
  * Execute cmd and insert its standard output into fp.
  */
 static void
-insertcommand(fp, cmd)
-FILE *fp;
-char *cmd;
+insertcommand(FILE *fp, char *cmd)
 {
 	FILE *obuf = NULL;
 	char *cp;
@@ -118,14 +121,14 @@ char *cmd;
 	if (sigsetjmp(pipejmp, 1))
 		goto endpipe;
 	if (cp == NULL)
-		cp = PATH_CSHELL;
-	if ((obuf = Popen(cmd, "r", cp, 0)) == (FILE *) NULL) {
+		cp = SHELL;
+	if ((obuf = Popen(cmd, "r", cp, 0)) == NULL) {
 		perror(cmd);
 		return;
 	}
 	safe_signal(SIGPIPE, onpipe);
-	while ((c = sgetc(obuf)) != EOF)
-		sputc(c, fp);
+	while ((c = getc(obuf)) != EOF)
+		putc(c, fp);
 endpipe:
 	safe_signal(SIGPIPE, SIG_IGN);
 	Pclose(obuf);
@@ -136,14 +139,13 @@ endpipe:
  * ~p command.
  */
 static void
-print_collf(collf, hp)
-FILE *collf;
-struct header *hp;
+print_collf(FILE *collf, struct header *hp)
 {
 	char *lbuf = NULL;
 	FILE *obuf = stdout;
 	struct attachment *ap;
 	char *cp;
+	enum gfield	gf;
 	size_t	linecnt, maxlines, linesize = 0, linelen, count, count2;
 
 	(void)&obuf;
@@ -166,6 +168,8 @@ struct header *hp;
 			maxlines--;
 		if (hp->h_bcc)
 			maxlines--;
+		if (hp->h_replyto)
+			maxlines--;
 		if (hp->h_attach)
 			maxlines--;
 		if (linecnt > maxlines) {
@@ -182,11 +186,12 @@ struct header *hp;
 	}
 	fprintf(obuf, catgets(catd, CATSET, 62,
 				"-------\nMessage contains:\n"));
-	puthead(hp, obuf, GTO|GSUBJECT|GCC|GBCC|GNL, CONV_TODISP, NULL, NULL);
-	while (fgetline(&lbuf, &linesize, &count, &linelen, collf, 1)) {
-		makeprint(lbuf, linelen);
-		fwrite(lbuf, sizeof *lbuf, linelen, obuf);
-	}
+	gf = GTO|GSUBJECT|GCC|GBCC|GNL|GREPLYTO|GFILES;
+	if (value("fullnames"))
+		gf |= GCOMMA;
+	puthead(hp, obuf, gf, SEND_TODISP, CONV_NONE, NULL, NULL);
+	while (fgetline(&lbuf, &linesize, &count, &linelen, collf, 1))
+		prout(lbuf, linelen, obuf);
 	if (hp->h_attach != NULL) {
 		fputs(catgets(catd, CATSET, 63, "Attachments:"), obuf);
 		for (ap = hp->h_attach; ap != NULL; ap = ap->a_flink)
@@ -204,10 +209,7 @@ endpipe:
 }
 
 static int
-include_file(fbuf, name, linecount, charcount, echo)
-FILE *fbuf;
-char *name;
-int *linecount, *charcount;
+include_file(FILE *fbuf, char *name, int *linecount, int *charcount, int echo)
 {
 	char *interactive, *linebuf = NULL;
 	size_t linesize = 0, linelen, count;
@@ -247,9 +249,7 @@ int *linecount, *charcount;
  * attachment. NULL is returned if no file name is given.
  */
 static struct attachment *
-read_attachment_data(ap, number)
-	struct attachment *ap;
-	unsigned number;
+read_attachment_data(struct attachment *ap, unsigned number)
 {
 	char prefix[80];
 
@@ -304,8 +304,7 @@ read_attachment_data(ap, number)
  * Interactively edit the attachment list.
  */
 struct attachment *
-edit_attachments(attach)
-	struct attachment *attach;
+edit_attachments(struct attachment *attach)
 {
 	struct attachment *ap, *nap;
 	unsigned attno = 1;
@@ -340,9 +339,7 @@ edit_attachments(attach)
  * Put the given file to the end of the attachment list.
  */
 struct attachment *
-add_attachment(attach, file)
-	struct attachment *attach;
-	const char *file;
+add_attachment(struct attachment *attach, const char *file)
 {
 	struct attachment *ap, *nap;
 
@@ -373,9 +370,7 @@ add_attachment(attach, file)
  * the attachment list.
  */
 static struct attachment *
-append_attachments(attach, names)
-	struct attachment *attach;
-	char *names;
+append_attachments(struct attachment *attach, char *names)
 {
 	char *cp;
 	int c;
@@ -405,57 +400,54 @@ append_attachments(attach, names)
 }
 
 FILE *
-collect(hp, printheaders, mp, quotefile, tflag)
-	struct header *hp;
-	int printheaders;
-	struct message *mp;
-	char *quotefile;
+collect(struct header *hp, int printheaders, struct message *mp,
+		char *quotefile, int doprefix, int tflag)
 {
 	FILE *fbuf;
 	struct ignoretab *quoteig;
 	int lc, cc, escape, eofcount;
 	int c, t;
-	char *linebuf = NULL, *cp, *quote;
+	char *linebuf = NULL, *cp, *quote = NULL;
 	size_t linesize;
 	char *tempMail = NULL;
 	int getfields;
 	sigset_t oset, nset;
-	ssize_t count;
-	const char *tildehelp = catgets(catd, CATSET, 49,
-"-----------------------------------------------------------\n\
-The following ~ escapes are defined:\n\
+	long count;
+	enum sendaction	action;
+	sighandler_type	savedtop;
+	const char tildehelp[] =
+"-------------------- ~ ESCAPES ----------------------------\n\
 ~~              Quote a single tilde\n\
 ~@ [file ...]   Edit attachment list\n\
 ~b users        Add users to \"blind\" cc list\n\
 ~c users        Add users to cc list\n\
 ~d              Read in dead.letter\n\
 ~e              Edit the message buffer\n\
-~f messages     Read in messages\n\
+~f messages     Read in messages without indenting lines\n\
 ~F messages     Same as ~f, but keep all header lines\n\
-~h              Prompt for to list, subject and cc list\n\
+~h              Prompt for to list, subject, cc, and \"blind\" cc list\n\
 ~r file         Read a file into the message buffer\n\
 ~p              Print the message buffer\n\
-~m messages     Read in messages, right shifted by a tab\n\
+~q              Abort message composition and save text to dead.letter\n\
+~m messages     Read in messages with each line indented\n\
 ~M messages     Same as ~m, but keep all header lines\n\
 ~s subject      Set subject\n\
 ~t users        Add users to to list\n\
 ~v              Invoke display editor on message\n\
-~w file         Write message onto file.\n\
-~?              Print this message\n\
+~w file         Write message onto file\n\
+~x              Abort message composition and discard text written so far\n\
 ~!command       Invoke the shell\n\
-~|command       Pipe the message through the command\n\
------------------------------------------------------------\n");
+~:command       Execute a regular command\n\
+-----------------------------------------------------------\n";
 
-#ifdef __GNUC__
-	/* Avoid longjmp clobbering */
 	(void) &escape;
 	(void) &eofcount;
 	(void) &getfields;
 	(void) &tempMail;
 	(void) &tflag;
-#endif
+	(void) &quote;
 
-	collf = (FILE*)NULL;
+	collf = NULL;
 	/*
 	 * Start catching signals from here, but we're still die on interrupts
 	 * until we're in the main loop.
@@ -464,6 +456,7 @@ The following ~ escapes are defined:\n\
 	sigaddset(&nset, SIGINT);
 	sigaddset(&nset, SIGHUP);
 	sigprocmask(SIG_BLOCK, &nset, &oset);
+	handlerpush(collint);
 	if ((saveint = safe_signal(SIGINT, SIG_IGN)) != SIG_IGN)
 		safe_signal(SIGINT, collint);
 	if ((savehup = safe_signal(SIGHUP, SIG_IGN)) != SIG_IGN)
@@ -488,7 +481,7 @@ The following ~ escapes are defined:\n\
 	sigprocmask(SIG_SETMASK, &oset, (sigset_t *)NULL);
 
 	noreset++;
-	if ((collf = Ftemp(&tempMail, "Rs", "w+", 0600, 1)) == (FILE *)NULL) {
+	if ((collf = Ftemp(&tempMail, "Rs", "w+", 0600, 1)) == NULL) {
 		perror(catgets(catd, CATSET, 51, "temporary mail file"));
 		goto err;
 	}
@@ -497,9 +490,8 @@ The following ~ escapes are defined:\n\
 
 	if ((cp = value("MAILX_HEAD")) != NULL) {
 		if (is_a_tty[0])
-			puts(cp);
-		fputs(cp, collf);
-		sputc('\n', collf);
+			putesc(cp, stdout);
+		putesc(cp, collf);
 	}
 
 	/*
@@ -510,6 +502,8 @@ The following ~ escapes are defined:\n\
 	getfields = 0;
 	if (!tflag) {
 		t = GTO|GSUBJECT|GCC|GNL;
+		if (value("fullnames"))
+			t |= GCOMMA;
 		if (hp->h_subject == NULL && value("interactive") != NULL &&
 			    (value("ask") != NULL || value("asksub") != NULL))
 			t &= ~GNL, getfields |= GSUBJECT;
@@ -523,7 +517,8 @@ The following ~ escapes are defined:\n\
 				t &= ~GNL, getfields |= GCC;
 		}
 		if (printheaders) {
-			puthead(hp, stdout, t, CONV_TODISP, NULL, NULL);
+			puthead(hp, stdout, t, SEND_TODISP, CONV_NONE,
+					NULL, NULL);
 			fflush(stdout);
 		}
 	}
@@ -531,14 +526,24 @@ The following ~ escapes are defined:\n\
 	/*
 	 * Quote an original message
 	 */
-	if (mp != NULL && (quote = value("quote")) != NULL) {
+	if (mp != NULL && (doprefix || (quote = value("quote")) != NULL)) {
 		quoteig = allignore;
-		if (strcmp(quote, "noheading") == 0) {
-			/*EMPTY*/
+		action = SEND_QUOTE;
+		if (doprefix) {
+			quoteig = fwdignore;
+			if ((cp = value("fwdheading")) == NULL)
+				cp = "-------- Original Message --------";
+			if (*cp) {
+				fprintf(collf, "%s\n", cp);
+				fprintf(stdout, "%s\n", cp);
+			}
+		} else if (strcmp(quote, "noheading") == 0) {
+			/*EMPTY*/;
 		} else if (strcmp(quote, "headers") == 0) {
 			quoteig = ignore;
 		} else if (strcmp(quote, "allheaders") == 0) {
-			quoteig = (struct ignoretab *) NULL;
+			quoteig = NULL;
+			action = SEND_QUOTE_ALL;
 		} else {
 			cp = hfield("from", mp);
 			if (cp != NULL) {
@@ -557,8 +562,8 @@ The following ~ escapes are defined:\n\
 		cp = value("indentprefix");
 		if (cp != NULL && *cp == '\0')
 			cp = "\t";
-		send_message(mp, collf, quoteig, cp, CONV_QUOTE, NULL);
-		send_message(mp, stdout, quoteig, cp, CONV_QUOTE, NULL);
+		send(mp, collf, quoteig, doprefix ? NULL : cp, action, NULL);
+		send(mp, stdout, quoteig, doprefix ? NULL : cp, action, NULL);
 	}
 
 	if ((cp = value("escape")) != NULL)
@@ -660,7 +665,7 @@ cont:
 			/*
 			 * Dump core.
 			 */
-			core((void *)NULL);
+			core(NULL);
 			break;
 #endif	/* DEBUG_COMMANDS */
 		case '!':
@@ -675,6 +680,7 @@ cont:
 			/*
 			 * Escape to command mode, but be nice!
 			 */
+			inhook = 0;
 			execute(&linebuf[2], 1, count - 2);
 			goto cont;
 		case '.':
@@ -714,8 +720,8 @@ cont:
 			 * Add to the To list.
 			 */
 			while ((hp->h_to = checkaddrs(cat(hp->h_to,
-						extract(&linebuf[2], GTO))))
-					== NULL);
+					sextract(&linebuf[2], GTO|GFULL))))
+				== NULL);
 			break;
 		case 's':
 			/*
@@ -741,14 +747,14 @@ cont:
 			 * Add to the CC list.
 			 */
 			hp->h_cc = checkaddrs(cat(hp->h_cc,
-						extract(&linebuf[2], GCC)));
+				sextract(&linebuf[2], GCC|GFULL)));
 			break;
 		case 'b':
 			/*
 			 * Add stuff to blind carbon copies list.
 			 */
 			hp->h_bcc = checkaddrs(cat(hp->h_bcc,
-					extract(&linebuf[2], GBCC)));
+				sextract(&linebuf[2], GBCC|GFULL)));
 			break;
 		case 'd':
 			strncpy(linebuf + 2, getdeadletter(), linesize - 2);
@@ -781,7 +787,7 @@ cont:
 						"%s: Directory\n"), cp);
 				break;
 			}
-			if ((fbuf = Fopen(cp, "r")) == (FILE *)NULL) {
+			if ((fbuf = Fopen(cp, "r")) == NULL) {
 				perror(cp);
 				break;
 			}
@@ -800,8 +806,9 @@ cont:
 				cp++;
 			if ((cp = value(cp)) == NULL || *cp == '\0')
 				break;
-			fputs(cp, collf);
-			sputc('\n', collf);
+			if (is_a_tty[0])
+				putesc(cp, stdout);
+			putesc(cp, collf);
 			break;
 		case 'a':
 		case 'A':
@@ -809,8 +816,9 @@ cont:
 			 * Insert the contents of a signature variable.
 			 */
 			if ((cp = value(c == 'a' ? "sign" : "Sign")) != NULL) {
-				fputs(cp, collf);
-				sputc('\n', collf);
+				if (is_a_tty[0])
+					putesc(cp, stdout);
+				putesc(cp, collf);
 			}
 			break;
 		case 'w':
@@ -875,20 +883,21 @@ cont:
 	}
 	goto out;
 err:
-	if (collf != (FILE *)NULL) {
+	if (collf != NULL) {
 		Fclose(collf);
-		collf = (FILE *)NULL;
+		collf = NULL;
 	}
 out:
-	if (collf != (FILE *)NULL) {
+	if (collf != NULL) {
 		if ((cp = value("MAILX_TAIL")) != NULL) {
 			if (is_a_tty[0])
-				puts(cp);
-			fputs(cp, collf);
-			sputc('\n', collf);
+				putesc(cp, stdout);
+			fflush(collf);
+			putesc(cp, collf);
 		}
 		rewind(collf);
 	}
+	handlerpop();
 	noreset--;
 	sigemptyset(&nset);
 	sigaddset(&nset, SIGINT);
@@ -911,38 +920,28 @@ out:
  * Write a file, ex-like if f set.
  */
 static int
-exwrite(name, fp, f)
-	char name[];
-	FILE *fp;
-	int f;
+exwrite(char *name, FILE *fp, int f)
 {
 	FILE *of;
 	int c;
 	long cc;
 	int lc;
-	struct stat junk;
 
 	if (f) {
 		printf("\"%s\" ", name);
 		fflush(stdout);
 	}
-	if (stat(name, &junk) >= 0 && S_ISREG(junk.st_mode)) {
-		if (!f)
-			fprintf(stderr, "%s: ", name);
-		fprintf(stderr, catgets(catd, CATSET, 64, "File exists\n"));
-		return(-1);
-	}
-	if ((of = Fopen(name, "wx")) == (FILE *)NULL) {
+	if ((of = Fopen(name, "a")) == NULL) {
 		perror(NULL);
 		return(-1);
 	}
 	lc = 0;
 	cc = 0;
-	while ((c = sgetc(fp)) != EOF) {
+	while ((c = getc(fp)) != EOF) {
 		cc++;
 		if (c == '\n')
 			lc++;
-		(void) sputc(c, of);
+		putc(c, of);
 		if (ferror(of)) {
 			perror(name);
 			Fclose(of);
@@ -956,9 +955,7 @@ exwrite(name, fp, f)
 }
 
 static enum okay
-makeheader(fp, hp)
-	FILE *fp;
-	struct header *hp;
+makeheader(FILE *fp, struct header *hp)
 {
 	char *tempEdit;
 	FILE *nf;
@@ -974,8 +971,8 @@ makeheader(fp, hp)
 	unlink(tempEdit);
 	Ftfree(&tempEdit);
 	extract_header(fp, hp);
-	while ((c = sgetc(fp)) != EOF)
-		sputc(c, nf);
+	while ((c = getc(fp)) != EOF)
+		putc(c, nf);
 	if (fp != collf)
 		Fclose(collf);
 	Fclose(fp);
@@ -987,13 +984,12 @@ makeheader(fp, hp)
  * Edit the message being collected on fp.
  * On return, make the edit file the new temp file.
  */
-static void
-mesedit(c, hp)
-	int c;
-	struct header *hp;
+static void 
+mesedit(int c, struct header *hp)
 {
 	sighandler_type sigint = safe_signal(SIGINT, SIG_IGN);
-	FILE *nf = run_editor(collf, (off_t)-1, c, 0, NULL, hp);
+	FILE *nf = run_editor(collf, (off_t)-1, c, 0, hp, NULL, SEND_MBOX,
+			sigint);
 
 	if (nf != NULL) {
 		if (hp) {
@@ -1005,7 +1001,7 @@ mesedit(c, hp)
 			collf = nf;
 		}
 	}
-	(void) safe_signal(SIGINT, sigint);
+	safe_signal(SIGINT, sigint);
 }
 
 /*
@@ -1014,47 +1010,46 @@ mesedit(c, hp)
  * New message collected from stdout.
  * Sh -c must return 0 to accept the new message.
  */
-static void
-mespipe(cmd)
-	char cmd[];
+static void 
+mespipe(char *cmd)
 {
 	FILE *nf;
 	sighandler_type sigint = safe_signal(SIGINT, SIG_IGN);
 	char *tempEdit;
 	char *shell;
 
-	if ((nf = Ftemp(&tempEdit, "Re", "w+", 0600, 1)) == (FILE *)NULL) {
+	if ((nf = Ftemp(&tempEdit, "Re", "w+", 0600, 1)) == NULL) {
 		perror(catgets(catd, CATSET, 66, "temporary mail edit file"));
 		goto out;
 	}
 	fflush(collf);
-	(void) unlink(tempEdit);
+	unlink(tempEdit);
 	Ftfree(&tempEdit);
 	/*
 	 * stdin = current message.
 	 * stdout = new message.
 	 */
 	if ((shell = value("SHELL")) == NULL)
-		shell = PATH_CSHELL;
+		shell = SHELL;
 	if (run_command(shell,
 	    0, fileno(collf), fileno(nf), "-c", cmd, NULL) < 0) {
-		(void) Fclose(nf);
+		Fclose(nf);
 		goto out;
 	}
 	if (fsize(nf) == 0) {
 		fprintf(stderr, catgets(catd, CATSET, 67,
 				"No bytes from \"%s\" !?\n"), cmd);
-		(void) Fclose(nf);
+		Fclose(nf);
 		goto out;
 	}
 	/*
 	 * Take new files.
 	 */
-	(void) fseek(nf, 0L, SEEK_END);
-	(void) Fclose(collf);
+	fseek(nf, 0L, SEEK_END);
+	Fclose(collf);
 	collf = nf;
 out:
-	(void) safe_signal(SIGINT, sigint);
+	safe_signal(SIGINT, sigint);
 }
 
 /*
@@ -1066,18 +1061,16 @@ out:
  * should shift over and 'f' if not.
  */
 static int
-forward(ms, fp, f)
-	char ms[];
-	FILE *fp;
-	int f;
+forward(char *ms, FILE *fp, int f)
 {
 	int *msgvec;
 	struct ignoretab *ig;
 	char *tabst;
+	enum sendaction	action;
 
 	/*LINTED*/
-	msgvec = (int *)salloc((msgcount+1) * sizeof *msgvec);
-	if (msgvec == (int *)NULL)
+	msgvec = (int *)salloc((msgCount+1) * sizeof *msgvec);
+	if (msgvec == NULL)
 		return(0);
 	if (getmsglist(ms, msgvec, 0) < 0)
 		return(0);
@@ -1095,13 +1088,14 @@ forward(ms, fp, f)
 	else if ((tabst = value("indentprefix")) == NULL)
 		tabst = "\t";
 	ig = upperchar(f) ? (struct ignoretab *)NULL : ignore;
+	action = upperchar(f) ? SEND_QUOTE_ALL : SEND_QUOTE;
 	printf(catgets(catd, CATSET, 69, "Interpolating:"));
 	for (; *msgvec != 0; msgvec++) {
 		struct message *mp = message + *msgvec - 1;
 
 		touch(mp);
 		printf(" %d", *msgvec);
-		if (send_message(mp, fp, ig, tabst, CONV_QUOTE, NULL) < 0) {
+		if (send(mp, fp, ig, tabst, action, NULL) < 0) {
 			perror(catgets(catd, CATSET, 70,
 					"temporary mail file"));
 			return(-1);
@@ -1115,9 +1109,8 @@ forward(ms, fp, f)
  * Print (continue) when continued after ^Z.
  */
 /*ARGSUSED*/
-static RETSIGTYPE
-collstop(s)
-	int s;
+static void 
+collstop(int s)
 {
 	sighandler_type old_action = safe_signal(s, SIG_DFL);
 	sigset_t nset;
@@ -1140,9 +1133,8 @@ collstop(s)
  * Then jump out of the collection loop.
  */
 /*ARGSUSED*/
-static RETSIGTYPE
-collint(s)
-	int s;
+static void 
+collint(int s)
 {
 	/*
 	 * the control flow is subtle, because we can be called from ~q.
@@ -1165,9 +1157,8 @@ collint(s)
 }
 
 /*ARGSUSED*/
-static RETSIGTYPE
-collhup(s)
-	int s;
+static void 
+collhup(int s)
 {
 	rewind(collf);
 	savedeadletter(collf);
@@ -1179,8 +1170,7 @@ collhup(s)
 }
 
 void
-savedeadletter(fp)
-	FILE *fp;
+savedeadletter(FILE *fp)
 {
 	FILE *dbuf;
 	int c;
@@ -1191,11 +1181,39 @@ savedeadletter(fp)
 	cp = getdeadletter();
 	c = umask(077);
 	dbuf = Fopen(cp, "a");
-	(void) umask(c);
-	if (dbuf == (FILE *)NULL)
+	umask(c);
+	if (dbuf == NULL)
 		return;
-	while ((c = sgetc(fp)) != EOF)
-		(void) sputc(c, dbuf);
+	while ((c = getc(fp)) != EOF)
+		putc(c, dbuf);
 	Fclose(dbuf);
 	rewind(fp);
 }
+
+static int
+putesc(const char *s, FILE *stream)
+{
+	int	n = 0;
+
+	while (s[0]) {
+		if (s[0] == '\\') {
+			if (s[1] == 't') {
+				putc('\t', stream);
+				n++;
+				s += 2;
+				continue;
+			}
+			if (s[1] == 'n') {
+				putc('\n', stream);
+				n++;
+				s += 2;
+				continue;
+			}
+		}
+		putc(s[0]&0377, stream);
+		n++;
+		s++;
+	}
+	putc('\n', stream);
+	return ++n;
+}
