@@ -1,7 +1,7 @@
 /*
  * Nail - a mail user agent derived from Berkeley Mail.
  *
- * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ * Copyright (c) 2000-2004 Gunnar Ritter, Freiburg i. Br., Germany.
  */
 /*
  * Copyright (c) 1980, 1993
@@ -35,7 +35,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *	Sccsid @(#)def.h	2.20 (gritter) 1/13/03
+ *	Sccsid @(#)def.h	2.93 (gritter) 12/25/04
  */
 
 /*
@@ -44,21 +44,11 @@
  * Author: Kurt Shoens (UCB) March 25, 1978
  */
 
-#include "config.h"
-
 #if !defined (NI_MAXHOST) || (NI_MAXHOST) < 1025
 #undef	NI_MAXHOST
 #define	NI_MAXHOST	1025
 #endif
 
-#if defined (HAVE_GETADDRINFO) && defined (HAVE_GETNAMEINFO)
-#define	HAVE_IPv6_FUNCS
-#endif
-
-#ifndef	FD_CLOEXEC
-#define	FD_CLOEXEC	1
-#endif
-
 #define	APPEND				/* New mail goes to end of mailbox */
 
 #define	ESCAPE		'~'		/* Default escape for sending */
@@ -80,21 +70,13 @@
 
 #define	equal(a, b)	(strcmp(a,b)==0)/* A nice function to string compare */
 
-#ifndef	__P
-#ifdef	__STDC__
-#define	__P(a)	a
-#else
-#define	__P(a)	()
-#endif
-#endif
-
 #ifdef	HAVE_CATGETS
 #define	CATSET	1
 #else	/* !HAVE_CATGETS */
 #define	catgets(a, b, c, d)	(d)
 #endif	/* !HAVE_CATGETS */
 
-typedef RETSIGTYPE (*sighandler_type) __P((int));
+typedef void (*sighandler_type)(int);
 
 enum okay {
 	STOP = 0,
@@ -113,9 +95,6 @@ enum mimeenc {
 enum conversion {
 	CONV_NONE,			/* no conversion */
 	CONV_7BIT,			/* no conversion, is 7bit */
-	CONV_TODISP,			/* convert in displayable form */
-	CONV_TOFILE,			/* convert for saving to a file */
-	CONV_QUOTE,			/* first part body only */
 	CONV_FROMQP,			/* convert from quoted-printable */
 	CONV_TOQP,			/* convert to quoted-printable */
 	CONV_FROMB64,			/* convert from base64 */
@@ -126,13 +105,31 @@ enum conversion {
 	CONV_TOHDR_A			/* convert addresses for header */
 };
 
+enum sendaction {
+	SEND_MBOX,			/* no conversion to perform */
+	SEND_TODISP,			/* convert to displayable form */
+	SEND_TODISP_ALL,		/* same, include all MIME parts */
+	SEND_SHOW,			/* convert to 'show' command form */
+	SEND_TOSRCH,			/* convert for IMAP SEARCH */
+	SEND_TOFLTR,			/* convert for junk mail filtering */
+	SEND_TOFILE,			/* convert for saving body to a file */
+	SEND_QUOTE,			/* convert for quoting */
+	SEND_QUOTE_ALL,			/* same, include all MIME parts */
+	SEND_DECRYPT			/* decrypt */
+};
+
 enum mimecontent {
 	MIME_UNKNOWN,			/* unknown content */
 	MIME_SUBHDR,			/* inside a multipart subheader */
 	MIME_822,			/* message/rfc822 content */
-	MIME_MESSAGE,			/* message/ content */
-	MIME_TEXT,			/* text/ content */
-	MIME_MULTI,			/* multipart/ content */
+	MIME_MESSAGE,			/* other message/ content */
+	MIME_TEXT_PLAIN,		/* text/plain content */
+	MIME_TEXT_HTML,			/* text/html content */
+	MIME_TEXT,			/* other text/ content */
+	MIME_ALTERNATIVE,		/* multipart/alternative content */
+	MIME_DIGEST,			/* multipart/digest content */
+	MIME_MULTI,			/* other multipart/ content */
+	MIME_PKCS7,			/* PKCS7 content */
 	MIME_DISCARD			/* content is discarded */
 };
 
@@ -145,9 +142,10 @@ enum mimeclean {
 };
 
 enum tdflags {
-	TD_NONE	= 0,		/* no display conversion */
-	TD_ISPR	= 01,		/* use isprint() checks */
-	TD_ICONV= 02		/* use iconv() */
+	TD_NONE		= 0,	/* no display conversion */
+	TD_ISPR		= 01,	/* use isprint() checks */
+	TD_ICONV	= 02,	/* use iconv() */
+	TD_DELCTRL	= 04	/* delete control characters */
 };
 
 struct str {
@@ -158,37 +156,70 @@ struct str {
 enum protocol {
 	PROTO_FILE,			/* refers to a local file */
 	PROTO_POP3,			/* is a pop3 server string */
+	PROTO_IMAP,			/* is an imap server string */
+	PROTO_MAILDIR,			/* refers to a maildir folder */
 	PROTO_UNKNOWN			/* unknown protocol */
 };
 
+struct sock {				/* data associated with a socket */
+	int	s_fd;			/* file descriptor */
+#ifdef	USE_SSL
+	int	s_use_ssl;		/* SSL is used */
+#if defined (USE_NSS)
+	void	*s_prfd;		/* NSPR file descriptor */
+#elif defined (USE_OPENSSL)
+	void	*s_ssl;			/* SSL object */
+	void	*s_ctx;			/* SSL context object */
+#endif	/* SSL library specific */
+#endif	/* USE_SSL */
+	char	*s_wbuf;		/* for buffered writes */
+	int	s_wbufsize;		/* allocated size of s_buf */
+	int	s_wbufpos;		/* position of first empty data byte */
+	char	s_rbuf[LINESIZE+1];	/* for buffered reads */
+	char	*s_rbufptr;		/* read pointer to s_rbuf */
+	int	s_rsz;			/* size of last read in s_rbuf */
+	char	*s_desc;		/* description of error messages */
+	void	(*s_onclose)(void);	/* execute on close */
+};
+
 struct mailbox {
-	int mb_sock;			/* socket file descriptor */
-	char *mb_ptr;			/* read pointer to mb_buf */
-	int mb_sz;			/* size of last read in mb_buf */
+	struct sock	mb_sock;	/* socket structure */
 	enum {
-		MB_NONE = 00,		/* no reply expected */
-		MB_COMD = 01,		/* command reply expected */
-		MB_MULT = 02		/* multiline reply expected */
+		MB_NONE		= 000,	/* no reply expected */
+		MB_COMD		= 001,	/* command reply expected */
+		MB_MULT		= 002,	/* multiline reply expected */
+		MB_PREAUTH	= 004,	/* not in authenticated state */
+		MB_BYE		= 010	/* may accept a BYE state */
 	} mb_active;
-	char mb_buf[LINESIZE+1];	/* read buffer */
 	FILE *mb_itf;			/* temp file with messages, read open */
 	FILE *mb_otf;			/* same, write open */
+	char *mb_sorted;		/* sort method */
 	enum {
 		MB_VOID,		/* no type (e. g. connection failed) */
 		MB_FILE,		/* local file */
-		MB_POP3			/* POP3 mailbox */
+		MB_POP3,		/* POP3 mailbox */
+		MB_IMAP,		/* IMAP mailbox */
+		MB_MAILDIR,		/* maildir folder */
+		MB_CACHE		/* cached mailbox */
 	} mb_type;			/* type of mailbox */
 	enum {
 		MB_DELE = 01,		/* may delete messages in mailbox */
 		MB_EDIT = 02		/* may edit messages in mailbox */
 	} mb_perm;
-#ifdef	USE_SSL
-	void *mb_ssl;			/* SSL object */
-	void *mb_ctx;			/* SSL context object */
-#endif	/* USE_SSL */
+	int mb_compressed;		/* is a compressed mbox file */
+	int mb_threaded;		/* mailbox has been threaded */
+	enum mbflags {
+		MB_NOFLAGS	= 000,
+		MB_UIDPLUS	= 001	/* supports IMAP UIDPLUS */
+	} mb_flags;
+	unsigned long	mb_uidvalidity;	/* IMAP unique identifier validity */
+	char	*mb_imap_account;	/* name of current IMAP account */
+	char	*mb_imap_mailbox;	/* name of current IMAP mailbox */
+	char	*mb_cache_directory;	/* name of cache directory */
 };
 
 enum needspec {
+	NEED_UNSPEC,			/* unspecified need, don't fetch */
 	NEED_HEADER,			/* need the header of a message */
 	NEED_BODY			/* need header and body of a message */
 };
@@ -200,7 +231,7 @@ enum havespec {
 };
 
 /*
- * flag bits.
+ * flag bits. Attention: Flags that are used in cache.c may not change.
  */
 enum mflag {
 	MUSED		= (1<<0),	/* entry is used, but this bit isn't */
@@ -214,19 +245,73 @@ enum mflag {
 	MREAD		= (1<<8),	/* message has been read sometime. */
 	MSTATUS		= (1<<9),	/* message status has changed */
 	MBOX		= (1<<10),	/* Send this to mbox, regardless */
-	MNOFROM		= (1<<11)	/* no From line */
+	MNOFROM		= (1<<11),	/* no From line */
+	MHIDDEN		= (1<<12),	/* message is hidden to user */
+	MFULLYCACHED	= (1<<13),	/* message is completely cached */
+	MBOXED		= (1<<14),	/* message has been sent to mbox */
+	MUNLINKED	= (1<<15),	/* message was unlinked from cache */
+	MNEWEST		= (1<<16),	/* message is very new (newmail) */
+	MFLAG		= (1<<17),	/* message has been flagged recently */
+	MUNFLAG		= (1<<18),	/* message has been unflagged */
+	MFLAGGED	= (1<<19),	/* message is `flagged' */
+	MANSWER		= (1<<20),	/* message has been answered recently */
+	MUNANSWER	= (1<<21),	/* message has been unanswered */
+	MANSWERED	= (1<<22),	/* message is `answered' */
+	MDRAFT		= (1<<23),	/* message has been drafted recently */
+	MUNDRAFT	= (1<<24),	/* message has been undrafted */
+	MDRAFTED	= (1<<25),	/* message is marked as `draft' */
+	MKILL		= (1<<26),	/* message has been killed */
+	MOLDMARK	= (1<<27),	/* messages was marked previously */
+	MJUNK		= (1<<28)	/* message is classified as junk */
+};
+
+struct mimepart {
+	enum mflag	m_flag;		/* flags */
+	enum havespec	m_have;		/* downloaded parts of the part */
+	int	m_block;		/* block number of this part */
+	size_t	m_offset;		/* offset in block of part */
+	size_t	m_size;			/* Bytes in the part */
+	size_t	m_xsize;		/* Bytes in the full part */
+	long	m_lines;		/* Lines in the message */
+	long	m_xlines;		/* Lines in the full message */
+	time_t	m_time;			/* time the message was sent */
+	char	*m_from;		/* message sender */
+	struct mimepart	*m_nextpart;	/* next part at same level */
+	struct mimepart	*m_multipart;	/* parts of multipart */
+	struct mimepart	*m_parent;	/* enclosing multipart part */
+	char	*m_ct_type;		/* content-type */
+	char	*m_ct_type_plain;	/* content-type without specs */
+	enum mimecontent	m_mimecontent;	/* same in enum */
+	char	*m_charset;		/* charset */
+	char	*m_ct_transfer_enc;	/* content-transfer-encoding */
+	enum mimeenc	m_mimeenc;	/* same in enum */
+	char	*m_partstring;		/* part level string */
+	char	*m_filename;		/* attachment filename */
 };
 
 struct message {
 	enum mflag	m_flag;		/* flags */
-	time_t	m_time;			/* time in the 'Date:' field */
+	enum havespec	m_have;		/* downloaded parts of the message */
 	int	m_block;		/* block number of this message */
 	size_t	m_offset;		/* offset in block of message */
 	size_t	m_size;			/* Bytes in the message */
 	size_t	m_xsize;		/* Bytes in the full message */
-	int	m_lines;		/* Lines in the message */
-	int	m_xlines;		/* Lines in the full message */
-	enum havespec	m_have;		/* downloaded parts of the message */
+	long	m_lines;		/* Lines in the message */
+	long	m_xlines;		/* Lines in the full message */
+	time_t	m_time;			/* time the message was sent */
+	time_t	m_date;			/* time in the 'Date' field */
+	unsigned	m_idhash;	/* hash on Message-ID for threads */
+	unsigned long	m_uid;		/* IMAP unique identifier */
+	struct message	*m_child;	/* first child of this message */
+	struct message	*m_younger;	/* younger brother of this message */
+	struct message	*m_elder;	/* elder brother of this message */
+	struct message	*m_parent;	/* parent of this message */
+	unsigned	m_level;	/* thread level of message */
+	long		m_threadpos;	/* position in threaded display */
+	float		m_score;	/* score of message */
+	char	*m_maildir_file;	/* original maildir file of msg */
+	unsigned	m_maildir_hash;	/* hash of file name in maildir sub */
+	int	m_collapsed;		/* collapsed thread information */
 };
 
 /*
@@ -245,6 +330,7 @@ enum argtype {
 	RAWLIST	= 2,		/* Shell string list */
 	NOLIST	= 3,		/* Just plain 0 */
 	NDMLIST	= 4,		/* Message list, no defaults */
+	ECHOLIST= 5,		/* Like raw list, but keep quote chars */
 	P	= 040,		/* Autoprint dot after command */
 	I	= 0100,		/* Interactive command bit */
 	M	= 0200,		/* Legal from send mode bit */
@@ -259,8 +345,8 @@ enum argtype {
  * Oft-used mask values
  */
 
-#define	MMNORM		(MDELETED|MSAVED)/* Look at both save and delete bits */
-#define	MMNDEL		MDELETED	/* Look only at deleted bit */
+#define	MMNORM		(MDELETED|MSAVED|MHIDDEN)/* Look at both save and delete bits */
+#define	MMNDEL		(MDELETED|MHIDDEN)	/* Look only at deleted bit */
 
 /*
  * Format of the command description table.
@@ -269,7 +355,7 @@ enum argtype {
  */
 struct cmd {
 	char	*c_name;		/* Name of command */
-	int	(*c_func) __P((void *));/* Implementor of the command */
+	int	(*c_func)(void *);	/* Implementor of the command */
 	enum argtype	c_argtype;	/* Type of arglist (see below) */
 	short	c_msgflag;		/* Required flags of messages */
 	short	c_msgmask;		/* Relevant flags of messages */
@@ -306,10 +392,14 @@ enum gfield {
 	/*	  1024 */	/* unused */
 	GIDENT	= 2048,		/* From:, Reply-To: and Organization header */
 	GREF	= 4096,		/* References: header */
-	GDATE	= 8192		/* Date: header */
+	GDATE	= 8192,		/* Date: header */
+	GFULL	= 16384,	/* include full names */
+	GSKIN	= 32768,	/* skin names */
+	GREPLYTO= 65536,	/* a Reply-To field */
+	GFILES	= 131072	/* include filename addresses */
 };
 
-#define	GMASK	(GTO|GSUBJECT|GCC|GBCC)	/* Mask of places from whence */
+#define	GMASK	(GTO|GSUBJECT|GCC|GBCC|GREPLYTO)/* Mask of places from whence */
 
 /*
  * Structure used to pass about the current
@@ -321,6 +411,7 @@ struct header {
 	char *h_subject;		/* Subject string */
 	struct name *h_cc;		/* Carbon copies string */
 	struct name *h_bcc;		/* Blind carbon copies */
+	struct name *h_replyto;		/* Reply-To address */
 	struct name *h_ref;		/* References */
 	struct name *h_smopts;		/* Sendmail options */
 	struct attachment *h_attach;	/* MIME attachments */
@@ -335,8 +426,9 @@ struct header {
 struct name {
 	struct	name *n_flink;		/* Forward link in list. */
 	struct	name *n_blink;		/* Backward list link */
-	short	n_type;			/* From which list it came */
+	enum gfield	n_type;		/* From which list it came */
 	char	*n_name;		/* This fella's name */
+	char	*n_fullname;		/* Sometimes, name including comment */
 };
 
 /*
@@ -376,12 +468,6 @@ struct grouphead {
 	struct	group *g_list;		/* Users in group. */
 };
 
-#define	NIL	((struct name *) 0)	/* The nil pointer for namelists */
-#define	NONE	((struct cmd *) 0)	/* The nil pointer to command tab */
-#define	NOVAR	((struct var *) 0)	/* The nil pointer to variables */
-#define	NOGRP	((struct grouphead *) 0)/* The nil grouphead pointer */
-#define	NOGE	((struct group *) 0)	/* The nil group pointer */
-
 /*
  * Structure of the hash table of ignored header fields
  */
@@ -411,11 +497,11 @@ enum ltoken {
 	TPLUS	= 10,			/* A '+' */
 	TERROR	= 11,			/* A lexical error */
 	TCOMMA	= 12,			/* A ',' */
-	TSEMI	= 13			/* A ';' */
+	TSEMI	= 13,			/* A ';' */
+	TBACK	= 14			/* A '`' */
 };
 
 #define	REGDEP	2			/* Maximum regret depth. */
-#define	STRINGLEN	1024		/* Maximum length of string token */
 
 /*
  * Constants for conditional commands.  These describe whether
@@ -449,33 +535,35 @@ struct shortcut {
  * Locale-independent character classes.
  */
 enum {
-	C_CNTRL	= 0001,
-	C_BLANK	= 0002,
+	C_CNTRL	= 0000,
+	C_BLANK	= 0001,
+	C_WHITE = 0002,
 	C_SPACE	= 0004,
 	C_PUNCT	= 0010,
-	C_DIGIT	= 0020,
-	C_UPPER	= 0040,
-	C_LOWER	= 0100,
-	C_WHITE = 0200
+	C_OCTAL	= 0020,
+	C_DIGIT	= 0040,
+	C_UPPER	= 0100,
+	C_LOWER	= 0200
 };
 
 extern const unsigned char	class_char[];
 
 #define	asciichar(c) ((unsigned)(c) <= 0177)
-#define	alnumchar(c) (asciichar(c)&&(class_char[c]&(C_DIGIT|C_UPPER|C_LOWER)))
+#define	alnumchar(c) (asciichar(c)&&(class_char[c]&\
+			(C_DIGIT|C_OCTAL|C_UPPER|C_LOWER)))
 #define	alphachar(c) (asciichar(c)&&(class_char[c]&(C_UPPER|C_LOWER)))
 #define	blankchar(c) (asciichar(c)&&(class_char[c]&(C_BLANK)))
-#define	cntrlchar(c) (asciichar(c)&&(class_char[c]&(C_CNTRL)))
-#define	digitchar(c) (asciichar(c)&&(class_char[c]&(C_DIGIT)))
+#define	cntrlchar(c) (asciichar(c)&&(class_char[c]==C_CNTRL))
+#define	digitchar(c) (asciichar(c)&&(class_char[c]&(C_DIGIT|C_OCTAL)))
 #define	lowerchar(c) (asciichar(c)&&(class_char[c]&(C_LOWER)))
 #define	punctchar(c) (asciichar(c)&&(class_char[c]&(C_PUNCT)))
 #define	spacechar(c) (asciichar(c)&&(class_char[c]&(C_BLANK|C_SPACE|C_WHITE)))
 #define	upperchar(c) (asciichar(c)&&(class_char[c]&(C_UPPER)))
 #define	whitechar(c) (asciichar(c)&&(class_char[c]&(C_BLANK|C_WHITE)))
+#define	octalchar(c) (asciichar(c)&&(class_char[c]&(C_OCTAL)))
 
 #define	upperconv(c) (lowerchar(c) ? (c)-'a'+'A' : (c))
 #define	lowerconv(c) (upperchar(c) ? (c)-'A'+'a' : (c))
-
 /*	RFC 822, 3.2.	*/
 #define	fieldnamechar(c) (asciichar(c)&&(c)>040&&(c)!=0177&&(c)!=':')
 
@@ -485,8 +573,8 @@ extern const unsigned char	class_char[];
  * for read/write.
  */
 #define trunc(stream) {							\
-	(void)fflush(stream); 						\
-	(void)ftruncate(fileno(stream), (off_t)ftell(stream));		\
+	fflush(stream); 						\
+	ftruncate(fileno(stream), (off_t)ftell(stream));		\
 }
 
 /*
@@ -506,14 +594,38 @@ extern const unsigned char	class_char[];
  * glibc uses the slow thread-safe getc() even if _REENTRANT is not
  * defined. Work around it.
  */
-#if defined (HAVE_GETC_UNLOCKED) && defined (__GLIBC__)
-#define	sgetc(c)	getc_unlocked(c)
-#else	/* !HAVE_GETC_UNLOCKED */
-#define	sgetc(c)	getc(c)
-#endif	/* !HAVE_GETC_UNLOCKED */
+#ifdef	__GLIBC__
+#undef	getc
+#define	getc(c)		getc_unlocked(c)
+#undef	putc
+#define	putc(c, f)	putc_unlocked(c, f)
+#undef	putchar
+#define	putchar(c)	putc_unlocked((c), stdout)
+#endif	/* __GLIBC__ */
 
-#if defined (HAVE_PUTC_UNLOCKED) && defined (__GLIBC__)
-#define	sputc(c, f)	putc_unlocked(c, f)
-#else	/* !HAVE_PUTC_UNLOCKED */
-#define	sputc(c, f)	putc(c, f)
-#endif	/* !HAVE_PUTC_UNLOCKED */
+#define	CBAD		(-15555)
+
+#define	smin(a, b)	((a) < (b) ? (a) : (b))
+#define	smax(a, b)	((a) < (b) ? (b) : (a))
+
+/*
+ * For saving the current directory and later returning.
+ */
+#ifdef	HAVE_FCHDIR
+struct	cw {
+	int	cw_fd;
+};
+#else	/* !HAVE_FCHDIR */
+struct	cw {
+	char	cw_wd[PATHSIZE];
+};
+#endif	/* !HAVE_FCHDIR */
+
+#ifdef	USE_SSL
+enum ssl_vrfy_level {
+	VRFY_IGNORE,
+	VRFY_WARN,
+	VRFY_ASK,
+	VRFY_STRICT
+};
+#endif	/* USE_SSL */
