@@ -1,7 +1,7 @@
 /*
  * Nail - a mail user agent derived from Berkeley Mail.
  *
- * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ * Copyright (c) 2000-2004 Gunnar Ritter, Freiburg i. Br., Germany.
  */
 /*
  * Copyright (c) 1980, 1993
@@ -35,7 +35,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *	Sccsid @(#)glob.h	2.9 (gritter) 11/1/02
+ *	Sccsid @(#)glob.h	2.22 (gritter) 12/2/04
  */
 
 /*
@@ -50,7 +50,7 @@
 #  define _E	extern
 #endif
 
-_E int	msgcount;			/* Count of messages read in */
+_E int	msgCount;			/* Count of messages read in */
 _E int	rcvmode;			/* True if receiving mail */
 _E int	sawcom;				/* Set after first command */
 _E int	Iflag;				/* -I show Newsgroups: field */
@@ -65,15 +65,17 @@ _E struct mailbox mb;			/* Current mailbox */
 _E int	image;				/* File descriptor for image of msg */
 _E FILE	*input;				/* Current command input file */
 _E char	mailname[PATHSIZE];		/* Name of current file */
+_E char	mboxname[PATHSIZE];		/* Name of mbox */
 _E char	prevfile[PATHSIZE];		/* Name of previous file */
 _E char	*homedir;			/* Path name of home directory */
 _E char *progname;			/* our name */
 _E char	*myname;			/* My login name */
-_E const char *version;			/* version string */
+extern const char *version;		/* version string */
 _E off_t mailsize;			/* Size of system mailbox */
 _E struct message *dot;			/* Pointer to current message */
 _E struct message *prevdot;		/* Previous current message */
 _E struct message *message;		/* The actual message structure */
+_E struct message *threadroot;		/* first threaded message */
 _E int msgspace;			/* Number of allocated struct m */
 _E struct var *variables[HSHSIZE];	/* Pointer to active var list */
 _E struct grouphead *groups[HSHSIZE];	/* Pointer to active groups */
@@ -82,6 +84,7 @@ _E struct ignoretab ignore[2];		/* ignored and retained fields
 _E struct ignoretab saveignore[2];	/* ignored and retained fields
 					   on save to folder */
 _E struct ignoretab allignore[2];	/* special, ignore all headers */
+_E struct ignoretab fwdignore[2];	/* fields to ignore for forwarding */
 _E char	**altnames;			/* List of alternate names for user */
 _E int	debug;				/* Debug flag set */
 _E int	scrnwidth;			/* Screen width, or best guess */
@@ -97,6 +100,14 @@ _E int	tildeflag;			/* enable tilde escapes */
 _E char	*uflag;				/* name given with -u option */
 _E struct shortcut	*shortcuts;	/* list of shortcuts */
 _E int	mb_cur_max;			/* value of MB_CUR_MAX */
+_E int	imap_created_mailbox;		/* hack to get feedback from imap */
+_E int	unset_allow_undefined;		/* allow to unset undefined variables */
+_E int	inhook;				/* currently executing a hook */
+_E int	starting;			/* still in startup code */
+
+#ifdef	USE_SSL
+_E enum ssl_vrfy_level	ssl_vrfy_level;	/* SSL verification level */
+#endif
 
 #ifdef	HAVE_ICONV
 _E iconv_t iconvd;
@@ -115,7 +126,10 @@ extern const char *month_names[];
 #include <setjmp.h>
 
 _E sigjmp_buf	srbuf;
-
+_E int		interrupts;
+_E sighandler_type	handlerstacktop;
+#define	handlerpush(f)	(savedtop = handlerstacktop, handlerstacktop = (f))
+#define	handlerpop()	(handlerstacktop = savedtop)
 
 /*
  * The pointers for the string allocation routines,
