@@ -1,7 +1,7 @@
 /*
  * Nail - a mail user agent derived from Berkeley Mail.
  *
- * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ * Copyright (c) 2000-2004 Gunnar Ritter, Freiburg i. Br., Germany.
  */
 /*
  * Copyright (c) 1980, 1993
@@ -38,39 +38,51 @@
 
 #ifndef lint
 #ifdef	DOSCCS
-static char sccsid[] = "@(#)list.c	2.13 (gritter) 3/28/03";
+static char sccsid[] = "@(#)list.c	2.57 (gritter) 12/30/04";
 #endif
 #endif /* not lint */
 
 #include "rcv.h"
 #include <ctype.h>
 #include "extern.h"
+#ifdef	HAVE_WCTYPE_H
+#include <wctype.h>
+#endif	/* HAVE_WCTYPE_H */
 
 /*
  * Mail -- a mail program
  *
  * Message list handling.
  */
-static char	**add_to_namelist __P((char ***, size_t *, char **, char *));
-static int	markall __P((char [], int));
-static int	evalcol __P((int));
-static int	check __P((int, int));
-static int	scan __P((char **));
-static void	regret __P((int));
-static void	scaninit __P((void));
-static int	matchsender __P((char *, int, int));
-static int	matchmid __P((char *, int));
-static int	matchsubj __P((char *, int));
-static void	mark __P((int));
-static void	unmark __P((int));
-static int	metamess __P((int, int));
+
+enum idtype {
+	ID_REFERENCES,
+	ID_IN_REPLY_TO
+};
+
+static char **add_to_namelist(char ***namelist, size_t *nmlsize,
+		char **np, char *string);
+static int markall(char *buf, int f);
+static int evalcol(int col);
+static int check(int mesg, int f);
+static int scan(char **sp);
+static void regret(int token);
+static void scaninit(void);
+static int matchsender(char *str, int mesg, int allnet);
+static int matchmid(char *id, enum idtype idtype, int mesg);
+static int matchsubj(char *str, int mesg);
+static void unmark(int mesg);
+static int metamess(int meta, int f);
+
+static size_t	STRINGLEN;
 
 static int	lexnumber;		/* Number of TNUMBER from scan() */
-static char	lexstring[STRINGLEN];	/* String from TSTRING, scan() */
+static char	*lexstring;		/* String from TSTRING, scan() */
 static int	regretp;		/* Pointer to TOS of regret tokens */
 static int	regretstack[REGDEP];	/* Stack of regretted tokens */
 static char	*string_stack[REGDEP];	/* Stack of regretted strings */
 static int	numberstack[REGDEP];	/* Stack of regretted numbers */
+static int	threadflag;		/* mark entire threads */
 
 /*
  * Convert the user string of message numbers and
@@ -78,24 +90,41 @@ static int	numberstack[REGDEP];	/* Stack of regretted numbers */
  *
  * Returns the count of messages picked up or -1 on error.
  */
-int
-getmsglist(buf, vector, flags)
-	char *buf;
-	int *vector, flags;
+int 
+getmsglist(char *buf, int *vector, int flags)
 {
 	int *ip;
 	struct message *mp;
+	int	mc;
 
-	if (msgcount == 0) {
+	if (msgCount == 0) {
 		*vector = 0;
 		return 0;
 	}
 	if (markall(buf, flags) < 0)
 		return(-1);
 	ip = vector;
-	for (mp = &message[0]; mp < &message[msgcount]; mp++)
-		if (mp->m_flag & MMARK)
-			*ip++ = mp - &message[0] + 1;
+	if (inhook & 2) {
+		mc = 0;
+		for (mp = &message[0]; mp < &message[msgCount]; mp++)
+			if (mp->m_flag & MMARK) {
+				if ((mp->m_flag & MNEWEST) == 0)
+					unmark(mp - &message[0] + 1);
+				else
+					mc++;
+			}
+		if (mc == 0)
+			return -1;
+	}
+	if (mb.mb_threaded == 0) {
+		for (mp = &message[0]; mp < &message[msgCount]; mp++)
+			if (mp->m_flag & MMARK)
+				*ip++ = mp - &message[0] + 1;
+	} else {
+		for (mp = threadroot; mp; mp = next_in_thread(mp))
+			if (mp->m_flag & MMARK)
+				*ip++ = mp - &message[0] + 1;
+	}
 	*ip = 0;
 	return(ip - vector);
 }
@@ -115,6 +144,11 @@ getmsglist(buf, vector, flags)
 #define	CMUNREAD	04		/* Unread messages */
 #define	CMDELETED	010		/* Deleted messages */
 #define	CMREAD		020		/* Read messages */
+#define	CMFLAG		040		/* Flagged messages */
+#define	CMANSWER	0100		/* Answered messages */
+#define	CMDRAFT		0200		/* Draft messages */
+#define	CMKILL		0400		/* Killed messages */
+#define	CMJUNK		01000		/* Junk messages */
 
 /*
  * The following table describes the letters which can follow
@@ -132,17 +166,18 @@ static struct coltab {
 	{ 'u',		CMUNREAD,	MREAD,		0 },
 	{ 'd',		CMDELETED,	MDELETED,	MDELETED },
 	{ 'r',		CMREAD,		MREAD,		MREAD },
+	{ 'f',		CMFLAG,		MFLAGGED,	MFLAGGED },
+	{ 'a',		CMANSWER,	MANSWERED,	MANSWERED },
+	{ 't',		CMDRAFT,	MDRAFTED,	MDRAFTED },
+	{ 'k',		CMKILL,		MKILL,		MKILL },
+	{ 'j',		CMJUNK,		MJUNK,		MJUNK },
 	{ 0,		0,		0,		0 }
 };
 
 static	int	lastcolmod;
 
 static char **
-add_to_namelist(namelist, nmlsize, np, string)
-	char ***namelist;
-	size_t *nmlsize;
-	char **np;
-	char *string;
+add_to_namelist(char ***namelist, size_t *nmlsize, char **np, char *string)
 {
 	size_t idx;
 
@@ -156,25 +191,30 @@ add_to_namelist(namelist, nmlsize, np, string)
 
 #define	markall_ret(i)		{ \
 					retval = i; \
+					ac_free(lexstring); \
 					goto out; \
 				}
 
-static int
-markall(buf, f)
-	char buf[];
-	int f;
+static int 
+markall(char *buf, int f)
 {
 	char **np, **nq;
-	int i, retval;
-	struct message *mp;
+	int i, retval, gotheaders;
+	struct message *mp, *mx;
 	char **namelist, *bufp, *id = NULL, *cp;
-	int tok, beg, mc, star, other, valdot, colmod, colresult;
+	int tok, beg, mc, star, other, valdot, colmod, colresult, topen, tback;
 	size_t nmlsize;
+	enum idtype	idtype = ID_REFERENCES;
 
+	lexstring = ac_alloc(STRINGLEN = 2 * strlen(buf) + 1);
 	valdot = dot - &message[0] + 1;
 	colmod = 0;
-	for (i = 1; i <= msgcount; i++)
+	for (i = 1; i <= msgCount; i++) {
+		message[i-1].m_flag &= ~MOLDMARK;
+		if (message[i-1].m_flag & MMARK)
+			message[i-1].m_flag |= MOLDMARK;
 		unmark(i);
+	}
 	bufp = buf;
 	mc = 0;
 	namelist = smalloc((nmlsize = 8) * sizeof *namelist);
@@ -184,6 +224,9 @@ markall(buf, f)
 	star = 0;
 	other = 0;
 	beg = 0;
+	topen = 0;
+	tback = 0;
+	gotheaders = 0;
 	while (tok != TEOL) {
 		switch (tok) {
 		case TNUMBER:
@@ -198,9 +241,23 @@ number:
 			if (beg != 0) {
 				if (check(lexnumber, f))
 					markall_ret(-1)
-				for (i = beg; i <= lexnumber; i++)
-					if (f == MDELETED || (message[i - 1].m_flag & MDELETED) == 0)
-						mark(i);
+				i = beg;
+				while (mb.mb_threaded ? 1 : i <= lexnumber) {
+					if (!(message[i-1].m_flag&MHIDDEN) &&
+							(f == MDELETED ||
+							 (message[i-1].m_flag &
+							  MDELETED) == 0))
+						mark(i, f);
+					if (mb.mb_threaded) {
+						if (i == lexnumber)
+							break;
+						mx = next_in_thread(&message[i-1]);
+						if (mx == NULL)
+							markall_ret(-1)
+						i = mx-message+1;
+					} else
+						i++;
+				}
 				beg = 0;
 				break;
 			}
@@ -210,7 +267,7 @@ number:
 			tok = scan(&bufp);
 			regret(tok);
 			if (tok != TDASH) {
-				mark(beg);
+				mark(beg, f);
 				beg = 0;
 			}
 			break;
@@ -223,29 +280,43 @@ number:
 			}
 			i = valdot;
 			do {
-				i++;
-				if (i > msgcount) {
+				if (mb.mb_threaded) {
+					mx = next_in_thread(&message[i-1]);
+					i = mx ? mx-message+1 : msgCount+1;
+				} else
+					i++;
+				if (i > msgCount) {
 					printf(catgets(catd, CATSET, 114,
 						"Referencing beyond EOF\n"));
 					markall_ret(-1)
 				}
-			} while ((message[i - 1].m_flag & MDELETED) != f);
-			mark(i);
+			} while (message[i-1].m_flag == MHIDDEN ||
+					(message[i-1].m_flag & MDELETED) != f ||
+					message[i-1].m_flag & MKILL);
+			mark(i, f);
 			break;
 
 		case TDASH:
 			if (beg == 0) {
 				i = valdot;
 				do {
-					i--;
+					if (mb.mb_threaded) {
+						mx = prev_in_thread(
+								&message[i-1]);
+						i = mx ? mx-message+1 : 0;
+					} else
+						i--;
 					if (i <= 0) {
 						printf(catgets(catd, CATSET,
 							115,
 						"Referencing before 1\n"));
 						markall_ret(-1)
 					}
-				} while ((message[i - 1].m_flag & MDELETED) != f);
-				mark(i);
+				} while (message[i-1].m_flag & MHIDDEN ||
+						(message[i-1].m_flag & MDELETED)
+							!= f ||
+						message[i-1].m_flag & MKILL);
+				mark(i, f);
 			}
 			break;
 
@@ -271,6 +342,12 @@ number:
 						np, savestr(lexstring));
 			break;
 
+		case TOPEN:
+			if (imap_search(lexstring, f) == STOP)
+				markall_ret(-1)
+			topen++;
+			break;
+
 		case TDOLLAR:
 		case TUP:
 		case TDOT:
@@ -280,6 +357,18 @@ number:
 				markall_ret(-1)
 			goto number;
 
+		case TBACK:
+			tback = 1;
+			for (i = 1; i <= msgCount; i++) {
+				if (message[i-1].m_flag&MHIDDEN ||
+						(message[i-1].m_flag&MDELETED)
+							!= f)
+					continue;
+				if (message[i-1].m_flag&MOLDMARK)
+					mark(i, f);
+			}
+			break;
+
 		case TSTAR:
 			if (other) {
 				printf(catgets(catd, CATSET, 118,
@@ -290,17 +379,22 @@ number:
 			break;
 
 		case TCOMMA:
-			if ((cp = hfield("in-reply-to", dot)) != NULL)
-				id = savestr(cp);
-			else if ((cp = hfield("references", dot)) != NULL) {
-				struct name *n;
-
-				if ((n = extract(cp, GREF)) != NULL) {
-					while (n->n_flink != NULL)
-						n = n->n_flink;
-					id = savestr(n->n_name);
+			if (mb.mb_type == MB_IMAP && gotheaders++ == 0)
+				imap_getheaders(1, msgCount);
+			if ((cp = hfield("references", dot)) != NULL) {
+				struct name	*np;
+				if ((np = extract(cp, GREF)) != NULL) {
+					while (np->n_flink != NULL)
+						np = np->n_flink;
+					id = savestr(np->n_name);
+					idtype = ID_REFERENCES;
 				}
 			}
+			if (id == NULL && (cp = hfield("in-reply-to", dot))
+					!= NULL) {
+				id = savestr(cp);
+				idtype = ID_IN_REPLY_TO;
+			}
 			if (id == NULL) {
 				printf(catgets(catd, CATSET, 227,
 		"Cannot determine parent Message-ID of the current message\n"));
@@ -311,6 +405,7 @@ number:
 		case TERROR:
 			markall_ret(-1)
 		}
+		threadflag = 0;
 		tok = scan(&bufp);
 	}
 	lastcolmod = colmod;
@@ -318,19 +413,35 @@ number:
 	np--;
 	mc = 0;
 	if (star) {
-		for (i = 0; i < msgcount; i++)
-			if ((message[i].m_flag & MDELETED) == f) {
-				mark(i+1);
+		for (i = 0; i < msgCount; i++) {
+			if (!(message[i].m_flag & MHIDDEN) &&
+					(message[i].m_flag & MDELETED) == f) {
+				mark(i+1, f);
 				mc++;
 			}
+		}
 		if (mc == 0) {
-			printf(catgets(catd, CATSET, 119,
+			if (!inhook)
+				printf(catgets(catd, CATSET, 119,
 					"No applicable messages.\n"));
 			markall_ret(-1)
 		}
 		markall_ret(0)
 	}
 
+	if ((topen || tback) && mc == 0) {
+		for (i = 0; i < msgCount; i++)
+			if (message[i].m_flag & MMARK)
+				mc++;
+		if (mc == 0) {
+			if (!inhook)
+				printf(tback ?
+					"No previously marked messages.\n" :
+					"No messages satify (criteria).\n");
+			markall_ret(-1)
+		}
+	}
+
 	/*
 	 * If no numbers were given, mark all of the messages,
 	 * so that we can unmark any whose sender was not selected
@@ -338,9 +449,11 @@ number:
 	 */
 
 	if ((np > namelist || colmod != 0 || id) && mc == 0)
-		for (i = 1; i <= msgcount; i++)
-			if ((message[i-1].m_flag & MDELETED) == f)
-				mark(i);
+		for (i = 1; i <= msgCount; i++) {
+			if (!(message[i-1].m_flag & MHIDDEN) &&
+					(message[i-1].m_flag & MDELETED) == f)
+				mark(i, f);
+		}
 
 	/*
 	 * If any names were given, go through and eliminate any
@@ -350,7 +463,9 @@ number:
 	if (np > namelist || id) {
 		int	allnet = value("allnet") != NULL;
 
-		for (i = 1; i <= msgcount; i++) {
+		if (mb.mb_type == MB_IMAP && gotheaders++ == 0)
+			imap_getheaders(1, msgCount);
+		for (i = 1; i <= msgCount; i++) {
 			mc = 0;
 			if (np > namelist) {
 				for (nq = &namelist[0]; *nq != NULL; nq++) {
@@ -369,7 +484,7 @@ number:
 					}
 				}
 			}
-			if (mc == 0 && id && matchmid(id, i))
+			if (mc == 0 && id && matchmid(id, idtype, i))
 				mc++;
 			if (mc == 0)
 				unmark(i);
@@ -380,13 +495,13 @@ number:
 		 */
 
 		mc = 0;
-		for (i = 1; i <= msgcount; i++)
+		for (i = 1; i <= msgCount; i++)
 			if (message[i-1].m_flag & MMARK) {
 				mc++;
 				break;
 			}
 		if (mc == 0) {
-			if (np > namelist) {
+			if (!inhook && np > namelist) {
 				printf(catgets(catd, CATSET, 120,
 					"No applicable messages from {%s"),
 					namelist[0]);
@@ -408,7 +523,7 @@ number:
 	 */
 
 	if (colmod != 0) {
-		for (i = 1; i <= msgcount; i++) {
+		for (i = 1; i <= msgCount; i++) {
 			struct coltab *colp;
 
 			mp = &message[i - 1];
@@ -419,18 +534,20 @@ number:
 						unmark(i);
 			
 		}
-		for (mp = &message[0]; mp < &message[msgcount]; mp++)
+		for (mp = &message[0]; mp < &message[msgCount]; mp++)
 			if (mp->m_flag & MMARK)
 				break;
-		if (mp >= &message[msgcount]) {
+		if (mp >= &message[msgCount]) {
 			struct coltab *colp;
 
-			printf(catgets(catd, CATSET, 123,
+			if (!inhook) {
+				printf(catgets(catd, CATSET, 123,
 						"No messages satisfy"));
-			for (colp = &coltab[0]; colp->co_char; colp++)
-				if (colp->co_bit & colmod)
-					printf(" :%c", colp->co_char);
-			printf("\n");
+				for (colp = &coltab[0]; colp->co_char; colp++)
+					if (colp->co_bit & colmod)
+						printf(" :%c", colp->co_char);
+				printf("\n");
+			}
 			markall_ret(-1)
 		}
 	}
@@ -444,9 +561,8 @@ out:
  * Turn the character after a colon modifier into a bit
  * value.
  */
-static int
-evalcol(col)
-	int col;
+static int 
+evalcol(int col)
 {
 	struct coltab *colp;
 
@@ -463,19 +579,19 @@ evalcol(col)
  * If f is MDELETED, then either kind will do.  Otherwise, the message
  * has to be undeleted.
  */
-static int
-check(mesg, f)
-	int mesg, f;
+static int 
+check(int mesg, int f)
 {
 	struct message *mp;
 
-	if (mesg < 1 || mesg > msgcount) {
+	if (mesg < 1 || mesg > msgCount) {
 		printf(catgets(catd, CATSET, 124,
 			"%d: Invalid message number\n"), mesg);
 		return(-1);
 	}
 	mp = &message[mesg-1];
-	if (f != MDELETED && (mp->m_flag & MDELETED) != 0) {
+	if (mp->m_flag & MHIDDEN || (f != MDELETED &&
+				(mp->m_flag & MDELETED) != 0)) {
 		printf(catgets(catd, CATSET, 125,
 			"%d: Inappropriate message\n"), mesg);
 		return(-1);
@@ -488,13 +604,11 @@ check(mesg, f)
  * for a RAWLIST.
  */
 int
-getrawlist(line, linesize, argv, argc)
-	char line[];
-	size_t linesize;
-	char **argv;
-	int  argc;
+getrawlist(const char *line, size_t linesize, char **argv, int argc,
+		int echolist)
 {
-	char c, *cp, *cp2, quotec;
+	char c, *cp2, quotec;
+	const char	*cp;
 	int argn;
 	char *linebuf;
 
@@ -515,14 +629,17 @@ getrawlist(line, linesize, argv, argc)
 		while ((c = *cp) != '\0') {
 			cp++;
 			if (quotec != '\0') {
-				if (c == quotec)
+				if (c == quotec) {
 					quotec = '\0';
-				else if (c == '\\')
+					if (echolist)
+						*cp2++ = c;
+				} else if (c == '\\')
 					switch (c = *cp++) {
 					case '\0':
 						*cp2++ = '\\';
 						cp--;
 						break;
+					/*
 					case '0': case '1': case '2': case '3':
 					case '4': case '5': case '6': case '7':
 						c -= '0';
@@ -550,14 +667,17 @@ getrawlist(line, linesize, argv, argc)
 					case 'v':
 						*cp2++ = '\v';
 						break;
+					*/
 					default:
+						if (cp[-1]!=quotec || echolist)
+							*cp2++ = '\\';
 						*cp2++ = c;
 					}
-				else if (c == '^') {
+				/*else if (c == '^') {
 					c = *cp++;
 					if (c == '?')
 						*cp2++ = '\177';
-					/* null doesn't show up anyway */
+					/\* null doesn't show up anyway *\/
 					else if ((c >= 'A' && c <= '_') ||
 						 (c >= 'a' && c <= 'z'))
 						*cp2++ = c & 037;
@@ -565,11 +685,18 @@ getrawlist(line, linesize, argv, argc)
 						*cp2++ = '^';
 						cp--;
 					}
-				} else
+				}*/ else
+					*cp2++ = c;
+			} else if (c == '"' || c == '\'') {
+				if (echolist)
 					*cp2++ = c;
-			} else if (c == '"' || c == '\'')
 				quotec = c;
-			else if (blankchar(c & 0377))
+			} else if (c == '\\' && !echolist) {
+				if (*cp)
+					*cp2++ = *cp++;
+				else
+					*cp2++ = c;
+			} else if (blankchar(c & 0377))
 				break;
 			else
 				*cp2++ = c;
@@ -603,15 +730,15 @@ static struct lex {
 	{ ')',	TCLOSE },
 	{ ',',	TCOMMA },
 	{ ';',	TSEMI },
+	{ '`',	TBACK },
 	{ 0,	0 }
 };
 
-static int
-scan(sp)
-	char **sp;
+static int 
+scan(char **sp)
 {
 	char *cp, *cp2;
-	int c;
+	int c, level, inquote;
 	struct lex *lp;
 	int quotec;
 
@@ -642,6 +769,20 @@ scan(sp)
 		return(TEOL);
 	}
 
+	/*
+	 * Select members of a message thread.
+	 */
+	if (c == '&') {
+		threadflag = 1;
+		if (*cp == '\0' || spacechar(*cp&0377)) {
+			lexstring[0] = '.';
+			lexstring[1] = '\0';
+			*sp = cp;
+			return TDOT;
+		}
+		c = *cp++;
+	}
+
 	/*
 	 * If the leading character is a digit, scan
 	 * the number and convert it on the fly.
@@ -660,6 +801,53 @@ scan(sp)
 		return(TNUMBER);
 	}
 
+	/*
+	 * An IMAP SEARCH list. Note that TOPEN has always been included
+	 * in singles[] in Mail and mailx. Thus although there is no formal
+	 * definition for (LIST) lists, they do not collide with historical
+	 * practice because a subject string (LIST) could never been matched
+	 * this way.
+	 */
+
+	if (c == '(') {
+		level = 1;
+		inquote = 0;
+		*cp2++ = c;
+		do {
+			if ((c = *cp++&0377) == '\0') {
+			mtop:	fprintf(stderr, "Missing \")\".\n");
+				return TERROR;
+			}
+			if (inquote && c == '\\') {
+				*cp2++ = c;
+				c = *cp++&0377;
+				if (c == '\0')
+					goto mtop;
+			} else if (c == '"')
+				inquote = !inquote;
+			else if (inquote)
+				/*EMPTY*/;
+			else if (c == '(')
+				level++;
+			else if (c == ')')
+				level--;
+			else if (spacechar(c)) {
+				/*
+				 * Replace unquoted whitespace by single
+				 * space characters, to make the string
+				 * IMAP SEARCH conformant.
+				 */
+				c = ' ';
+				if (cp2[-1] == ' ')
+					cp2--;
+			}
+			*cp2++ = c;
+		} while (c != ')' || level > 0);
+		*cp2 = '\0';
+		*sp = cp;
+		return TOPEN;
+	}
+
 	/*
 	 * Check for single character tokens; return such
 	 * if found.
@@ -687,6 +875,8 @@ scan(sp)
 		c = *cp++;
 	}
 	while (c != '\0') {
+		if (quotec == 0 && c == '\\' && *cp)
+			c = *cp++;
 		if (c == quotec) {
 			cp++;
 			break;
@@ -710,9 +900,8 @@ scan(sp)
 /*
  * Unscan the named token by pushing it onto the regret stack.
  */
-static void
-regret(token)
-	int token;
+static void 
+regret(int token)
 {
 	if (++regretp >= REGDEP)
 		panic(catgets(catd, CATSET, 128, "Too many regrets"));
@@ -725,32 +914,36 @@ regret(token)
 /*
  * Reset all the scanner global variables.
  */
-static void
-scaninit()
+static void 
+scaninit(void)
 {
 	regretp = -1;
+	threadflag = 0;
 }
 
 /*
  * Find the first message whose flags & m == f  and return
  * its message number.
  */
-int
-first(f, m)
-	int f, m;
+int 
+first(int f, int m)
 {
 	struct message *mp;
 
-	if (msgcount == 0)
+	if (msgCount == 0)
 		return 0;
 	f &= MDELETED;
 	m &= MDELETED;
-	for (mp = dot; mp < &message[msgcount]; mp++)
-		if ((mp->m_flag & m) == f)
+	for (mp = dot; mb.mb_threaded ? mp != NULL : mp < &message[msgCount];
+			mb.mb_threaded ? mp = next_in_thread(mp) : mp++) {
+		if (!(mp->m_flag & MHIDDEN) && (mp->m_flag & m) == f)
 			return mp - message + 1;
-	for (mp = dot-1; mp >= &message[0]; mp--)
-		if ((mp->m_flag & m) == f)
+	}
+	for (mp = dot-1; mb.mb_threaded ? mp != NULL : mp >= &message[0];
+			mb.mb_threaded ? mp = prev_in_thread(mp) : mp--) {
+		if (!(mp->m_flag & MHIDDEN) && (mp->m_flag & m) == f)
 			return mp - message + 1;
+	}
 	return 0;
 }
 
@@ -758,10 +951,8 @@ first(f, m)
  * See if the passed name sent the passed message number.  Return true
  * if so.
  */
-static int
-matchsender(str, mesg, allnet)
-	char *str;
-	int mesg, allnet;
+static int 
+matchsender(char *str, int mesg, int allnet)
 {
 	if (allnet) {
 		char *cp = nameof(&message[mesg - 1], 0);
@@ -779,16 +970,25 @@ matchsender(str, mesg, allnet)
 			(name1(&message[mesg - 1], 0)));
 }
 
-static int
-matchmid(id, mesg)
-	char *id;
-	int mesg;
+static int 
+matchmid(char *id, enum idtype idtype, int mesg)
 {
+	struct name	*np;
 	char *cp;
 
-	if ((cp = hfield("message-id", &message[mesg - 1])) != NULL &&
-			strcmp(cp, id) == 0)
-		return 1;
+	if ((cp = hfield("message-id", &message[mesg - 1])) != NULL) {
+		switch (idtype) {
+		case ID_REFERENCES:
+			return msgidcmp(id, cp) == 0;
+		case ID_IN_REPLY_TO:
+			if ((np = extract(id, GREF)) != NULL)
+				do {
+					if (msgidcmp(np->n_name, cp) == 0)
+						return 1;
+				} while ((np = np->n_flink) != NULL);
+			break;
+		}
+	}
 	return 0;
 }
 
@@ -802,14 +1002,13 @@ matchmid(id, mesg)
 
 static char lastscan[128];
 
-static int
-matchsubj(str, mesg)
-	char *str;
-	int mesg;
+static int 
+matchsubj(char *str, int mesg)
 {
 	struct message *mp;
-	char *cp, *cp2, *backup;
+	char *cp, *cp2;
 	struct str in, out;
+	int	i;
 
 	str++;
 	if (strlen(str) == 0) {
@@ -838,81 +1037,49 @@ matchsubj(str, mesg)
 	in.s = cp2;
 	in.l = strlen(cp2);
 	mime_fromhdr(&in, &out, TD_ICONV);
-	cp2 = out.s;
-	backup = cp2;
-	while (*cp2) {
-		if (*cp == 0) {
-			free(out.s);
-			return(1);
-		}
-#if defined (HAVE_MBTOWC) && defined (HAVE_TOWUPPER)
-		if (mb_cur_max > 1) {
-			wchar_t c, c2;
-			int sz;
-
-			if ((sz = mbtowc(&c, cp, mb_cur_max)) < 0)
-				goto singlebyte;
-			cp += sz;
-			if ((sz = mbtowc(&c2, cp2, mb_cur_max)) < 0)
-				goto singlebyte;
-			cp2 += sz;
-			c = towupper(c);
-			c2 = towupper(c2);
-			if (c != c2) {
-				if ((sz = mbtowc(&c, backup, mb_cur_max)) > 0) {
-					backup += sz;
-					cp2 = backup;
-				} else
-					cp2 = ++backup;
-				cp = str;
-			}
-		} else
-#endif	/* HAVE_MBTOWC && HAVE_TOWUPPER */
-		{
-			int c, c2;
-
-	singlebyte:	c = *cp++ & 0377;
-			if (islower(c))
-				c = toupper(c);
-			c2 = *cp2++ & 0377;
-			if (islower(c2))
-				c2 = toupper(c2);
-			if (c != c2) {
-				cp2 = ++backup;
-				cp = str;
-			}
-		}
-	}
+	i = substr(out.s, cp);
 	free(out.s);
-	return(*cp == 0);
+	return i;
 }
 
 /*
  * Mark the named message by setting its mark bit.
  */
-static void
-mark(mesg)
-	int mesg;
+void 
+mark(int mesg, int f)
 {
+	struct message	*mp;
 	int i;
 
 	i = mesg;
-	if (i < 1 || i > msgcount)
+	if (i < 1 || i > msgCount)
 		panic(catgets(catd, CATSET, 129, "Bad message number to mark"));
-	message[i-1].m_flag |= MMARK;
+	if (mb.mb_threaded == 1 && threadflag) {
+		if ((message[i-1].m_flag & MHIDDEN) == 0) {
+			if (f == MDELETED ||
+					(message[i-1].m_flag&MDELETED) == 0)
+			message[i-1].m_flag |= MMARK;
+		}
+		if (message[i-1].m_child) {
+			mp = message[i-1].m_child;
+			mark(mp-message+1, f);
+			for (mp = mp->m_younger; mp; mp = mp->m_younger)
+				mark(mp-message+1, f);
+		}
+	} else
+		message[i-1].m_flag |= MMARK;
 }
 
 /*
  * Unmark the named message.
  */
-static void
-unmark(mesg)
-	int mesg;
+static void 
+unmark(int mesg)
 {
 	int i;
 
 	i = mesg;
-	if (i < 1 || i > msgcount)
+	if (i < 1 || i > msgCount)
 		panic(catgets(catd, CATSET, 130,
 					"Bad message number to unmark"));
 	message[i-1].m_flag &= ~MMARK;
@@ -921,9 +1088,8 @@ unmark(mesg)
 /*
  * Return the message number corresponding to the passed meta character.
  */
-static int
-metamess(meta, f)
-	int meta, f;
+static int 
+metamess(int meta, int f)
 {
 	int c, m;
 	struct message *mp;
@@ -934,20 +1100,43 @@ metamess(meta, f)
 		/*
 		 * First 'good' message left.
 		 */
-		for (mp = &message[0]; mp < &message[msgcount]; mp++)
-			if ((mp->m_flag & MDELETED) == f)
+		mp = mb.mb_threaded ? threadroot : &message[0];
+		while (mp < &message[msgCount]) {
+			if (!(mp->m_flag & (MHIDDEN|MKILL)) &&
+					(mp->m_flag & MDELETED) == f)
 				return(mp - &message[0] + 1);
-		printf(catgets(catd, CATSET, 131, "No applicable messages\n"));
+			if (mb.mb_threaded) {
+				mp = next_in_thread(mp);
+				if (mp == NULL)
+					break;
+			} else
+				mp++;
+		}
+		if (!inhook)
+			printf(catgets(catd, CATSET, 131,
+						"No applicable messages\n"));
 		return(-1);
 
 	case '$':
 		/*
 		 * Last 'good message left.
 		 */
-		for (mp = &message[msgcount-1]; mp >= &message[0]; mp--)
-			if ((mp->m_flag & MDELETED) == f)
+		mp = mb.mb_threaded ? this_in_thread(threadroot, -1) :
+			&message[msgCount-1];
+		while (mp >= &message[0]) {
+			if (!(mp->m_flag & (MHIDDEN|MKILL)) &&
+					(mp->m_flag & MDELETED) == f)
 				return(mp - &message[0] + 1);
-		printf(catgets(catd, CATSET, 132, "No applicable messages\n"));
+			if (mb.mb_threaded) {
+				mp = prev_in_thread(mp);
+				if (mp == NULL)
+					break;
+			} else
+				mp--;
+		}
+		if (!inhook)
+			printf(catgets(catd, CATSET, 132,
+						"No applicable messages\n"));
 		return(-1);
 
 	case '.':
@@ -955,7 +1144,7 @@ metamess(meta, f)
 		 * Current message.
 		 */
 		m = dot - &message[0] + 1;
-		if ((dot->m_flag & MDELETED) != f) {
+		if (dot->m_flag & MHIDDEN || (dot->m_flag & MDELETED) != f) {
 			printf(catgets(catd, CATSET, 133,
 				"%d: Inappropriate message\n"), m);
 			return(-1);
@@ -972,7 +1161,7 @@ metamess(meta, f)
 			return(-1);
 		}
 		m = prevdot - &message[0] + 1;
-		if ((prevdot->m_flag & MDELETED) != f) {
+		if (prevdot->m_flag&MHIDDEN || (prevdot->m_flag&MDELETED)!=f) {
 			printf(catgets(catd, CATSET, 133,
 				"%d: Inappropriate message\n"), m);
 			return(-1);
