@@ -1,7 +1,7 @@
 /*
  * Nail - a mail user agent derived from Berkeley Mail.
  *
- * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ * Copyright (c) 2000-2004 Gunnar Ritter, Freiburg i. Br., Germany.
  */
 /*
  * Copyright (c) 2000
@@ -40,17 +40,17 @@
 #ifdef	DOSCCS
 static char copyright[]
 = "@(#) Copyright (c) 2000, 2002 Gunnar Ritter. All rights reserved.\n";
-static char sccsid[]  = "@(#)mime.c	2.12 (gritter) 11/24/02";
+static char sccsid[]  = "@(#)mime.c	2.51 (gritter) 12/25/04";
 #endif /* DOSCCS */
 #endif /* not lint */
 
 #include "rcv.h"
 #include "extern.h"
+#include <ctype.h>
 #include <errno.h>
-
-#ifdef	HAVE_STRINGS_H
-#include <strings.h>
-#endif
+#ifdef	HAVE_WCTYPE_H
+#include <wctype.h>
+#endif	/* HAVE_WCTYPE_H */
 
 /*
  * Mail -- a mail program
@@ -66,39 +66,42 @@ static char *mimetypes_world = "/etc/mime.types";
 static char *mimetypes_user = "~/.mime.types";
 char *us_ascii = "us-ascii";
 
-static int	mustquote_body __P((int));
-static int	mustquote_hdr __P((int));
-static int	mustquote_inhdrq __P((int));
-#if defined (HAVE_MBTOWC) && defined (HAVE_ISWPRINT)
-static size_t	xmbstowcs __P((wchar_t *, const char *, size_t));
-#endif
-static char	*getcharset __P((int));
-static int	has_highbit __P((const char *));
-static void	uppercopy __P((char *, const char *));
-static void	stripdash __P((char *));
+static int mustquote_body(int c);
+static int mustquote_hdr(const char *cp, int wordstart, int wordend);
+static int mustquote_inhdrq(int c);
+static size_t	delctrl(char *cp, size_t sz);
+static char *getcharset(int isclean);
+static int has_highbit(register const char *s);
 #ifdef	HAVE_ICONV
-static size_t	iconv_ft __P((iconv_t, char **, size_t *, char **, size_t *));
-#endif
-static int	is_this_enc __P((const char *, const char *));
-static char	*mime_tline __P((char *, char *));
-static char	*mime_type __P((char *, char *));
-static int	gettextconversion __P((void));
-static enum mimeclean	mime_isclean __P((FILE*));
-static void	invalid_seq __P((int));
-static char	*ctohex __P((int, char *));
-static size_t	mime_write_toqp __P((struct str *, FILE *, int (*)(int)));
-static void	mime_str_toqp __P((struct str *, struct str *, int (*)(int)));
-static void	mime_fromqp __P((struct str *, struct str *, int));
-static size_t	mime_write_tohdr __P((struct str *, FILE *));
-static size_t	mime_write_tohdr_a __P((struct str *, FILE *));
-static size_t	fwrite_td __P((void *, size_t, size_t, FILE *, enum tdflags,
-			char *, size_t));
+static void uppercopy(char *dest, const char *src);
+static void stripdash(char *p);
+static size_t iconv_ft(iconv_t cd, char **inb, size_t *inbleft,
+		char **outb, size_t *outbleft);
+static void invalid_seq(int c);
+#endif	/* HAVE_ICONV */
+static int is_this_enc(const char *line, const char *encoding);
+static char *mime_tline(char *x, char *l);
+static char *mime_type(char *ext, char *filename);
+static enum mimeclean mime_isclean(FILE *f);
+static enum conversion gettextconversion(void);
+static char *ctohex(int c, char *hex);
+static size_t mime_write_toqp(struct str *in, FILE *fo, int (*mustquote)(int));
+static void mime_str_toqp(struct str *in, struct str *out,
+		int (*mustquote)(int), int inhdr);
+static void mime_fromqp(struct str *in, struct str *out, int ishdr);
+static size_t mime_write_tohdr(struct str *in, FILE *fo);
+static size_t convhdra(char *str, size_t len, FILE *fp);
+static size_t mime_write_tohdr_a(struct str *in, FILE *f);
+static void addstr(char **buf, size_t *sz, size_t *pos, char *str, size_t len);
+static void addconv(char **buf, size_t *sz, size_t *pos, char *str, size_t len);
+static size_t fwrite_td(void *ptr, size_t size, size_t nmemb, FILE *f,
+		enum tdflags flags, char *prefix, size_t prefixlen);
 
 /*
  * Check if c must be quoted inside a message's body.
  */
-static int
-mustquote_body(c)
+static int 
+mustquote_body(int c)
 {
 	if (c != '\n' && (c < 040 || c == '=' || c >= 0177))
 		return 1;
@@ -108,121 +111,51 @@ mustquote_body(c)
 /*
  * Check if c must be quoted inside a message's header.
  */
-static int
-mustquote_hdr(c)
+static int 
+mustquote_hdr(const char *cp, int wordstart, int wordend)
 {
+	int	c = *cp & 0377;
+
 	if (c != '\n' && (c < 040 || c >= 0177))
 		return 1;
+	if (wordstart && cp[0] == '=' && cp[1] == '?')
+		return 1;
+	if (cp[0] == '?' && cp[1] == '=' &&
+			(wordend || cp[2] == '\0' || whitechar(cp[2]&0377)))
+		return 1;
 	return 0;
 }
 
 /*
  * Check if c must be quoted inside a quoting in a message's header.
  */
-static int
-mustquote_inhdrq(c)
+static int 
+mustquote_inhdrq(int c)
 {
 	if (c != '\n'
-		&& (c < 040 || c == '=' || c == '?' || c == '_' || c >= 0177))
+		&& (c <= 040 || c == '=' || c == '?' || c == '_' || c >= 0177))
 		return 1;
 	return 0;
 }
 
-#if defined (HAVE_MBTOWC) && defined (HAVE_ISWPRINT)
-/*
- * A mbstowcs()-alike function that transparently handles invalid sequences.
- */
 static size_t
-xmbstowcs(pwcs, s, nwcs)
-register wchar_t *pwcs;
-register const char *s;
-size_t nwcs;
+delctrl(char *cp, size_t sz)
 {
-	size_t n = nwcs;
-	register int c;
+	size_t	x = 0, y = 0;
 
-	mbtowc(pwcs, NULL, mb_cur_max);
-	while (*s && n) {
-		if ((c = mbtowc(pwcs, s, mb_cur_max)) < 0) {
-			s++;
-			*pwcs = L'?';
-		} else
-			s += c;
-		pwcs++;
-		n--;
+	while (x < sz) {
+		if (!cntrlchar(cp[x]&0377))
+			cp[y++] = cp[x];
+		x++;
 	}
-	if (n)
-		*pwcs = L'\0';
-	mbtowc(pwcs, NULL, mb_cur_max);
-	return nwcs - n;
-}
-#endif	/* HAVE_MBTOWC && HAVE_ISWPRINT */
-
-/*
- * Replace non-printable characters in s with question marks.
- */
-size_t
-makeprint(s, l)
-char *s;
-size_t l;
-{
-#ifdef	HAVE_SETLOCALE
-	size_t sz;
-#if defined (HAVE_MBTOWC) && defined (HAVE_ISWPRINT)
-	int i;
-	wchar_t *w, *p;
-	char *t;
-	size_t ret;
-
-#ifdef	__GLIBC_MINOR__
-#if __GLIBC__ <= 2 && __GLIBC_MINOR__ <= 1
-	/*
-	 * This libc does not have correct utf-8 locales.
-	 * Just a strange hack that will disappear one
-	 * day, but we need it for testing.
-	 */
-	if (asccasecmp(gettcharset(), "utf-8") == 0)
-		return l;
-#endif
-#endif
-	if (*s == '\0' || l == (size_t) 0)
-		return 0;
-	if (mb_cur_max > 1) {
-		w = ac_alloc((l + 1) * sizeof *w);
-		t = s;
-		if ((sz = xmbstowcs(w, t, l + 1)) == (size_t)-1) {
-			ac_free(w);
-			return 0;
-		}
-		for (p = w, i = 0; *p && i < sz; p++, i++) {
-			if (!iswprint(*p) && *p != '\n' && *p != '\r'
-					&& *p != '\b' && *p != '\t')
-				*p = '?';
-		}
-		p = w;
-		ret = wcstombs(s, p, l + 1);
-		ac_free(w);
-		return ret;
-	}
-#endif	/* HAVE_MBTOWC && HAVE_ISWPRINT */
-	sz = l;
-	for (; l > 0 && *s; s++, l--) {
-		if (!isprint(*s & 0377) && *s != '\n' && *s != '\r'
-				&& *s != '\b' && *s != '\t')
-			*s = '?';
-	}
-	return sz;
-#else	/* !HAVE_SETLOCALE */
-	return l;
-#endif	/* !HAVE_SETLOCALE */
+	return y;
 }
 
 /*
  * Check if a name's address part contains invalid characters.
  */
-int
-mime_name_invalid(name, putmsg)
-char *name;
+int 
+mime_name_invalid(char *name, int putmsg)
 {
 	char *addr, *p;
 	int in_quote = 0, in_domain = 0, err = 0, hadat = 0;
@@ -279,12 +212,12 @@ char *name;
 				"%s contains invalid character '"), addr);
 #ifdef	HAVE_SETLOCALE
 		if (isprint(err))
-#endif	/* HAVE_SETLOCALE */
-			sputc(err, stderr);
-#ifdef	HAVE_SETLOCALE
+#else	/* !HAVE_SETLOCALE */
+		if (err >= 040 && err <= 0177)
+#endif	/* !HAVE_SETLOCALE */
+			putc(err, stderr);
 		else
 			fprintf(stderr, "\\%03o", err);
-#endif	/* HAVE_SETLOCALE */
 		fprintf(stderr, catgets(catd, CATSET, 144, "'\n"));
 	}
 	return err;
@@ -294,8 +227,7 @@ char *name;
  * Check all addresses in np and delete invalid ones.
  */
 struct name *
-checkaddrs(np)
-struct name *np;
+checkaddrs(struct name *np)
 {
 	struct name *n = np;
 
@@ -319,7 +251,7 @@ static char defcharset[] = "iso-8859-1";
  * Get the character set dependant on the conversion.
  */
 static char *
-getcharset(isclean)
+getcharset(int isclean)
 {
 	char *charset;
 
@@ -345,7 +277,7 @@ getcharset(isclean)
  * Get the setting of the terminal's character set.
  */
 char *
-gettcharset()
+gettcharset(void)
 {
 	char *t;
 
@@ -355,9 +287,8 @@ gettcharset()
 	return t;
 }
 
-static int
-has_highbit(s)
-	register const char *s;
+static int 
+has_highbit(const char *s)
 {
 	if (s) {
 		do
@@ -369,9 +300,7 @@ has_highbit(s)
 }
 
 char *
-need_hdrconv(hp, w)
-	struct header *hp;
-	enum gfield w;
+need_hdrconv(struct header *hp, enum gfield w)
 {
 	struct name *np;
 
@@ -398,6 +327,11 @@ need_hdrconv(hp, w)
 			if (has_highbit(np->n_name))
 				goto needs;
 	}
+	if (w & GREPLYTO) {
+		for (np = hp->h_replyto; np; np = np->n_flink)
+			if (has_highbit(np->n_name))
+				goto needs;
+	}
 	if (w & GSUBJECT) {
 		if (has_highbit(hp->h_subject))
 			goto needs;
@@ -410,10 +344,8 @@ needs:	return getcharset(MIME_HIGHBIT);
 /*
  * Convert a string, upper-casing the characters.
  */
-static void
-uppercopy(dest, src)
-char *dest;
-const char *src;
+static void 
+uppercopy(char *dest, const char *src)
 {
 	do
 		*dest++ = upperconv(*src & 0377);
@@ -423,9 +355,8 @@ const char *src;
 /*
  * Strip dashes.
  */
-static void
-stripdash(p)
-char *p;
+static void 
+stripdash(char *p)
 {
 	char *q = p;
 
@@ -440,8 +371,7 @@ char *p;
  * naming conventions.
  */
 iconv_t
-iconv_open_ft(tocode, fromcode)
-const char *tocode, *fromcode;
+iconv_open_ft(const char *tocode, const char *fromcode)
 {
 	iconv_t id;
 	char *t, *f;
@@ -477,9 +407,9 @@ const char *tocode, *fromcode;
 	/*
 	 * Solaris prefers upper-case charset names. Don't ask...
 	 */
-	t = (char *)salloc(strlen(tocode) + 1);
+	t = salloc(strlen(tocode) + 1);
 	uppercopy(t, tocode);
-	f = (char *)salloc(strlen(fromcode) + 1);
+	f = salloc(strlen(fromcode) + 1);
 	uppercopy(f, fromcode);
 	if (strcmp(t, f) == 0) {
 		errno = 0;
@@ -508,11 +438,7 @@ const char *tocode, *fromcode;
  * Fault-tolerant iconv() function.
  */
 static size_t
-iconv_ft(cd, inb, inbleft, outb, outbleft)
-iconv_t cd;
-char **inb;
-size_t *inbleft, *outbleft;
-char **outb;
+iconv_ft(iconv_t cd, char **inb, size_t *inbleft, char **outb, size_t *outbleft)
 {
 	size_t sz = 0;
 
@@ -535,6 +461,16 @@ char **outb;
 	}
 	return sz;
 }
+
+/*
+ * Print an error because of an invalid character sequence.
+ */
+/*ARGSUSED*/
+static void 
+invalid_seq(int c)
+{
+	/*fprintf(stderr, "iconv: cannot convert %c\n", c);*/
+}
 #endif	/* HAVE_ICONV */
 
 static int
@@ -555,19 +491,11 @@ is_this_enc(const char *line, const char *encoding)
 }
 
 /*
- * Get the mime encoding from a Content-Transfer-Encoding header line.
+ * Get the mime encoding from a Content-Transfer-Encoding header field.
  */
-enum mimeenc
-mime_getenc(h)
-char *h;
+enum mimeenc 
+mime_getenc(char *p)
 {
-	char *p;
-
-	if ((p = strchr(h, ':')) == NULL)
-		return MIME_NONE;
-	p++;
-	while (*p && whitechar(*p & 0377))
-		p++;
 	if (is_this_enc(p, "7bit"))
 		return MIME_7B;
 	if (is_this_enc(p, "8bit"))
@@ -582,38 +510,33 @@ char *h;
 }
 
 /*
- * Get the mime content from a Content-Type header line.
+ * Get the mime content from a Content-Type header field, other parameters
+ * already stripped.
  */
-int
-mime_getcontent(h)
-char *h;
+int 
+mime_getcontent(char *s)
 {
-	char *p, *q, *r, *s;
-
-	if ((s = strchr(h, ':')) == NULL)
-		return 1;
-	s++;
-	while (*s && whitechar(*s & 0377))
-		s++;
-	if (*s == '"')
-		s++;
-	r = salloc(strlen(s) + 1);
-	p = s, q = r;
-	while (*p && !whitechar(*p & 0377) && *p != ';') {
-		*q++ = lowerconv(*p & 0377);
-		p++;
-	}
-	*q = '\0';
-	if (strchr(r, '/') == NULL)	/* for compatibility with non-MIME */
+	if (strchr(s, '/') == NULL)	/* for compatibility with non-MIME */
 		return MIME_TEXT;
-	if (strncmp(r, "text/", 5) == 0)
+	if (asccasecmp(s, "text/plain") == 0)
+		return MIME_TEXT_PLAIN;
+	if (asccasecmp(s, "text/html") == 0)
+		return MIME_TEXT_HTML;
+	if (ascncasecmp(s, "text/", 5) == 0)
 		return MIME_TEXT;
-	if (strncmp(r, "message/rfc822", 14) == 0)
+	if (asccasecmp(s, "message/rfc822") == 0)
 		return MIME_822;
-	if (strncmp(r, "message/", 8) == 0)
+	if (ascncasecmp(s, "message/", 8) == 0)
 		return MIME_MESSAGE;
-	if (strncmp(r, "multipart/", 10) == 0)
+	if (asccasecmp(s, "multipart/alternative") == 0)
+		return MIME_ALTERNATIVE;
+	if (asccasecmp(s, "multipart/digest") == 0)
+		return MIME_DIGEST;
+	if (ascncasecmp(s, "multipart/", 10) == 0)
 		return MIME_MULTI;
+	if (asccasecmp(s, "application/x-pkcs7-mime") == 0 ||
+				asccasecmp(s, "application/pkcs7-mime") == 0)
+		return MIME_PKCS7;
 	return MIME_UNKNOWN;
 }
 
@@ -621,8 +544,7 @@ char *h;
  * Get a mime style parameter from a header line.
  */
 char *
-mime_getparam(param, h)
-char *param, *h;
+mime_getparam(char *param, char *h)
 {
 	char *p = h, *q, *r;
 	int c;
@@ -679,7 +601,7 @@ char *param, *h;
 			q++;
 	}
 	sz = q - p;
-	r = (char*)salloc(q - p + 1);
+	r = salloc(q - p + 1);
 	memcpy(r, p, sz);
 	*(r + sz) = '\0';
 	return r;
@@ -689,8 +611,7 @@ char *param, *h;
  * Get the boundary out of a Content-Type: multipart/xyz header field.
  */
 char *
-mime_getboundary(h)
-char *h;
+mime_getboundary(char *h)
 {
 	char *p, *q;
 	size_t sz;
@@ -698,7 +619,7 @@ char *h;
 	if ((p = mime_getparam("boundary", h)) == NULL)
 		return NULL;
 	sz = strlen(p);
-	q = (char*)smalloc(sz + 3);
+	q = salloc(sz + 3);
 	memcpy(q, "--", 2);
 	memcpy(q + 2, p, sz);
 	*(q + sz + 2) = '\0';
@@ -709,8 +630,7 @@ char *h;
  * Get a line like "text/html html" and look if x matches the extension.
  */
 static char *
-mime_tline(x, l)
-char *x, *l;
+mime_tline(char *x, char *l)
 {
 	char *type, *n;
 	int match = 0;
@@ -741,7 +661,7 @@ char *x, *l;
 			l++;
 	}
 	if (match != 0) {
-		n = (char *)salloc(strlen(type) + 1);
+		n = salloc(strlen(type) + 1);
 		strcpy(n, type);
 		return n;
 	}
@@ -752,8 +672,7 @@ char *x, *l;
  * Check the given MIME type file for extension ext.
  */
 static char *
-mime_type(ext, filename)
-char *ext, *filename;
+mime_type(char *ext, char *filename)
 {
 	FILE *f;
 	char *line = NULL;
@@ -776,8 +695,7 @@ char *ext, *filename;
  * Return the Content-Type matching the extension of name.
  */
 char *
-mime_filecontent(name)
-char *name;
+mime_filecontent(char *name)
 {
 	char *ext, *content;
 
@@ -794,8 +712,7 @@ char *name;
  * Check file contents.
  */
 static enum mimeclean
-mime_isclean(f)
-FILE *f;
+mime_isclean(FILE *f)
 {
 	long initial_pos;
 	unsigned curlen = 1, maxlen = 0;
@@ -804,7 +721,7 @@ FILE *f;
 
 	initial_pos = ftell(f);
 	do {
-		c = sgetc(f);
+		c = getc(f);
 		curlen++;
 		if (c == '\n' || c == EOF) {
 			/*
@@ -834,8 +751,8 @@ FILE *f;
 /*
  * Get the conversion that matches the encoding specified in the environment.
  */
-static int
-gettextconversion()
+static enum conversion
+gettextconversion(void)
 {
 	char *p;
 	int convert;
@@ -855,11 +772,8 @@ gettextconversion()
 }
 
 int
-get_mime_convert(fp, contenttype, charset, isclean)
-FILE *fp;
-char **contenttype;
-char **charset;
-enum mimeclean *isclean;
+get_mime_convert(FILE *fp, char **contenttype, char **charset,
+		enum mimeclean *isclean, int dosign)
 {
 	int convert;
 
@@ -870,7 +784,7 @@ enum mimeclean *isclean;
 				ascncasecmp(*contenttype, "text/", 5) == 0)
 			*contenttype = "application/octet-stream";
 		*charset = NULL;
-	} else if (*isclean & MIME_LONGLINES)
+	} else if (*isclean & (MIME_LONGLINES|MIME_CTRLCHAR) || dosign)
 		convert = CONV_TOQP;
 	else if (*isclean & MIME_HIGHBIT)
 		convert = gettextconversion();
@@ -893,50 +807,11 @@ enum mimeclean *isclean;
 	return convert;
 }
 
-/*
- * Print an error because of an invalid character sequence.
- */
-/*ARGSUSED*/
-static void
-invalid_seq(c)
-char c;
-{
-	/*fprintf(stderr, "iconv: cannot convert %c\n", c);*/
-}
-
-/*
- * Convert i to a baseX character string and store it in b.
- * The caller has to ensure that the size of b is sufficient.
- */
-char *
-itostr(base, i, b)
-unsigned base, i;
-char *b;
-{
-	char *p, *q, c;
-	
-	p = b;
-	while (i != 0) {
-		*p++ = basetable[i % base];
-		i /= base;
-	}
-	*p-- = '\0';
-	q = b;
-	while (p >= q) {
-		c = *q;
-		*q++ = *p;
-		*p-- = c;
-	}
-	return b;
-}
-
 /*
  * Convert c to a hexadecimal character string and store it in hex.
  */
 static char *
-ctohex(c, hex)
-unsigned char c;
-char *hex;
+ctohex(int c, char *hex)
 {
 	unsigned char d;
 
@@ -955,10 +830,7 @@ char *hex;
  * The mustquote function determines whether a character must be quoted.
  */
 static size_t
-mime_write_toqp(in, fo, mustquote)
-struct str *in;
-FILE *fo;
-int (*mustquote)(int);
+mime_write_toqp(struct str *in, FILE *fo, int (*mustquote)(int))
 {
 	char *p, *upper, *h, hex[3];
 	int l;
@@ -967,16 +839,22 @@ int (*mustquote)(int);
 	sz = in->l;
 	upper = in->s + in->l;
 	for (p = in->s, l = 0; p < upper; p++) {
-		if (mustquote(*p) 
-				|| (*(p + 1) == '\n' && blankchar(*p & 0377))) {
+		if (mustquote(*p&0377) ||
+				p < upper-1 && p[1] == '\n' &&
+					blankchar(p[0]&0377) ||
+				p < upper-4 && l == 0 &&
+					p[0] == 'F' && p[1] == 'r' &&
+					p[2] == 'o' && p[3] == 'm' ||
+				*p == '.' && l == 0 && p < upper-1 &&
+					p[1] == '\n') {
 			if (l >= 69) {
 				sz += 2;
 				fwrite("=\n", sizeof (char), 2, fo);
 				l = 0;
 			}
 			sz += 2;
-			sputc('=', fo);
-			h = ctohex(*p, hex);
+			putc('=', fo);
+			h = ctohex(*p&0377, hex);
 			fwrite(h, sizeof *h, 2, fo);
 			l += 3;
 		} else {
@@ -987,7 +865,7 @@ int (*mustquote)(int);
 				fwrite("=\n", sizeof (char), 2, fo);
 				l = 0;
 			}
-			sputc(*p, fo);
+			putc(*p, fo);
 			l++;
 		}
 	}
@@ -998,24 +876,26 @@ int (*mustquote)(int);
  * Write to a stringstruct converting to quoted-printable.
  * The mustquote function determines whether a character must be quoted.
  */
-static void
-mime_str_toqp(in, out, mustquote)
-struct str *in, *out;
-int (*mustquote)(int);
+static void 
+mime_str_toqp(struct str *in, struct str *out, int (*mustquote)(int), int inhdr)
 {
 	char *p, *q, *upper;
 
-	out->s = (char*)smalloc(in->l * 3 + 1);
+	out->s = smalloc(in->l * 3 + 1);
 	q = out->s;
 	out->l = in->l;
 	upper = in->s + in->l;
 	for (p = in->s; p < upper; p++) {
-		if (mustquote(*p) ||
-				(*(p + 1) == '\n' && blankchar(*p & 0377))) {
-			out->l += 2;
-			*q++ = '=';
-			ctohex(*p, q);
-			q += 2;
+		if (mustquote(*p&0377) || p+1 < upper && *(p + 1) == '\n' &&
+				blankchar(*p & 0377)) {
+			if (inhdr && *p == ' ') {
+				*q++ = '_';
+			} else {
+				out->l += 2;
+				*q++ = '=';
+				ctohex(*p&0377, q);
+				q += 2;
+			}
 		} else {
 			*q++ = *p;
 		}
@@ -1026,9 +906,8 @@ int (*mustquote)(int);
 /*
  * Write to a stringstruct converting from quoted-printable.
  */
-static void
-mime_fromqp(in, out, ishdr)
-struct str *in, *out;
+static void 
+mime_fromqp(struct str *in, struct str *out, int ishdr)
 {
 	char *p, *q, *upper;
 	char quote[4];
@@ -1072,15 +951,13 @@ struct str *in, *out;
 /*
  * Convert header fields from RFC 1522 format
  */
-void
-mime_fromhdr(in, out, flags)
-struct str *in, *out;
-enum tdflags flags;
+void 
+mime_fromhdr(struct str *in, struct str *out, enum tdflags flags)
 {
-	char *p, *q, *op, *upper, *cs, *cbeg, *tcs;
+	char *p, *q, *op, *upper, *cs, *cbeg, *tcs, *lastwordend = NULL;
 	struct str cin, cout;
 	int convert;
-	size_t maxstor;
+	size_t maxstor, lastoutl = 0;
 #ifdef	HAVE_ICONV
 	iconv_t fhicd = (iconv_t)-1;
 #endif
@@ -1099,7 +976,7 @@ enum tdflags flags;
 				p++;	/* strip charset */
 			if (p >= upper)
 				goto notmime;
-			cs = (char *)salloc(++p - cbeg);
+			cs = salloc(++p - cbeg);
 			memcpy(cs, cbeg, p - cbeg - 1);
 			cs[p - cbeg - 1] = '\0';
 #ifdef	HAVE_ICONV
@@ -1140,6 +1017,10 @@ enum tdflags flags;
 					mime_fromqp(&cin, &cout, 1);
 					break;
 			}
+			if (lastwordend) {
+				q = lastwordend;
+				out->l = lastoutl;
+			}
 #ifdef	HAVE_ICONV
 			if ((flags & TD_ICONV) && fhicd != (iconv_t)-1) {
 				char *iptr, *mptr, *nptr, *uptr;
@@ -1170,6 +1051,8 @@ enum tdflags flags;
 			}
 #endif
 			free(cout.s);
+			lastwordend = q;
+			lastoutl = out->l;
 		} else {
 notmime:
 			p = op;
@@ -1177,12 +1060,20 @@ notmime:
 				mime_fromhdr_inc(16);
 			*q++ = *p;
 			out->l++;
+			if (!blankchar(*p&0377))
+				lastwordend = NULL;
 		}
 	}
 fromhdr_end:
 	*q = '\0';
-	if (flags & TD_ISPR)
-		out->l = makeprint(out->s, out->l);
+	if (flags & TD_ISPR) {
+		struct str	new;
+		makeprint(out, &new);
+		free(out->s);
+		*out = new;
+	}
+	if (flags & TD_DELCTRL)
+		out->l = delctrl(out->s, out->l);
 #ifdef	HAVE_ICONV
 	if (fhicd != (iconv_t)-1)
 		iconv_close(fhicd);
@@ -1194,23 +1085,29 @@ fromhdr_end:
  * Convert header fields to RFC 1522 format and write to the file fo.
  */
 static size_t
-mime_write_tohdr(in, fo)
-struct str *in;
-FILE *fo;
+mime_write_tohdr(struct str *in, FILE *fo)
 {
-	char *upper, *wbeg, *wend, *charset;
+	char *upper, *wbeg, *wend, *charset, *lastwordend = NULL, *lastspc, b,
+		*charset7;
 	struct str cin, cout;
-	size_t sz = 0, col = 0, wr, charsetlen;
-	int mustquote,
+	size_t sz = 0, col = 0, wr, charsetlen, charset7len;
+	int quoteany, mustquote, broken,
 		maxcol = 65 /* there is the header field's name, too */;
 
 	upper = in->s + in->l;
 	charset = getcharset(MIME_HIGHBIT);
+	if ((charset7 = value("charset7")) == NULL)
+		charset7 = us_ascii;
 	charsetlen = strlen(charset);
-	for (wbeg = in->s, mustquote = 0; wbeg < upper; wbeg++)
-		if (mustquote_hdr(*wbeg))
-			mustquote++;
-	if (2 * mustquote > in->l) {
+	charset7len = strlen(charset7);
+	charsetlen = smax(charsetlen, charset7len);
+	b = 0;
+	for (wbeg = in->s, quoteany = 0; wbeg < upper; wbeg++) {
+		b |= *wbeg;
+		if (mustquote_hdr(wbeg, wbeg == in->s, wbeg == &upper[-1]))
+			quoteany++;
+	}
+	if (2 * quoteany > in->l) {
 		/*
 		 * Print the entire field in base64.
 		 */
@@ -1221,7 +1118,8 @@ FILE *fo;
 				cin.l = wend - wbeg;
 				if (cin.l * 4/3 + 7 + charsetlen
 						< maxcol - col) {
-					fprintf(fo, "=?%s?B?", charset);
+					fprintf(fo, "=?%s?B?",
+						b&0200 ? charset : charset7);
 					wr = mime_write_tob64(&cin, fo, 1);
 					fwrite("?=", sizeof (char), 2, fo);
 					wr += 7 + charsetlen;
@@ -1249,59 +1147,103 @@ FILE *fo;
 		/*
 		 * Print the field word-wise in quoted-printable.
 		 */
+		broken = 0;
 		for (wbeg = in->s; wbeg < upper; wbeg = wend) {
+			lastspc = NULL;
 			while (wbeg < upper && whitechar(*wbeg & 0377)) {
-				sputc(*wbeg++, fo);
-				sz++, col++;
+				lastspc = lastspc ? lastspc : wbeg;
+				wbeg++;
+				col++;
+				broken = 0;
 			}
-			if (wbeg == upper)
+			if (wbeg == upper) {
+				if (lastspc)
+					while (lastspc < wbeg) {
+						putc(*lastspc&0377, fo);
+							lastspc++,
+							sz++;
+						}
 				break;
+			}
 			mustquote = 0;
+			b = 0;
 			for (wend = wbeg;
 				wend < upper && !whitechar(*wend & 0377);
 					wend++) {
-				if (mustquote_hdr(*wend))
+				b |= *wend;
+				if (mustquote_hdr(wend, wend == wbeg,
+							wbeg == &upper[-1]))
 					mustquote++;
 			}
-			if (mustquote) {
+			if (mustquote || broken || (wend - wbeg) >= 74 &&
+					quoteany) {
 				for (;;) {
-					cin.s = wbeg;
-					cin.l = wend - wbeg;
+					cin.s = lastwordend ? lastwordend :
+						wbeg;
+					cin.l = wend - cin.s;
 					mime_str_toqp(&cin, &cout,
-							mustquote_inhdrq);
+							mustquote_inhdrq, 1);
 					if ((wr = cout.l + charsetlen + 7)
 							< maxcol - col) {
-						fprintf(fo, "=?%s?Q?",
-								charset);
+						if (lastspc)
+							while (lastspc < wbeg) {
+								putc(*lastspc
+									&0377,
+									fo);
+								lastspc++,
+								sz++;
+							}
+						fprintf(fo, "=?%s?Q?", b&0200 ?
+							charset : charset7);
 						fwrite(cout.s, sizeof *cout.s,
 								cout.l, fo);
-						fwrite("?=", sizeof (char),
-								2, fo);
+						fwrite("?=", 1, 2, fo);
 						sz += wr, col += wr;
 						free(cout.s);
 						break;
 					} else {
+						broken = 1;
 						if (col) {
-							fprintf(fo, "\n ");
-							sz += 2;
+							putc('\n', fo);
+							sz++;
 							col = 0;
 							maxcol = 76;
+							if (lastspc == NULL) {
+								putc(' ', fo);
+								sz++;
+								maxcol--;
+							} else
+								maxcol -= wbeg -
+									lastspc;
 						} else {
 							wend -= 4;
 						}
 						free(cout.s);
 					}
 				}
+				lastwordend = wend;
 			} else {
 				if (col && wend - wbeg > maxcol - col) {
-					fwrite("\n ", sizeof (char), 2, fo);
-					sz += 2;
+					putc('\n', fo);
+					sz++;
 					col = 0;
 					maxcol = 76;
+					if (lastspc == NULL) {
+						putc(' ', fo);
+						sz++;
+						maxcol--;
+					} else
+						maxcol -= wbeg - lastspc;
 				}
+				if (lastspc)
+					while (lastspc < wbeg) {
+						putc(*lastspc&0377, fo);
+						lastspc++, sz++;
+					}
 				wr = fwrite(wbeg, sizeof *wbeg,
 						wend - wbeg, fo);
 				sz += wr, col += wr;
+				lastwordend = NULL;
 			}
 		}
 	}
@@ -1309,131 +1251,164 @@ FILE *fo;
 }
 
 /*
- * Write an address to a header field.
+ * Write len characters of the passed string to the passed file, 
+ * doing charset and header conversion.
  */
 static size_t
-mime_write_tohdr_a(in, f)
-struct str *in;
-FILE *f;
+convhdra(char *str, size_t len, FILE *fp)
 {
-	char *p, *q, *ip, *op;
-	struct str cin;
-	size_t sz = 0, isz, osz, cbufsz;
-	char *cbuf;
+#ifdef	HAVE_ICONV
+	char	*ip, *op;
+	size_t	isz, osz;
+#endif
+	struct str	cin;
+	size_t	cbufsz;
+	char	*cbuf;
+	size_t	sz;
 
 	cbuf = ac_alloc(cbufsz = 1);
-	*(in->s + in->l) = '\0';
-	if (strpbrk(in->s, "(<") == NULL) {
-		ac_free(cbuf);
-		return fwrite(in->s, sizeof *in->s, in->l, f);
-	}
-	if ((p = strchr(in->s, '<')) != NULL) {
-		q = strchr(p, '>');
-		if (q++ == NULL) {
-			ac_free(cbuf);
-			return 0;
-		}
 #ifdef	HAVE_ICONV
-		if (iconvd == (iconv_t)-1) {
+	if (iconvd == (iconv_t)-1) {
 #endif
-			cin.s = in->s;
-			cin.l = p - in->s;
+		cin.s = str;
+		cin.l = len;
 #ifdef	HAVE_ICONV
-		} else {
-		again1:	ip = in->s;
-			isz = p - in->s;
-			op = cbuf;
-			osz = cbufsz;
-			if (iconv(iconvd, &ip, &isz, &op, &osz) != 0) {
-				if (errno != E2BIG) {
-					ac_free(cbuf);
-					return 0;
-				}
-				cbuf = ac_alloc(cbufsz += isz);
-				goto again1;
-			}
-			cin.s = cbuf;
-			cin.l = cbufsz - osz;
-		}
-#endif
-		sz = mime_write_tohdr(&cin, f);
-		sz += fwrite(p, sizeof *p, q - p, f);
-#ifdef	HAVE_ICONV
-		if (iconvd == (iconv_t)-1) {
-#endif
-			cin.s = q;
-			cin.l = in->s + in->l - q;
-#ifdef	HAVE_ICONV
-		} else {
-		again2:	ip = q;
-			isz = in->s + in->l - q;
-			op = cbuf;
-			osz = cbufsz;
-			if (iconv(iconvd, &ip, &isz, &op, &osz) != 0) {
-				if (errno != E2BIG) {
-					ac_free(cbuf);
-					return 0;
-				}
-				cbuf = ac_alloc(cbufsz += isz);
-				goto again2;
-			}
-			cin.s = cbuf;
-			cin.l = cbufsz - osz;
-		}
-#endif
-		sz += mime_write_tohdr(&cin, f);
-	} else if ((p = strchr(in->s, '(')) != NULL) {
-		p++;
-		q = in->s;
-		do {
-			sz += fwrite(q, sizeof *q, p - q, f);
-			q = strchr(p, ')');
-			if (q++ == NULL) {
+	} else {
+	again:	ip = str;
+		isz = len;
+		op = cbuf;
+		osz = cbufsz;
+		if (iconv(iconvd, &ip, &isz, &op, &osz) != 0) {
+			if (errno != E2BIG) {
 				ac_free(cbuf);
 				return 0;
 			}
-#ifdef	HAVE_ICONV
-			if (iconvd == (iconv_t)-1) {
-#endif
-				cin.s = p;
-				cin.l = q - p;
-#ifdef	HAVE_ICONV
-			} else {
-			again3:	ip = p;
-				isz = q - p;
-				op = cbuf;
-				osz = cbufsz;
-				if (iconv(iconvd, &ip, &isz, &op, &osz) != 0) {
-					if (errno != E2BIG) {
-						ac_free(cbuf);
-						return 0;
-					}
-					cbuf = ac_alloc(cbufsz += isz);
-					goto again3;
-				}
-				cin.s = cbuf;
-				cin.l = cbufsz - osz;
-			}
-#endif
-			sz += mime_write_tohdr(&cin, f);
-			p = strchr(q, '(');
-		} while (p != NULL);
-		if (*q != '\0')
-			sz += fwrite(q, sizeof *q, in->s + in->l - q, f);
+			cbuf = ac_alloc(cbufsz += isz);
+			goto again;
+		}
+		cin.s = cbuf;
+		cin.l = cbufsz - osz;
 	}
+#endif	/* HAVE_ICONV */
+	sz = mime_write_tohdr(&cin, fp);
 	ac_free(cbuf);
 	return sz;
 }
 
+
+/*
+ * Write an address to a header field.
+ */
+static size_t
+mime_write_tohdr_a(struct str *in, FILE *f)
+{
+	char	*cp, *lastcp;
+	size_t	sz = 0;
+
+	in->s[in->l] = '\0';
+	lastcp = in->s;
+	if ((cp = routeaddr(in->s)) != NULL && cp > lastcp) {
+		sz += convhdra(lastcp, cp - lastcp, f);
+		lastcp = cp;
+	} else
+		cp = in->s;
+	for ( ; *cp; cp++) {
+		switch (*cp) {
+		case '(':
+			sz += fwrite(lastcp, 1, cp - lastcp + 1, f);
+			lastcp = ++cp;
+			cp = skip_comment(cp);
+			if (--cp > lastcp)
+				sz += convhdra(lastcp, cp - lastcp, f);
+			lastcp = cp;
+			break;
+		case '"':
+			while (*cp) {
+				if (*++cp == '"')
+					break;
+				if (*cp == '\\' && cp[1])
+					cp++;
+			}
+			break;
+		}
+	}
+	if (cp > lastcp)
+		sz += fwrite(lastcp, 1, cp - lastcp, f);
+	return sz;
+}
+
+static void
+addstr(char **buf, size_t *sz, size_t *pos, char *str, size_t len)
+{
+	*buf = srealloc(*buf, *sz += len);
+	memcpy(&(*buf)[*pos], str, len);
+	*pos += len;
+}
+
+static void
+addconv(char **buf, size_t *sz, size_t *pos, char *str, size_t len)
+{
+	struct str	in, out;
+
+	in.s = str;
+	in.l = len;
+	mime_fromhdr(&in, &out, TD_ISPR|TD_ICONV);
+	addstr(buf, sz, pos, out.s, out.l);
+	free(out.s);
+}
+
+/*
+ * Interpret MIME strings in parts of an address field.
+ */
+char *
+mime_fromaddr(char *name)
+{
+	char	*cp, *lastcp;
+	char	*res = NULL;
+	size_t	ressz = 1, rescur = 0;
+
+	if (name == NULL || *name == '\0')
+		return name;
+	if ((cp = routeaddr(name)) != NULL && cp > name) {
+		addconv(&res, &ressz, &rescur, name, cp - name);
+		lastcp = cp;
+	} else
+		cp = lastcp = name;
+	for ( ; *cp; cp++) {
+		switch (*cp) {
+		case '(':
+			addstr(&res, &ressz, &rescur, lastcp, cp - lastcp + 1);
+			lastcp = ++cp;
+			cp = skip_comment(cp);
+			if (--cp > lastcp)
+				addconv(&res, &ressz, &rescur, lastcp,
+						cp - lastcp);
+			lastcp = cp;
+			break;
+		case '"':
+			while (*cp) {
+				if (*++cp == '"')
+					break;
+				if (*cp == '\\' && cp[1])
+					cp++;
+			}
+			break;
+		}
+	}
+	if (cp > lastcp)
+		addstr(&res, &ressz, &rescur, lastcp, cp - lastcp);
+	res[rescur] = '\0';
+	cp = savestr(res);
+	free(res);
+	return cp;
+}
+
 /*
  * fwrite whilst adding prefix, if present.
  */
 size_t
-prefixwrite(ptr, size, nmemb, f, prefix, prefixlen)
-void *ptr;
-size_t size, nmemb, prefixlen;
-FILE *f;
-char *prefix;
+prefixwrite(void *ptr, size_t size, size_t nmemb, FILE *f,
+		char *prefix, size_t prefixlen)
 {
 	static FILE *lastf;
 	static char lastc = '\n';
@@ -1457,7 +1432,7 @@ char *prefix;
 	}
 	lastf = f;
 	for (rsz = size * nmemb; rsz; rsz--, p++, wsz++) {
-		sputc(*p, f);
+		putc(*p, f);
 		if (*p != '\n' || rsz == 1) {
 			continue;
 		}
@@ -1476,12 +1451,8 @@ char *prefix;
  * fwrite while checking for displayability.
  */
 static size_t
-fwrite_td(ptr, size, nmemb, f, flags, prefix, prefixlen)
-void *ptr;
-char *prefix;
-size_t size, nmemb, prefixlen;
-FILE *f;
-enum tdflags flags;
+fwrite_td(void *ptr, size_t size, size_t nmemb, FILE *f, enum tdflags flags,
+		char *prefix, size_t prefixlen)
 {
 	char *upper;
 	size_t sz, csize;
@@ -1489,12 +1460,12 @@ enum tdflags flags;
 	char *iptr, *nptr;
 	size_t inleft, outleft;
 #endif
-	char *mptr;
+	char *mptr, *xmptr, *mlptr = NULL;
 	size_t mptrsz;
 
 	csize = size * nmemb;
 	mptrsz = csize;
-	mptr = ac_alloc(mptrsz + 1);
+	mptr = xmptr = ac_alloc(mptrsz + 1);
 #ifdef	HAVE_ICONV
 	if ((flags & TD_ICONV) && iconvd != (iconv_t)-1) {
 	again:	inleft = csize;
@@ -1520,12 +1491,18 @@ enum tdflags flags;
 	upper = mptr + csize;
 	*upper = '\0';
 	if (flags & TD_ISPR) {
-		if ((csize = makeprint(mptr, upper - mptr)) >
-				upper - mptr)
-			csize = upper - mptr;
+		struct str	in, out;
+		in.s = mptr;
+		in.l = csize;
+		makeprint(&in, &out);
+		mptr = mlptr = out.s;
+		csize = out.l;
 	}
+	if (flags & TD_DELCTRL)
+		csize = delctrl(mptr, csize);
 	sz = prefixwrite(mptr, sizeof *mptr, csize, f, prefix, prefixlen);
-	ac_free(mptr);
+	ac_free(xmptr);
+	free(mlptr);
 	return sz;
 }
 
@@ -1533,13 +1510,9 @@ enum tdflags flags;
  * fwrite performing the given MIME conversion.
  */
 size_t
-mime_write(ptr, size, nmemb, f, convert, dflags, prefix, prefixlen)
-void *ptr;
-size_t size, nmemb, prefixlen;
-char *prefix;
-FILE *f;
-enum conversion convert;
-enum tdflags dflags;
+mime_write(void *ptr, size_t size, size_t nmemb, FILE *f,
+		enum conversion convert, enum tdflags dflags,
+		char *prefix, size_t prefixlen)
 {
 	struct str in, out;
 	size_t sz, csize;
@@ -1602,8 +1575,8 @@ enum tdflags dflags;
 		break;
 	case CONV_FROMHDR:
 		mime_fromhdr(&in, &out, TD_ISPR|TD_ICONV);
-		sz = fwrite_td(out.s, sizeof *out.s, out.l, f, TD_NONE,
-				prefix, prefixlen);
+		sz = fwrite_td(out.s, sizeof *out.s, out.l, f,
+				dflags&TD_DELCTRL, prefix, prefixlen);
 		free(out.s);
 		break;
 	case CONV_TOHDR:
