@@ -1,7 +1,7 @@
 /*
  * Nail - a mail user agent derived from Berkeley Mail.
  *
- * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ * Copyright (c) 2000-2004 Gunnar Ritter, Freiburg i. Br., Germany.
  */
 /*
  * Copyright (c) 1980, 1993
@@ -38,17 +38,18 @@
 
 #ifndef lint
 #ifdef	DOSCCS
-static char sccsid[] = "@(#)sendout.c	2.19 (gritter) 4/9/03";
+static char sccsid[] = "@(#)sendout.c	2.72 (gritter) 12/2/04";
 #endif
 #endif /* not lint */
 
 #include "rcv.h"
 #include "extern.h"
 #include <errno.h>
-
-#ifdef	HAVE_STRINGS_H
-#include <strings.h>
-#endif
+#include <sys/stat.h>
+#include <fcntl.h>
+#include <unistd.h>
+#include <time.h>
+#include "md5.h"
 
 /*
  * Mail -- a mail program
@@ -56,75 +57,37 @@ static char sccsid[] = "@(#)sendout.c	2.19 (gritter) 4/9/03";
  * Mail to others.
  */
 
-static const char	randfile[] = "/dev/urandom";
 static char	*send_boundary;
 
-static char	*makeboundary __P((void));
-static char	*getencoding __P((int));
-static struct name	*fixhead __P((struct header *, struct name *, int));
-static int	put_signature __P((FILE *, int));
-static int	attach_file __P((struct attachment *, FILE *));
-static int	make_multipart __P((struct header *, int, FILE *, FILE *,
-			const char *, const char *));
-static FILE	*infix __P((struct header *, FILE *));
-static int	savemail __P((char [], FILE *));
-static int	sendmail_internal __P((void *, int));
-static int	start_mta __P((struct name *, struct name *, FILE *));
-static void	message_id __P((FILE *));
-static void	date_field __P((FILE *));
-static int	fmt __P((char *, struct name *, FILE *, int, int));
-static int	infix_fw __P((FILE *, FILE *, struct message *,
-			struct name *, int));
-
-#define	BOUNDARRAY	8
+static char *getencoding(enum conversion convert);
+static struct name *fixhead(struct header *hp, struct name *tolist,
+		enum gfield addauto);
+static int put_signature(FILE *fo, int convert);
+static int attach_file(struct attachment *ap, FILE *fo, int dosign);
+static int make_multipart(struct header *hp, int convert, FILE *fi, FILE *fo,
+		const char *contenttype, const char *charset, int dosign);
+static FILE *infix(struct header *hp, FILE *fi, int dosign);
+static int savemail(char *name, FILE *fi);
+static int sendmail_internal(void *v, int recipient_record);
+static enum okay transfer(struct name *to, struct name *mailargs, FILE *input);
+static enum okay start_mta(struct name *to, struct name *mailargs, FILE *input);
+static void message_id(FILE *fo);
+static int fmt(char *str, struct name *np, FILE *fo, int comma,
+		int dropinvalid, int domime);
+static int infix_resend(FILE *fi, FILE *fo, struct message *mp,
+		struct name *to, int add_resent);
 
 /*
  * Generate a boundary for MIME multipart messages.
  */
-static char *
-makeboundary()
+char *
+makeboundary(void)
 {
-	int i, bd;
-	static unsigned msgc;
-	static pid_t pid;
-	static char bound[73];
-	time_t t;
-	unsigned long r[BOUNDARRAY];
-	char b[BOUNDARRAY][sizeof r[0] * 2 + 1];
+	static char bound[70];
+	time_t	now;
 
-	if (pid == 0) {
-		pid = getpid();
-		msgc = (unsigned)pid;
-	}
-	msgc *= 2053 * (unsigned)time(&t);
-	if ((bd = open(randfile, O_RDONLY)) != -1) {
-		for (i = 0; i < BOUNDARRAY; i++) {
-			if (read(bd, &r[i], sizeof r[0]) != sizeof r[0]) {
-				r[0] = 0L;
-				break;
-			}
-		}
-		close(bd);
-	} else
-		r[0] = 0L;
-	if (r[0] == 0L) {
-		srand((unsigned)msgc);
-		for (i = 0; i < BOUNDARRAY; i++) {
-			r[i] = 1L;
-			while (r[i] < 60466176L)
-				r[i] *= (unsigned long)rand();
-		}
-	}
-	snprintf(bound, 73,
-			"%.5s%.5s-=-%.5s%.5s-CUT-HERE-%.5s%.5s-=-%.5s%.5s",
-			itostr(36, (unsigned)r[0], b[0]),
-			itostr(36, (unsigned)r[1], b[1]),
-			itostr(36, (unsigned)r[2], b[2]),
-			itostr(36, (unsigned)r[3], b[3]),
-			itostr(36, (unsigned)r[4], b[4]),
-			itostr(36, (unsigned)r[5], b[5]),
-			itostr(36, (unsigned)r[6], b[6]),
-			itostr(36, (unsigned)r[7], b[7]));
+	time(&now);
+	snprintf(bound, sizeof bound, "=_%lx.%s", (long)now, getrandstring(48));
 	send_boundary = bound;
 	return send_boundary;
 }
@@ -133,7 +96,7 @@ makeboundary()
  * Get an encoding flag based on the given string.
  */
 static char *
-getencoding(convert)
+getencoding(enum conversion convert)
 {
 	switch (convert) {
 	case CONV_7BIT:
@@ -144,6 +107,8 @@ getencoding(convert)
 		return "quoted-printable";
 	case CONV_TOB64:
 		return "base64";
+	default:
+		break;
 	}
 	/*NOTREACHED*/
 	return NULL;
@@ -154,35 +119,45 @@ getencoding(convert)
  * the distribution list into the appropriate fields.
  */
 static struct name *
-fixhead(hp, tolist, addauto)
-	struct header *hp;
-	struct name *tolist;
+fixhead(struct header *hp, struct name *tolist, enum gfield addauto)
 {
 	struct name *np;
 	char	*cp;
 
-	hp->h_to = NIL;
-	hp->h_cc = NIL;
-	hp->h_bcc = NIL;
-	for (np = tolist; np != NIL; np = np->n_flink)
+	hp->h_to = NULL;
+	hp->h_cc = NULL;
+	hp->h_bcc = NULL;
+	hp->h_replyto = NULL;
+	for (np = tolist; np != NULL; np = np->n_flink)
 		if ((np->n_type & GMASK) == GTO)
 			hp->h_to =
-				cat(hp->h_to, nalloc(np->n_name, np->n_type));
+				cat(hp->h_to, nalloc(np->n_fullname,
+							np->n_type|GFULL));
 		else if ((np->n_type & GMASK) == GCC)
 			hp->h_cc =
-				cat(hp->h_cc, nalloc(np->n_name, np->n_type));
+				cat(hp->h_cc, nalloc(np->n_fullname,
+							np->n_type|GFULL));
 		else if ((np->n_type & GMASK) == GBCC)
 			hp->h_bcc =
-				cat(hp->h_bcc, nalloc(np->n_name, np->n_type));
-	if (addauto && (cp = value("autocc")) != NULL && *cp) {
-		np = usermap(checkaddrs(extract(cp, GCC)));
-		hp->h_cc = cat(hp->h_cc, nalloc(np->n_name, np->n_type));
-		tolist = cat(tolist, nalloc(np->n_name, np->n_type));
+				cat(hp->h_bcc, nalloc(np->n_fullname,
+							np->n_type|GFULL));
+		else if ((np->n_type & GMASK) == GREPLYTO)
+			hp->h_replyto =
+				cat(hp->h_replyto, nalloc(np->n_fullname,
+							np->n_type|GFULL));
+	if (addauto&GCC && (cp = value("autocc")) != NULL && *cp) {
+		np = checkaddrs(sextract(cp, GCC|GFULL));
+		hp->h_cc = cat(hp->h_cc, np);
+		tolist = cat(tolist, np);
 	}
-	if (addauto && (cp = value("autobcc")) != NULL && *cp) {
-		np = usermap(checkaddrs(extract(cp, GBCC)));
-		hp->h_bcc = cat(hp->h_bcc, nalloc(np->n_name, np->n_type));
-		tolist = cat(tolist, nalloc(np->n_name, np->n_type));
+	if (addauto&GBCC && (cp = value("autobcc")) != NULL && *cp) {
+		np = checkaddrs(sextract(cp, GBCC|GFULL));
+		hp->h_bcc = cat(hp->h_bcc, np);
+		tolist = cat(tolist, np);
+	}
+	if (addauto&GREPLYTO && (cp = value("replyto")) != NULL && *cp) {
+		np = checkaddrs(sextract(cp, GREPLYTO|GFULL));
+		hp->h_replyto = cat(hp->h_replyto, np);
 	}
 	return tolist;
 }
@@ -197,8 +172,7 @@ fixhead(hp, tolist, addauto)
  * Put the signature file at fo.
  */
 static int
-put_signature(fo, convert)
-FILE *fo;
+put_signature(FILE *fo, int convert)
 {
 	char *sig, buf[INFIX_BUF], c = '\n';
 	FILE *fsig;
@@ -230,7 +204,7 @@ FILE *fo;
 	}
 	Fclose(fsig);
 	if (c != '\n')
-		sputc('\n', fo);
+		putc('\n', fo);
 	return 0;
 }
 
@@ -238,13 +212,11 @@ FILE *fo;
  * Write an attachment to the file buffer, converting to MIME.
  */
 static int
-attach_file(ap, fo)
-struct attachment *ap;
-FILE *fo;
+attach_file(struct attachment *ap, FILE *fo, int dosign)
 {
 	FILE *fi;
 	char *charset = NULL, *contenttype = NULL, *basename;
-	int convert = CONV_TOB64;
+	enum conversion convert = CONV_TOB64;
 	int err = 0;
 	enum mimeclean isclean;
 	size_t sz;
@@ -263,13 +235,14 @@ FILE *fo;
 		contenttype = ap->a_content_type;
 	else
 		contenttype = mime_filecontent(basename);
-	convert = get_mime_convert(fi, &contenttype, &charset, &isclean);
+	convert = get_mime_convert(fi, &contenttype, &charset, &isclean,
+			dosign);
 	fprintf(fo,
 		"\n--%s\n"
 		"Content-Type: %s",
 		send_boundary, contenttype);
 	if (charset == NULL)
-		sputc('\n', fo);
+		putc('\n', fo);
 	else
 		fprintf(fo, ";\n charset=%s\n", charset);
 	if (ap->a_content_disposition == NULL)
@@ -287,7 +260,7 @@ FILE *fo;
 	if (ap->a_content_description)
 		fprintf(fo, "Content-Description: %s\n",
 				ap->a_content_description);
-	sputc('\n', fo);
+	putc('\n', fo);
 	buf = smalloc(bufsize = INFIX_BUF);
 	if (convert == CONV_TOQP) {
 		fflush(fi);
@@ -317,11 +290,8 @@ FILE *fo;
  * Generate the body of a MIME multipart message.
  */
 static int
-make_multipart(hp, convert, fi, fo, contenttype, charset)
-struct header *hp;
-FILE *fi, *fo;
-const char *contenttype;
-const char *charset;
+make_multipart(struct header *hp, int convert, FILE *fi, FILE *fo,
+		const char *contenttype, const char *charset, int dosign)
 {
 	struct attachment *att;
 
@@ -334,7 +304,8 @@ const char *charset;
 		fprintf(fo, "Content-Type: %s", contenttype);
 		if (charset)
 			fprintf(fo, "; charset=%s", charset);
-		fprintf(fo, "\nContent-Transfer-Encoding: %s\n\n",
+		fprintf(fo, "\nContent-Transfer-Encoding: %s\n"
+				"Content-Disposition: inline\n\n",
 				getencoding(convert));
 		buf = smalloc(bufsize = INFIX_BUF);
 		if (convert == CONV_TOQP) {
@@ -362,12 +333,12 @@ const char *charset;
 		if (ferror(fi))
 			return -1;
 		if (c != '\n')
-			sputc('\n', fo);
+			putc('\n', fo);
 		if (charset != NULL)
 			put_signature(fo, convert);
 	}
 	for (att = hp->h_attach; att != NULL; att = att->a_flink)
-		if (attach_file(att, fo) != 0)
+		if (attach_file(att, fo, dosign) != 0)
 			return -1;
 	/* the final boundary with two attached dashes */
 	fprintf(fo, "\n--%s--\n", send_boundary);
@@ -379,37 +350,36 @@ const char *charset;
  * and return the new file.
  */
 static FILE *
-infix(hp, fi)
-	struct header *hp;
-	FILE *fi;
+infix(struct header *hp, FILE *fi, int dosign)
 {
 	FILE *nfo, *nfi;
 	char *tempMail;
 #ifdef	HAVE_ICONV
-	char *tcs;
+	char *tcs, *convhdr = NULL;
 #endif
 	enum mimeclean isclean;
-	int convert;
-	char *charset = NULL, *contenttype = NULL, *convhdr = NULL;
+	enum conversion convert;
+	char *charset = NULL, *contenttype = NULL;
 
 	if ((nfo = Ftemp(&tempMail, "Rs", "w", 0600, 1)) == NULL) {
 		perror(catgets(catd, CATSET, 178, "temporary mail file"));
 		return(NULL);
 	}
-	if ((nfi = Fopen(tempMail, "r")) == (FILE *)NULL) {
+	if ((nfi = Fopen(tempMail, "r")) == NULL) {
 		perror(tempMail);
-		(void) Fclose(nfo);
+		Fclose(nfo);
 		return(NULL);
 	}
-	(void) rm(tempMail);
+	rm(tempMail);
 	Ftfree(&tempMail);
-	convert = get_mime_convert(fi, &contenttype, &charset, &isclean);
+	convert = get_mime_convert(fi, &contenttype, &charset,
+			&isclean, dosign);
 #ifdef	HAVE_ICONV
 	tcs = gettcharset();
 	if (((isclean & MIME_HASNUL) == 0 && (isclean & MIME_HIGHBIT) &&
 				strcmp(charset, tcs)) ||
 			(convhdr = need_hdrconv(hp,
-					GTO|GSUBJECT|GCC|GBCC|GIDENT)) != 0) {
+				GTO|GSUBJECT|GCC|GBCC|GREPLYTO|GIDENT)) != 0) {
 		if (convhdr)
 			charset = convhdr;
 		if (iconvd != (iconv_t)-1)
@@ -428,10 +398,10 @@ infix(hp, fi)
 #endif	/* HAVE_ICONV */
 	if (puthead(hp, nfo,
 		   GTO|GSUBJECT|GCC|GBCC|GNL|GCOMMA|GUA|GMIME
-		   |GMSGID|GIDENT|GREF|GDATE,
-		   convert, contenttype, charset)) {
-		(void) Fclose(nfo);
-		(void) Fclose(nfi);
+		   |GMSGID|GIDENT|GREPLYTO|GREF|GDATE,
+		   SEND_MBOX, convert, contenttype, charset)) {
+		Fclose(nfo);
+		Fclose(nfi);
 #ifdef	HAVE_ICONV
 		if (iconvd != (iconv_t)-1) {
 			iconv_close(iconvd);
@@ -448,9 +418,9 @@ infix(hp, fi)
 #endif
 	if (hp->h_attach != NULL) {
 		if (make_multipart(hp, convert, fi, nfo,
-					contenttype, charset) != 0) {
-			(void) Fclose(nfo);
-			(void) Fclose(nfi);
+					contenttype, charset, dosign) != 0) {
+			Fclose(nfo);
+			Fclose(nfi);
 #ifdef	HAVE_ICONV
 			if (iconvd != (iconv_t)-1) {
 				iconv_close(iconvd);
@@ -480,8 +450,8 @@ infix(hp, fi)
 			}
 			if (mime_write(buf, sizeof *buf, sz, nfo, convert,
 					TD_ICONV, NULL, (size_t)0) == 0) {
-				(void) Fclose(nfo);
-				(void) Fclose(nfi);
+				Fclose(nfo);
+				Fclose(nfi);
 				perror("read");
 #ifdef	HAVE_ICONV
 				if (iconvd != (iconv_t)-1) {
@@ -495,8 +465,8 @@ infix(hp, fi)
 		}
 		free(buf);
 		if (ferror(fi)) {
-			(void) Fclose(nfo);
-			(void) Fclose(nfi);
+			Fclose(nfo);
+			Fclose(nfi);
 			perror("read");
 #ifdef	HAVE_ICONV
 			if (iconvd != (iconv_t)-1) {
@@ -515,15 +485,15 @@ infix(hp, fi)
 		iconvd = (iconv_t)-1;
 	}
 #endif
-	(void) fflush(nfo);
+	fflush(nfo);
 	if (ferror(nfo)) {
 		perror(catgets(catd, CATSET, 180, "temporary mail file"));
-		(void) Fclose(nfo);
-		(void) Fclose(nfi);
+		Fclose(nfo);
+		Fclose(nfi);
 		return NULL;
 	}
-	(void) Fclose(nfo);
-	(void) Fclose(fi);
+	Fclose(nfo);
+	Fclose(fi);
 	fflush(nfi);
 	rewind(nfi);
 	return(nfi);
@@ -535,9 +505,7 @@ infix(hp, fi)
 
 /*ARGSUSED*/
 static int
-savemail(name, fi)
-	char name[];
-	FILE *fi;
+savemail(char *name, FILE *fi)
 {
 	FILE *fo;
 	char *buf;
@@ -545,11 +513,12 @@ savemail(name, fi)
 	char *p;
 	time_t now;
 	int prependnl = 0;
+	int error = 0;
 
 	buf = smalloc(bufsize = LINESIZE);
 	time(&now);
-	if ((fo = Fopen(name, "r+")) == (FILE *)NULL) {
-		if ((fo = Fopen(name, "wx")) == (FILE *)NULL) {
+	if ((fo = Zopen(name, "a+", NULL)) == NULL) {
+		if ((fo = Zopen(name, "wx", NULL)) == NULL) {
 			perror(name);
 			free(buf);
 			return (-1);
@@ -576,7 +545,7 @@ savemail(name, fi)
 			}
 			fflush(fo);
 			if (prependnl) {
-				sputc('\n', fo);
+				putc('\n', fo);
 				fflush(fo);
 			}
 		}
@@ -593,18 +562,21 @@ savemail(name, fi)
 				p++;
 			if (strncmp(p, "From ", 5) == 0)
 				/* we got a masked From line */
-				sputc('>', fo);
+				putc('>', fo);
 		} else if (strncmp(buf, "From ", 5) == 0)
-			sputc('>', fo);
+			putc('>', fo);
 		fwrite(buf, sizeof *buf, buflen, fo);
 	}
 	if (buflen && *(buf + buflen - 1) != '\n')
-		(void) sputc('\n', fo);
-	(void) sputc('\n', fo);
-	(void) fflush(fo);
-	if (ferror(fo))
+		putc('\n', fo);
+	putc('\n', fo);
+	fflush(fo);
+	if (ferror(fo)) {
 		perror(name);
-	(void) Fclose(fo);
+		error = -1;
+	}
+	if (Fclose(fo) != 0)
+		error = -1;
 	fflush(fi);
 	rewind(fi);
 	/*
@@ -615,43 +587,37 @@ savemail(name, fi)
 	 */
 	lseek(fileno(fi), 0, SEEK_SET);
 	free(buf);
-	return (0);
+	return error;
 }
 
 /*
  * Interface between the argument list and the mail1 routine
  * which does all the dirty work.
  */
-int
-mail(to, cc, bcc, smopts, subject, attach, quotefile, recipient_record, tflag)
-	struct name *to, *cc, *bcc, *smopts;
-	struct attachment *attach;
-	char *subject, *quotefile;
+int 
+mail(struct name *to, struct name *cc, struct name *bcc,
+		struct name *smopts, char *subject, struct attachment *attach,
+		char *quotefile, int recipient_record, int tflag)
 {
 	struct header head;
 	struct str in, out;
 
+	memset(&head, 0, sizeof head);
 	/* The given subject may be in RFC1522 format. */
 	if (subject != NULL) {
 		in.s = subject;
 		in.l = strlen(subject);
 		mime_fromhdr(&in, &out, TD_ISPR | TD_ICONV);
 		head.h_subject = out.s;
-	} else
-		head.h_subject = NULL;
+	}
 	if (tflag == 0) {
 		head.h_to = to;
 		head.h_cc = cc;
 		head.h_bcc = bcc;
-	} else {
-		head.h_to = NULL;
-		head.h_cc = NULL;
-		head.h_bcc = NULL;
 	}
-	head.h_ref = NULL;
 	head.h_attach = attach;
 	head.h_smopts = smopts;
-	mail1(&head, 0, NULL, quotefile, recipient_record, tflag);
+	mail1(&head, 0, NULL, quotefile, recipient_record, 0, tflag);
 	if (subject != NULL)
 		free(out.s);
 	return(0);
@@ -661,51 +627,87 @@ mail(to, cc, bcc, smopts, subject, attach, quotefile, recipient_record, tflag)
  * Send mail to a bunch of user names.  The interface is through
  * the mail routine below.
  */
-static int
-sendmail_internal(v, recipient_record)
-	void *v;
+static int 
+sendmail_internal(void *v, int recipient_record)
 {
 	char *str = v;
 	struct header head;
 
-	head.h_to = extract(str, GTO);
-	head.h_subject = NULL;
-	head.h_cc = NIL;
-	head.h_bcc = NIL;
-	head.h_ref = NIL;
-	head.h_attach = NULL;
-	head.h_smopts = NIL;
-	mail1(&head, 0, NULL, NULL, recipient_record, 0);
+	memset(&head, 0, sizeof head);
+	head.h_to = extract(str, GTO|GFULL);
+	mail1(&head, 0, NULL, NULL, recipient_record, 0, 0);
 	return(0);
 }
 
-int
-sendmail(v)
-	void *v;
+int 
+sendmail(void *v)
 {
 	return sendmail_internal(v, 0);
 }
 
-int
-Sendmail(v)
-	void *v;
+int 
+Sendmail(void *v)
 {
 	return sendmail_internal(v, 1);
 }
 
+static enum okay
+transfer(struct name *to, struct name *mailargs, FILE *input)
+{
+	char	o[LINESIZE], *cp;
+	struct name	*np, *nt;
+	int	cnt = 0;
+	FILE	*ef;
+	enum okay	ok = OKAY;
+
+	np = to;
+	while (np) {
+		snprintf(o, sizeof o, "smime-encrypt-%s", np->n_name);
+		if ((cp = value(o)) != NULL) {
+			if ((ef = smime_encrypt(input, cp, np->n_name)) != 0) {
+				nt = nalloc(np->n_name,
+					np->n_type & ~(GFULL|GSKIN));
+				if (start_mta(nt, mailargs, ef) != OKAY)
+					ok = STOP;
+				Fclose(ef);
+			} else {
+				fprintf(stderr, "Message not sent to <%s>\n",
+						np->n_name);
+				senderr++;
+			}
+			rewind(input);
+			if (np->n_flink)
+				np->n_flink->n_blink = np->n_blink;
+			if (np->n_blink)
+				np->n_blink->n_flink = np->n_flink;
+			if (np == to)
+				to = np->n_flink;
+			np = np->n_flink;
+		} else {
+			cnt++;
+			np = np->n_flink;
+		}
+	}
+	if (cnt) {
+		if (value("smime-force-encryption") ||
+				start_mta(to, mailargs, input) != OKAY)
+			ok = STOP;
+	}
+	return ok;
+}
+
 /*
  * Start the Mail Transfer Agent
  * mailing to namelist and stdin redirected to input.
  */
-static int
-start_mta(to, mailargs, input)
-struct name *to, *mailargs;
-FILE* input;
+static enum okay
+start_mta(struct name *to, struct name *mailargs, FILE *input)
 {
 	char **args = NULL, **t;
 	pid_t pid;
 	sigset_t nset;
 	char *cp, *smtp;
+	enum okay	ok = STOP;
 
 	if ((smtp = value("smtp")) == NULL) {
 		args = unpack(cat(mailargs, to));
@@ -715,10 +717,10 @@ FILE* input;
 			for (t = args; *t != NULL; t++)
 				printf(" \"%s\"", *t);
 			printf("\n");
-			return 0;
+			return OKAY;
 		}
 	} else if (debug || value("debug"))
-		return 0;
+		return OKAY;
 	/*
 	 * Fork, set up the temporary mail file as standard
 	 * input for "mail", and exec with the user list we generated
@@ -728,7 +730,7 @@ FILE* input;
 		perror("fork");
 		savedeadletter(input);
 		senderr++;
-		return 1;
+		return STOP;
 	}
 	if (pid == 0) {
 		sigemptyset(&nset);
@@ -738,6 +740,7 @@ FILE* input;
 		sigaddset(&nset, SIGTSTP);
 		sigaddset(&nset, SIGTTIN);
 		sigaddset(&nset, SIGTTOU);
+		freopen("/dev/null", "r", stdin);
 		if (smtp != NULL) {
 			prepare_child(&nset, 0, 1);
 			if (smtp_mta(smtp, to, input) == 0)
@@ -747,7 +750,7 @@ FILE* input;
 			if ((cp = value("sendmail")) != NULL)
 				cp = expand(cp);
 			else
-				cp = PATH_SENDMAIL;
+				cp = SENDMAIL;
 			execv(cp, args);
 			perror(cp);
 		}
@@ -756,54 +759,100 @@ FILE* input;
 				". . . message not sent.\n"), stderr);
 		_exit(1);
 	}
-	if (value("verbose") != NULL || value("sendwait"))
-		(void) wait_child(pid);
-	else
+	if (value("verbose") != NULL || value("sendwait")) {
+		if (wait_child(pid) == 0)
+			ok = OKAY;
+		else
+			senderr++;
+	} else {
+		ok = OKAY;
 		free_child(pid);
-	return 0;
+	}
+	return ok;
+}
+
+/*
+ * Record outgoing mail if instructed to do so.
+ */
+static enum okay
+mightrecord(FILE *fp, struct name *to, int recipient_record)
+{
+	char	*cp, *cq, *ep;
+
+	if (recipient_record) {
+		cq = skin(to->n_name);
+		cp = salloc(strlen(cq) + 1);
+		strcpy(cp, cq);
+		for (cq = cp; *cq && *cq != '@'; cq++);
+		*cq = '\0';
+	} else
+		cp = value("record");
+	if (cp != NULL) {
+		ep = expand(cp);
+		if (value("outfolder") && *ep != '/' && *ep != '+' &&
+				which_protocol(ep) == PROTO_FILE) {
+			cq = salloc(strlen(cp) + 2);
+			cq[0] = '+';
+			strcpy(&cq[1], cp);
+			cp = cq;
+			ep = expand(cp);
+		}
+		if (savemail(ep, fp) != 0) {
+			fprintf(stderr,
+				"Error while saving message to %s - "
+				"message not sent\n", ep);
+			rewind(fp);
+			exit_status |= 1;
+			savedeadletter(fp);
+			return STOP;
+		}
+	}
+	return OKAY;
 }
 
 /*
  * Mail a message on standard input to the people indicated
  * in the passed header.  (Internal interface).
  */
-void
-mail1(hp, printheaders, quote, quotefile, recipient_record, tflag)
-	struct header *hp;
-	int printheaders;
-	struct message *quote;
-	char *quotefile;
+enum okay 
+mail1(struct header *hp, int printheaders, struct message *quote,
+		char *quotefile, int recipient_record, int doprefix, int tflag)
 {
-	char *cp, *cq;
 	struct name *to;
 	FILE *mtf, *nmtf;
+	enum okay	ok = STOP;
+	int	dosign = -1;
 
 #ifdef	notdef
 	if ((hp->h_to = checkaddrs(hp->h_to)) == NULL) {
 		senderr++;
-		return;
+		return STOP;
 	}
 #endif
 	/*
 	 * Collect user's mail from standard input.
 	 * Get the result as mtf.
 	 */
-	if ((mtf = collect(hp, printheaders, quote, quotefile, tflag)) == NULL)
-		return;
+	if ((mtf = collect(hp, printheaders, quote, quotefile, doprefix,
+					tflag)) == NULL)
+		return STOP;
 	if (value("interactive") != NULL) {
 		if (((value("bsdcompat") || value("askatend"))
 					&& (value("askcc") != NULL ||
 					value("askbcc") != NULL)) ||
-				value("askattach") != NULL) {
+				value("askattach") != NULL ||
+				value("asksign") != NULL) {
 			if (value("askcc") != NULL)
 				grabh(hp, GCC, 1);
 			if (value("askbcc") != NULL)
 				grabh(hp, GBCC, 1);
 			if (value("askattach") != NULL)
 				hp->h_attach = edit_attachments(hp->h_attach);
+			if (value("asksign") != NULL)
+				dosign = yorn("Sign this message (y/n)? ");
 		} else {
 			printf(catgets(catd, CATSET, 183, "EOT\n"));
-			(void) fflush(stdout);
+			fflush(stdout);
 		}
 	}
 	if (fsize(mtf) == 0) {
@@ -814,28 +863,40 @@ mail1(hp, printheaders, quote, quotefile, recipient_record, tflag)
 			printf(catgets(catd, CATSET, 185,
 				"Null message body; hope that's ok\n"));
 	}
+	if (dosign < 0) {
+		if (value("smime-sign") != NULL)
+			dosign = 1;
+		else
+			dosign = 0;
+	}
 	/*
 	 * Now, take the user names from the combined
 	 * to and cc lists and do all the alias
 	 * processing.
 	 */
 	senderr = 0;
-	if ((to = usermap(cat(hp->h_bcc, cat(hp->h_to, hp->h_cc)))) == NIL) {
+	if ((to = usermap(cat(hp->h_bcc, cat(hp->h_to, hp->h_cc)))) == NULL) {
 		printf(catgets(catd, CATSET, 186, "No recipients specified\n"));
 		senderr++;
 	}
-	to = fixhead(hp, to, 1);
-	if ((nmtf = infix(hp, mtf)) == (FILE *)NULL) {
+	to = fixhead(hp, to, GCC|GBCC|GREPLYTO);
+	if ((nmtf = infix(hp, mtf, dosign)) == NULL) {
 		/* fprintf(stderr, ". . . message lost, sorry.\n"); */
 		perror("");
-		senderr++;
+	fail:	senderr++;
 		rewind(mtf);
 		savedeadletter(mtf);
 		fputs(catgets(catd, CATSET, 187,
 				". . . message not sent.\n"), stderr);
-		return;
+		return STOP;
 	}
 	mtf = nmtf;
+	if (dosign) {
+		if ((nmtf = smime_sign(mtf)) == NULL)
+			goto fail;
+		Fclose(mtf);
+		mtf = nmtf;
+	}
 	/*
 	 * Look through the recipient list for names with /'s
 	 * in them which we write to as files directly.
@@ -844,74 +905,35 @@ mail1(hp, printheaders, quote, quotefile, recipient_record, tflag)
 	if (senderr)
 		savedeadletter(mtf);
 	to = elide(to);
-	if (count(to) == 0)
+	if (count(to) == 0) {
+		ok = OKAY;
 		goto out;
-	if (recipient_record) {
-		cq = skin(to->n_name);
-		cp = salloc(strlen(cq) + 1);
-		strcpy(cp, cq);
-		for (cq = cp; *cq && *cq != '@'; cq++);
-		*cq = '\0';
-	} else
-		cp = value("record");
-	if (cp != NULL) {
-		if (value("outfolder") && *cp != '/' && *cp != '+') {
-			cq = salloc(strlen(cp) + 2);
-			cq[0] = '+';
-			strcpy(&cq[1], cp);
-			cp = cq;
-		}
-		savemail(expand(cp), mtf);
 	}
-	start_mta(to, hp->h_smopts, mtf);
+	if (mightrecord(mtf, to, recipient_record) != OKAY)
+		goto out;
+	ok = transfer(to, hp->h_smopts, mtf);
 out:
 	Fclose(mtf);
+	return ok;
 }
 
 /*
  * Create a Message-Id: header field.
- * Use either the host name or the host part of the from address.
+ * Use either the host name or the from address.
  */
 static void
-message_id(fo)
-FILE *fo;
+message_id(FILE *fo)
 {
-	static unsigned long msgc;
-	time_t t;
-	char *domainpart;
-	int rd;
-	long randbuf;
-	char datestr[sizeof (time_t) * 2 + 1];
-	char pidstr[sizeof (pid_t) * 2+ 1];
-	char countstr[sizeof msgc * 2 + 1];
-	char randstr[sizeof randbuf * 2+ 1];
+	char	*cp;
+	time_t	now;
 
-
-	msgc++;
-	time(&t);
-	rd = open(randfile, O_RDONLY);
-	if (rd != -1) {
-		if (read(rd, &randbuf, sizeof randbuf) != sizeof randbuf) {
-			srand((unsigned)(msgc * t));
-			randbuf = (long)rand();
-		}
-		close(rd);
-	} else {
-		srand((unsigned)(msgc * t));
-		randbuf = (long)rand();
-	}
-	itostr(16, (unsigned)t, datestr);
-	itostr(36, (unsigned)getpid(), pidstr);
-	itostr(36, (unsigned)msgc, countstr);
-	itostr(36, (unsigned)randbuf, randstr);
-	if ((domainpart = skin(value("from"))) != NULL
-			&& (domainpart = strchr(domainpart, '@')) != NULL)
-		domainpart++;
-	else
-		domainpart = nodename();
-	fprintf(fo, "Message-ID: <%s.%s%.5s%s%.5s@%s>\n",
-			datestr, progname, pidstr, countstr, randstr,
-			domainpart);
+	time(&now);
+	if ((cp = value("hostname")) != NULL)
+		fprintf(fo, "Message-ID: <%lx.%s@%s>\n",
+				(long)now, getrandstring(24), cp);
+	else if ((cp = skin(myaddr())) != NULL && strchr(cp, '@') != NULL)
+		fprintf(fo, "Message-ID: <%lx.%s%%%s>\n",
+				(long)now, getrandstring(16), cp);
 }
 
 static const char *weekday_names[] = {
@@ -929,9 +951,8 @@ const char *month_names[] = {
  * because numeric timezones are easier to read and because $TZ is
  * not set on most GNU systems.
  */
-static void
-date_field(fo)
-FILE *fo;
+int
+mkdate(FILE *fo, const char *field)
 {
 	time_t t;
 	struct tm *tmptr;
@@ -945,7 +966,8 @@ FILE *fo;
 	tmptr = localtime(&t);
 	if (tmptr->tm_isdst > 0)
 		tzdiff_hour++;
-	fprintf(fo, "Date: %s, %02d %s %04d %02d:%02d:%02d %+03d%02d\n",
+	return fprintf(fo, "%s: %s, %02d %s %04d %02d:%02d:%02d %+03d%02d\n",
+			field,
 			weekday_names[tmptr->tm_wday],
 			tmptr->tm_mday, month_names[tmptr->tm_mon],
 			tmptr->tm_year + 1900, tmptr->tm_hour,
@@ -954,15 +976,17 @@ FILE *fo;
 }
 
 #define	FMT_CC_AND_BCC	{ \
-				if (hp->h_cc != NIL && w & GCC) { \
+				if (hp->h_cc != NULL && w & GCC) { \
 					if (fmt("Cc:", hp->h_cc, fo, \
-							w&GCOMMA, 0)) \
+							w&(GCOMMA|GFILES), 0, \
+							action!=SEND_TODISP)) \
 						return 1; \
 					gotcha++; \
 				} \
-				if (hp->h_bcc != NIL && w & GBCC) { \
+				if (hp->h_bcc != NULL && w & GBCC) { \
 					if (fmt("Bcc:", hp->h_bcc, fo, \
-							w&GCOMMA, 0)) \
+							w&(GCOMMA|GFILES), 0, \
+							action!=SEND_TODISP)) \
 						return 1; \
 					gotcha++; \
 				} \
@@ -972,14 +996,12 @@ FILE *fo;
  * passed file buffer.
  */
 int
-puthead(hp, fo, w, convert, contenttype, charset)
-	struct header *hp;
-	FILE *fo;
-	enum gfield w;
-	char *contenttype, *charset;
+puthead(struct header *hp, FILE *fo, enum gfield w,
+		enum sendaction action, enum conversion convert,
+		char *contenttype, char *charset)
 {
 	int gotcha;
-	char *addr, *cp;
+	char *addr/*, *cp*/;
 	int stealthmua;
 	struct name *np;
 
@@ -990,58 +1012,52 @@ puthead(hp, fo, w, convert, contenttype, charset)
 		stealthmua = 0;
 	gotcha = 0;
 	if (w & GDATE) {
-		date_field(fo), gotcha++;
+		mkdate(fo, "Date"), gotcha++;
 	}
 	if (w & GIDENT) {
 		addr = myaddr();
 		if (addr != NULL) {
 			if (mime_name_invalid(addr, 1))
 				return 1;
+			/*
 			if ((cp = value("smtp")) == NULL
 					|| strcmp(cp, "localhost") == 0)
 				fprintf(fo, "Sender: %s\n", username());
+			*/
 			fwrite("From: ", sizeof (char), 6, fo);
 			if (mime_write(addr, sizeof *addr, strlen(addr), fo,
-					convert == CONV_TODISP ?
+					action == SEND_TODISP ?
 					CONV_NONE:CONV_TOHDR_A,
-					convert == CONV_TODISP ?
+					action == SEND_TODISP ?
 					TD_ISPR|TD_ICONV:TD_ICONV,
 					NULL, (size_t)0) == 0)
 				return 1;
 			gotcha++;
-			sputc('\n', fo);
+			putc('\n', fo);
 		}
 		addr = value("ORGANIZATION");
 		if (addr != NULL) {
 			fwrite("Organization: ", sizeof (char), 14, fo);
 			if (mime_write(addr, sizeof *addr, strlen(addr), fo,
-					convert == CONV_TODISP ?
+					action == SEND_TODISP ?
 					CONV_NONE:CONV_TOHDR,
-					convert == CONV_TODISP ?
+					action == SEND_TODISP ?
 					TD_ISPR|TD_ICONV:TD_ICONV,
 					NULL, (size_t)0) == 0)
 				return 1;
 			gotcha++;
-			sputc('\n', fo);
-		}
-		addr = value("replyto");
-		if (addr != NULL) {
-			if (mime_name_invalid(addr, 1))
-				return 1;
-			fwrite("Reply-To: ", sizeof (char), 10, fo);
-			if (mime_write(addr, sizeof *addr, strlen(addr), fo,
-					convert == CONV_TODISP ?
-					CONV_NONE:CONV_TOHDR_A,
-					convert == CONV_TODISP ?
-					TD_ISPR|TD_ICONV:TD_ICONV,
-					NULL, (size_t)0) == 0)
-				return 1;
-			gotcha++;
-			sputc('\n', fo);
+			putc('\n', fo);
 		}
 	}
-	if (hp->h_to != NIL && w & GTO) {
-		if (fmt("To:", hp->h_to, fo, w&GCOMMA, 0))
+	if (hp->h_replyto != NULL && w & GREPLYTO) {
+		if (fmt("Reply-To:", hp->h_replyto, fo, w&(GCOMMA|GFILES), 0,
+					action != SEND_TODISP))
+			return 1;
+		gotcha++;
+	}
+	if (hp->h_to != NULL && w & GTO) {
+		if (fmt("To:", hp->h_to, fo, w&(GCOMMA|GFILES), 0,
+					action!=SEND_TODISP))
 			return 1;
 		gotcha++;
 	}
@@ -1053,18 +1069,18 @@ puthead(hp, fo, w, convert, contenttype, charset)
 			fwrite("Re: ", sizeof (char), 4, fo);
 			if (mime_write(hp->h_subject + 4, sizeof *hp->h_subject,
 					strlen(hp->h_subject + 4),
-					fo, convert == CONV_TODISP ?
+					fo, action == SEND_TODISP ?
 					CONV_NONE:CONV_TOHDR,
-					convert == CONV_TODISP ?
+					action == SEND_TODISP ?
 					TD_ISPR|TD_ICONV:TD_ICONV,
 					NULL, (size_t)0) == 0)
 				return 1;
-		} else {
+		} else if (*hp->h_subject) {
 			if (mime_write(hp->h_subject, sizeof *hp->h_subject,
 					strlen(hp->h_subject),
-					fo, convert == CONV_TODISP ?
+					fo, action == SEND_TODISP ?
 					CONV_NONE:CONV_TOHDR,
-					convert == CONV_TODISP ?
+					action == SEND_TODISP ?
 					TD_ISPR|TD_ICONV:TD_ICONV,
 					NULL, (size_t)0) == 0)
 				return 1;
@@ -1076,8 +1092,8 @@ puthead(hp, fo, w, convert, contenttype, charset)
 		FMT_CC_AND_BCC
 	if (w & GMSGID && stealthmua == 0)
 		message_id(fo), gotcha++;
-	if (hp->h_ref != NIL && w & GREF) {
-		fmt("References:", hp->h_ref, fo, 0, 1);
+	if (hp->h_ref != NULL && w & GREF) {
+		fmt("References:", hp->h_ref, fo, 0, 1, 0);
 		if ((np = hp->h_ref) != NULL && np->n_name) {
 			while (np->n_flink)
 				np = np->n_flink;
@@ -1104,7 +1120,7 @@ puthead(hp, fo, w, convert, contenttype, charset)
 		}
 	}
 	if (gotcha && w & GNL)
-		(void) sputc('\n', fo);
+		putc('\n', fo);
 	return(0);
 }
 
@@ -1112,27 +1128,27 @@ puthead(hp, fo, w, convert, contenttype, charset)
  * Format the given header line to not exceed 72 characters.
  */
 static int
-fmt(str, np, fo, comma, dropinvalid)
-	char *str;
-	struct name *np;
-	FILE *fo;
-	int comma;
+fmt(char *str, struct name *np, FILE *fo, int flags, int dropinvalid,
+		int domime)
 {
 	int col, len, count = 0;
-	int is_to = 0;
+	int is_to = 0, comma;
 
-	comma = comma ? 1 : 0;
+	comma = flags&GCOMMA ? 1 : 0;
 	col = strlen(str);
 	if (col) {
 		fwrite(str, sizeof *str, strlen(str), fo);
-		if ((col == 3 && asccasecmp(str, "to:") == 0) ||
-			(col == 10 && asccasecmp(str, "Resent-To:") == 0))
+		if ((flags&GFILES) == 0 &&
+				col == 3 && asccasecmp(str, "to:") == 0 ||
+				col == 3 && asccasecmp(str, "cc:") == 0 ||
+				col == 4 && asccasecmp(str, "bcc:") == 0 ||
+				col == 10 && asccasecmp(str, "Resent-To:") == 0)
 			is_to = 1;
 	}
-	for (; np != NIL; np = np->n_flink) {
+	for (; np != NULL; np = np->n_flink) {
 		if (is_to && is_fileaddr(np->n_name))
 			continue;
-		if (np->n_flink == NIL)
+		if (np->n_flink == NULL)
 			comma = 0;
 		if (mime_name_invalid(np->n_name, !dropinvalid)) {
 			if (dropinvalid)
@@ -1140,114 +1156,114 @@ fmt(str, np, fo, comma, dropinvalid)
 			else
 				return 1;
 		}
-		len = strlen(np->n_name);
+		len = strlen(np->n_fullname);
 		col++;		/* for the space */
 		if (count && col + len + comma > 72 && col > 1) {
 			fputs("\n ", fo);
 			col = 1;
 		} else
-			sputc(' ', fo);
-		len = mime_write(np->n_name, sizeof *np->n_name, len, fo,
-				CONV_TOHDR_A, TD_ICONV, NULL, (size_t)0);
+			putc(' ', fo);
+		len = mime_write(np->n_fullname, sizeof *np->n_fullname,
+				len, fo,
+				domime?CONV_TOHDR_A:CONV_NONE,
+				TD_ICONV, NULL, (size_t)0);
 		if (comma && !(is_to && is_fileaddr(np->n_flink->n_name)))
-			sputc(',', fo);
+			putc(',', fo);
 		col += len + comma;
 		count++;
 	}
-	sputc('\n', fo);
+	putc('\n', fo);
 	return 0;
 }
 
 /*
- * Rewrite a message for forwarding, adding the Resent-Headers.
+ * Rewrite a message for resending, adding the Resent-Headers.
  */
 static int
-infix_fw(fi, fo, mp, to, add_resent)
-FILE *fi, *fo;
-struct message *mp;
-struct name *to;
+infix_resend(FILE *fi, FILE *fo, struct message *mp, struct name *to,
+		int add_resent)
 {
 	size_t count;
-	char *buf = NULL, *cp, *cp2;
+	char *buf = NULL, *cp/*, *cp2*/;
 	size_t c, bufsize = 0;
 
 	count = mp->m_size;
 	/*
-	 * Write the original headers first.
+	 * Write the Resent-Fields.
+	 */
+	if (add_resent) {
+		fputs("Resent-", fo);
+		mkdate(fo, "Date");
+		cp = myaddr();
+		if (cp != NULL) {
+			if (mime_name_invalid(cp, 1)) {
+				if (buf)
+					free(buf);
+				return 1;
+			}
+			fwrite("Resent-From: ", sizeof (char), 13, fo);
+			mime_write(cp, sizeof *cp, strlen(cp), fo,
+					CONV_TOHDR_A, TD_ICONV,
+					NULL, (size_t)0);
+			putc('\n', fo);
+			/*
+			if ((cp2 = value("smtp")) == NULL
+				|| strcmp(cp2, "localhost") == 0)
+				fprintf(fo, "Resent-Sender: %s\n",
+						username());
+			*/
+		}
+#ifdef	notdef
+		/*
+		 * RFC 2822 disallows generation of this field.
+		 */
+		cp = value("replyto");
+		if (cp != NULL) {
+			if (mime_name_invalid(cp, 1)) {
+				if (buf)
+					free(buf);
+				return 1;
+			}
+			fwrite("Resent-Reply-To: ", sizeof (char),
+					17, fo);
+			mime_write(cp, sizeof *cp, strlen(cp), fo,
+					CONV_TOHDR_A, TD_ICONV,
+					NULL, (size_t)0);
+			putc('\n', fo);
+		}
+#endif	/* notdef */
+		if (fmt("Resent-To:", to, fo, 1, 1, 0)) {
+			if (buf)
+				free(buf);
+			return 1;
+		}
+		if (value("stealthmua") == NULL) {
+			fputs("Resent-", fo);
+			message_id(fo);
+		}
+	}
+	/*
+	 * Write the original headers.
 	 */
 	while (count > 0) {
 		if ((cp = foldergets(&buf, &bufsize, &count, &c, fi)) == NULL)
 			break;
-		if (count > 0 && *buf == '\n')
-			break;
 		if (ascncasecmp("status: ", buf, 8) != 0
 				&& strncmp("From ", buf, 5) != 0) {
 			fwrite(buf, sizeof *buf, c, fo);
 		}
+		if (count > 0 && *buf == '\n')
+			break;
 	}
 	/*
-	 * Write the Resent-Headers, but only if the message
-	 * has headers at all.
+	 * Write the message body.
 	 */
-	if (count > 0) {
-		if (add_resent) {
-				fputs("Resent-", fo);
-			date_field(fo);
-			cp = myaddr();
-			if (cp != NULL) {
-				if (mime_name_invalid(cp, 1)) {
-					if (buf)
-						free(buf);
-					return 1;
-				}
-				fwrite("Resent-From: ", sizeof (char), 13, fo);
-				mime_write(cp, sizeof *cp, strlen(cp), fo,
-						CONV_TOHDR_A, TD_ICONV,
-						NULL, (size_t)0);
-				sputc('\n', fo);
-				if ((cp2 = value("smtp")) == NULL
-					|| strcmp(cp2, "localhost") == 0)
-					fprintf(fo, "Resent-Sender: %s\n",
-							username());
-			}
-#ifdef	notdef
-			/*
-			 * RFC 2822 disallows generation of this field.
-			 */
-			cp = value("replyto");
-			if (cp != NULL) {
-				if (mime_name_invalid(cp, 1)) {
-					if (buf)
-						free(buf);
-					return 1;
-				}
-				fwrite("Resent-Reply-To: ", sizeof (char),
-						17, fo);
-				mime_write(cp, sizeof *cp, strlen(cp), fo,
-						CONV_TOHDR_A, TD_ICONV,
-						NULL, (size_t)0);
-				sputc('\n', fo);
-			}
-#endif	/* notdef */
-			if (fmt("Resent-To:", to, fo, 1, 1)) {
-				if (buf)
-					free(buf);
-				return 1;
-			}
-			if (value("stealthmua") == NULL) {
-				fputs("Resent-", fo);
-				message_id(fo);
-			}
-		}
-		sputc('\n', fo);
-		/*
-		 * Write the message body.
-		 */
-		while (count > 0) {
-			if (foldergets(&buf, &bufsize, &count, &c, fi) == NULL)
-				break;
-			fwrite(buf, sizeof *buf, c, fo);
-		}
+	while (count > 0) {
+		if (foldergets(&buf, &bufsize, &count, &c, fi) == NULL)
+			break;
+		if (count == 0 && *buf == '\n')
+			break;
+		fwrite(buf, sizeof *buf, c, fo);
 	}
 	if (buf)
 		free(buf);
@@ -1258,15 +1274,15 @@ struct name *to;
 	return 0;
 }
 
-int
-forward_msg(mp, to, add_resent)
-struct message *mp;
-struct name *to;
+int 
+resend_msg(struct message *mp, struct name *to, int add_resent)
 {
 	FILE *ibuf, *nfo, *nfi;
 	char *tempMail;
 	struct header head;
+	enum okay	ok = STOP;
 
+	memset(&head, 0, sizeof head);
 	if ((to = checkaddrs(to)) == NULL) {
 		senderr++;
 		return 1;
@@ -1276,20 +1292,18 @@ struct name *to;
 		perror(catgets(catd, CATSET, 189, "temporary mail file"));
 		return 1;
 	}
-	if ((nfi = Fopen(tempMail, "r")) == (FILE *)NULL) {
+	if ((nfi = Fopen(tempMail, "r")) == NULL) {
 		senderr++;
 		perror(tempMail);
 		return 1;
 	}
 	rm(tempMail);
 	Ftfree(&tempMail);
-	if ((ibuf = setinput(mp, NEED_BODY)) == NULL)
+	if ((ibuf = setinput(&mb, mp, NEED_BODY)) == NULL)
 		return 1;
 	head.h_to = to;
-	head.h_cc = head.h_bcc = head.h_ref = head.h_smopts = NULL;
-	head.h_attach = NULL;
 	to = fixhead(&head, to, 0);
-	if (infix_fw(ibuf, nfo, mp, head.h_to, add_resent) != 0) {
+	if (infix_resend(ibuf, nfo, mp, head.h_to, add_resent) != 0) {
 		senderr++;
 		rewind(nfo);
 		savedeadletter(nfi);
@@ -1306,8 +1320,12 @@ struct name *to;
 	if (senderr)
 		savedeadletter(nfi);
 	to = elide(to);
-	if (count(to) != 0)
-		start_mta(to, head.h_smopts, nfi);
+	if (count(to) != 0) {
+		if (value("record-resent") == NULL ||
+				mightrecord(nfi, to, 0) == OKAY)
+			ok = transfer(to, head.h_smopts, nfi);
+	} else
+		ok = OKAY;
 	Fclose(nfi);
-	return 0;
+	return ok;
 }
