@@ -1,7 +1,7 @@
 /*
  * Nail - a mail user agent derived from Berkeley Mail.
  *
- * Copyright (c) 2000-2002 Gunnar Ritter, Freiburg i. Br., Germany.
+ * Copyright (c) 2000-2004 Gunnar Ritter, Freiburg i. Br., Germany.
  */
 /*
  * Copyright (c) 2000
@@ -38,16 +38,25 @@
 
 #ifndef lint
 #ifdef	DOSCCS
-static char sccsid[] = "@(#)smtp.c	2.6 (gritter) 1/13/03";
+static char sccsid[] = "@(#)smtp.c	2.30 (gritter) 10/30/04";
 #endif
 #endif /* not lint */
 
 #include "rcv.h"
-#include "extern.h"
 
-#if defined (HAVE_UNAME)
 #include <sys/utsname.h>
-#endif
+#ifdef	HAVE_SOCKETS
+#include <sys/socket.h>
+#include <netdb.h>
+#include <netinet/in.h>
+#ifdef	HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif	/* HAVE_ARPA_INET_H */
+#endif	/* HAVE_SOCKETS */
+#include <unistd.h>
+
+#include "extern.h"
+#include "md5.h"
 
 /*
  * Mail -- a mail program
@@ -55,21 +64,19 @@ static char sccsid[] = "@(#)smtp.c	2.6 (gritter) 1/13/03";
  * SMTP client and other internet related functions.
  */
 
+#ifdef	HAVE_SOCKETS
 static int verbose;
+#endif
 
 /*
  * Return our hostname.
  */
 char *
-nodename()
+nodename(int mayoverride)
 {
 	static char *hostname;
 	char *hn;
-#if defined (HAVE_UNAME)
         struct utsname ut;
-#elif defined (HAVE_GETHOSTNAME)
-        char host__name[MAXHOSTNAMELEN];
-#endif
 #ifdef	HAVE_SOCKETS
 #ifdef	HAVE_IPv6_FUNCS
 	struct addrinfo hints, *res;
@@ -78,16 +85,13 @@ nodename()
 #endif	/* !HAVE_IPv6_FUNCS */
 #endif	/* HAVE_SOCKETS */
 
+	if (mayoverride && (hn = value("hostname")) != NULL && *hn) {
+		free(hostname);
+		hostname = sstrdup(hn);
+	}
 	if (hostname == NULL) {
-#if defined (HAVE_UNAME)
 		uname(&ut);
 		hn = ut.nodename;
-#elif defined (HAVE_GETHOSTNAME)
-		gethostname(host__name, MAXHOSTNAMELEN);
-		hn = host__name;
-#else
-		hn = "unknown";
-#endif
 #ifdef	HAVE_SOCKETS
 #ifdef	HAVE_IPv6_FUNCS
 		memset(&hints, 0, sizeof hints);
@@ -107,7 +111,7 @@ nodename()
 		}
 #endif	/* !HAVE_IPv6_FUNCS */
 #endif	/* HAVE_SOCKETS */
-		hostname = (char *)smalloc(strlen(hn) + 1);
+		hostname = smalloc(strlen(hn) + 1);
 		strcpy(hostname, hn);
 	}
 	return hostname;
@@ -117,7 +121,7 @@ nodename()
  * Return the user's From: address.
  */
 char *
-myaddr()
+myaddr(void)
 {
 	char *cp, *hn;
 	static char *addr;
@@ -132,9 +136,9 @@ myaddr()
 	if (value("smtp") == NULL)
 		return NULL;
 	if (addr == NULL) {
-		hn = nodename();
+		hn = nodename(1);
 		sz = strlen(myname) + strlen(hn) + 2;
-		addr = (char *)smalloc(sz);
+		addr = smalloc(sz);
 		snprintf(addr, sz, "%s@%s", myname, hn);
 	}
 	return addr;
@@ -142,26 +146,45 @@ myaddr()
 
 #ifdef	HAVE_SOCKETS
 
+static char *auth_var(const char *type, const char *addr);
+static int read_smtp(struct sock *sp, int value);
+static int talk_smtp(struct name *to, FILE *fi, struct sock *sp,
+		char *server, char *uhp);
+
+static char *
+auth_var(const char *type, const char *addr)
+{
+	char	*var, *cp;
+	int	len;
+
+	var = ac_alloc(len = strlen(type) + strlen(addr) + 7);
+	snprintf(var, len, "smtp-auth%s-%s", type, addr);
+	if ((cp = value(var)) != NULL)
+		cp = savestr(cp);
+	else {
+		snprintf(var, len, "smtp-auth%s", type);
+		if ((cp = value(var)) != NULL)
+			cp = savestr(cp);
+	}
+	ac_free(var);
+	return cp;
+}
+
 static char	*smtpbuf;
 static size_t	smtpbufsize;
 
 /*
  * Get the SMTP server's answer, expecting value.
  */
-static int
-read_smtp(f, value)
-FILE *f;
+static int 
+read_smtp(struct sock *sp, int value)
 {
 	int ret;
-	size_t len;
+	int len;
 
 	do {
-		if (fgetline(&smtpbuf, &smtpbufsize, NULL, &len, f, 0) == NULL
-				|| len < 6) {
-			if (ferror(f))
-				perror(catgets(catd, CATSET, 255,
-						"SMTP read failed"));
-			else
+		if ((len = sgetline(&smtpbuf, &smtpbufsize, NULL, sp)) < 6) {
+			if (len >= 0)
 				fprintf(stderr, catgets(catd, CATSET, 241,
 					"Unexpected EOF on SMTP connection\n"));
 			return 5;
@@ -185,15 +208,8 @@ FILE *f;
 /*
  * Macros for talk_smtp.
  */
-#define	SMTP_ANSWER(x)	if (ferror(fsi)) { \
-				perror(catgets(catd, CATSET, 256, \
-					"SMTP write error")); \
-				if (b != NULL) \
-					free(b); \
-				return 1; \
-			} \
-			if (read_smtp(fso, x) != (x)) { \
-				fputs("QUIT\r\n", fsi); \
+#define	SMTP_ANSWER(x)	if (read_smtp(sp, x) != (x)) { \
+				swrite(sp, "QUIT\r\n"); \
 				if (b != NULL) \
 					free(b); \
 				return 1; \
@@ -201,32 +217,103 @@ FILE *f;
 
 #define	SMTP_OUT(x)	if (verbose) \
 				fprintf(stderr, ">>> %s", x); \
-			fputs(x, fsi); \
-			fflush(fsi);
+			swrite(sp, x);
 
 /*
  * Talk to a SMTP server.
  */
 static int
-talk_smtp(to, fi, fsi, fso)
-struct name *to;
-FILE *fi, *fsi, *fso;
+talk_smtp(struct name *to, FILE *fi, struct sock *sp,
+		char *server, char *uhp)
 {
 	struct name *n;
 	char *b = NULL, o[LINESIZE];
 	size_t blen, bsize = 0, count;
+	char	*user, *password, *b64, *skinned, *authstr, *cp;
+	enum	{ AUTH_NONE, AUTH_LOGIN, AUTH_CRAM_MD5 } auth;
 
+	skinned = skin(myaddr());
+	user = auth_var("-user", skinned);
+	password = auth_var("-password", skinned);
+	if ((authstr = auth_var("", skinned)) == NULL)
+		auth = user && password ? AUTH_LOGIN : AUTH_NONE;
+	else if (strcmp(authstr, "login") == 0)
+		auth = AUTH_LOGIN;
+	else if (strcmp(authstr, "cram-md5") == 0)
+		auth = AUTH_CRAM_MD5;
+	else {
+		fprintf(stderr, "Unknown SMTP authentication "
+				"method: \"%s\"\n", authstr);
+		return 1;
+	}
+	if (auth != AUTH_NONE && (user == NULL || password == NULL)) {
+		fprintf(stderr, "User and password are necessary "
+				"for SMTP authentication.\n");
+		return 1;
+	}
 	SMTP_ANSWER(2);
-	snprintf(o, sizeof o, "HELO %s\r\n", nodename());
-	SMTP_OUT(o);
-	SMTP_ANSWER(2);
-	snprintf(o, sizeof o, "MAIL FROM: <%s>\r\n", skin(myaddr()));
+#ifdef	USE_SSL
+	if (value("smtp-use-starttls") ||
+			value("smtp-use-tls") /* v11.0 compatibility */) {
+		snprintf(o, sizeof o, "EHLO %s\r\n", nodename(1));
+		SMTP_OUT(o);
+		SMTP_ANSWER(2);
+		SMTP_OUT("STARTTLS\r\n");
+		SMTP_ANSWER(2);
+		if (ssl_open(server, sp, uhp) != OKAY)
+			return 1;
+	}
+#else	/* !USE_SSL */
+	if (value("smtp-use-starttls") || value("smtp-use-tls")) {
+		fprintf(stderr, "No SSL support compiled in.\n");
+		return 1;
+	}
+#endif	/* !USE_SSL */
+	if (auth != AUTH_NONE) {
+		snprintf(o, sizeof o, "EHLO %s\r\n", nodename(1));
+		SMTP_OUT(o);
+		SMTP_ANSWER(2);
+		switch (auth) {
+		default:
+		case AUTH_LOGIN:
+			SMTP_OUT("AUTH LOGIN\r\n");
+			SMTP_ANSWER(3);
+			b64 = strtob64(user);
+			snprintf(o, sizeof o, "%s\r\n", b64);
+			free(b64);
+			SMTP_OUT(o);
+			SMTP_ANSWER(3);
+			b64 = strtob64(password);
+			snprintf(o, sizeof o, "%s\r\n", b64);
+			free(b64);
+			SMTP_OUT(o);
+			SMTP_ANSWER(2);
+			break;
+		case AUTH_CRAM_MD5:
+			SMTP_OUT("AUTH CRAM-MD5\r\n");
+			SMTP_ANSWER(3);
+			for (cp = smtpbuf; digitchar(*cp&0377); cp++);
+			while (blankchar(*cp&0377)) cp++;
+			cp = cram_md5_string(user, password, cp);
+			SMTP_OUT(cp);
+			SMTP_ANSWER(2);
+			break;
+		}
+	} else {
+		snprintf(o, sizeof o, "HELO %s\r\n", nodename(1));
+		SMTP_OUT(o);
+		SMTP_ANSWER(2);
+	}
+	snprintf(o, sizeof o, "MAIL FROM: <%s>\r\n", skinned);
 	SMTP_OUT(o);
 	SMTP_ANSWER(2);
 	for (n = to; n != NULL; n = n->n_flink) {
-		snprintf(o, sizeof o, "RCPT TO: <%s>\r\n", n->n_name);
-		SMTP_OUT(o);
-		SMTP_ANSWER(2);
+		if ((n->n_type & GDEL) == 0) {
+			snprintf(o, sizeof o, "RCPT TO: <%s>\r\n",
+					skin(n->n_name));
+			SMTP_OUT(o);
+			SMTP_ANSWER(2);
+		}
 	}
 	SMTP_OUT("DATA\r\n");
 	SMTP_ANSWER(3);
@@ -235,8 +322,10 @@ FILE *fi, *fsi, *fso;
 	count = fsize(fi);
 	while (fgetline(&b, &bsize, &count, &blen, fi, 1) != NULL) {
 		if (*b == '.')
-			sputc('.', fsi);
-		crlfputs(b, fsi);
+			swrite1(sp, ".", 1, 1);
+		b[blen-1] = '\r';
+		b[blen] = '\n';
+		swrite1(sp, b, blen+1, 1);
 	}
 	SMTP_OUT(".\r\n");
 	SMTP_ANSWER(2);
@@ -251,115 +340,29 @@ FILE *fi, *fsi, *fso;
  * Connect to a SMTP server.
  */
 int
-smtp_mta(server, to, fi)
-char *server;
-struct name *to;
-FILE *fi;
+smtp_mta(char *server, struct name *to, FILE *fi)
 {
-	int sockfd;
-#ifdef	HAVE_IPv6_FUNCS
-	struct addrinfo hints, *res, *res0;
-	char hbuf[NI_MAXHOST];
-#else	/* !HAVE_IPv6_FUNCS */
-	struct sockaddr_in servaddr;
-	struct in_addr **pptr;
-	struct hostent *hp;
-	struct servent *sp;
-	unsigned short port = 0;
-#endif	/* !HAVE_IPv6_FUNCS */
-	FILE *fsi, *fso;
-	int ret;
-	char *portstr;
+	struct sock	so;
+	int	use_ssl, ret;
 
+	memset(&so, 0, sizeof so);
 	verbose = value("verbose") != NULL;
-	portstr = strchr(server, ':');
-	if (portstr == NULL)
-		portstr = "smtp";
-	else {
-		*portstr++ = '\0';
-		if (*portstr == '\0')
-			portstr = "smtp";
-#ifndef	HAVE_IPv6_FUNCS
-		else
-			port = (unsigned short)strtol(portstr, NULL, 10);
+	if (strncmp(server, "smtp://", 7) == 0) {
+		use_ssl = 0;
+		server += 7;
+#ifdef	USE_SSL
+	} else if (strncmp(server, "smtps://", 8) == 0) {
+		use_ssl = 1;
+		server += 8;
 #endif
-	}
-#ifdef	HAVE_IPv6_FUNCS
-	memset(&hints, 0, sizeof hints);
-	hints.ai_socktype = SOCK_STREAM;
-	if (getaddrinfo(server, portstr, &hints, &res0) != 0) {
-		fprintf(stderr, catgets(catd, CATSET, 252,
-				"could not resolve host: %s\n"));
-		return 1;
-	}
-	sockfd = -1;
-	for (res = res0; res != NULL && sockfd < 0; res = res->ai_next) {
-		if (verbose) {
-			if (getnameinfo(res->ai_addr, res->ai_addrlen,
-						hbuf, sizeof hbuf, NULL, 0,
-						NI_NUMERICHOST) != 0)
-				strcpy(hbuf, "unknown host");
-			fprintf(stderr, catgets(catd, CATSET, 192,
-					"Connecting to %s . . ."), hbuf);
-		}
-		if ((sockfd = socket(res->ai_family, res->ai_socktype,
-						res->ai_protocol)) >= 0) {
-			if (connect(sockfd, res->ai_addr, res->ai_addrlen)!=0) {
-				close(sockfd);
-				sockfd = -1;
-			}
-		}
-	}
-	if (sockfd < 0) {
-		freeaddrinfo(res0);
-		perror(catgets(catd, CATSET, 254, "could not connect"));
-		return 1;
-	}
-	freeaddrinfo(res0);
-#else	/* !HAVE_IPv6_FUNCS */
-	if (port == 0) {
-		if ((sp = getservbyname(portstr, "tcp")) == NULL) {
-			if (equal(portstr, "smtp"))
-				port = htons(25);
-			else {
-				fprintf(stderr, catgets(catd, CATSET, 251,
-					"unknown service: %s\n"), portstr);
-				return 1;
-			}
-		}
-		port = sp->s_port;
 	} else
-		port = htons(port);
-	if ((hp = gethostbyname(server)) == NULL) {
-		fprintf(stderr, catgets(catd, CATSET, 252,
-				"could not resolve host: %s\n"));
+		use_ssl = 0;
+	if (sopen(server, &so, use_ssl, server, use_ssl ? "smtps" : "smtp",
+				verbose) != OKAY)
 		return 1;
-	}
-	pptr = (struct in_addr **) hp->h_addr_list;
-	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
-		perror(catgets(catd, CATSET, 253, "could not create socket"));
-		return 1;
-	}
-	memset(&servaddr, 0, sizeof servaddr);
-	servaddr.sin_family = AF_INET;
-	servaddr.sin_port = port;
-	memcpy(&servaddr.sin_addr, *pptr, sizeof(struct in_addr));
-	if (verbose)
-		fprintf(stderr, catgets(catd, CATSET, 192,
-				"Connecting to %s . . ."), inet_ntoa(**pptr));
-	if (connect(sockfd, (struct sockaddr *)&servaddr, sizeof servaddr)
-			!= 0) {
-		perror(catgets(catd, CATSET, 254, "could not connect"));
-		return 1;
-	}
-#endif	/* !HAVE_IPv6_FUNCS */
-	if (verbose)
-		fputs(catgets(catd, CATSET, 193, " connected.\n"), stderr);
-	fsi = (FILE *)Fdopen(sockfd, "w");
-	fso = (FILE *)Fdopen(sockfd, "r");
-	ret = talk_smtp(to, fi, fsi, fso);
-	Fclose(fsi);
-	Fclose(fso);
+	so.s_desc = "SMTP";
+	ret = talk_smtp(to, fi, &so, server, server);
+	sclose(&so);
 	if (smtpbuf) {
 		free(smtpbuf);
 		smtpbuf = NULL;
@@ -369,10 +372,7 @@ FILE *fi;
 }
 #else	/* !HAVE_SOCKETS */
 int
-smtp_mta(server, to, fi)
-char *server;
-struct name *to;
-FILE *fi;
+smtp_mta(char *server, struct name *to, FILE *fi)
 {
 	fputs(catgets(catd, CATSET, 194,
 			"No SMTP support compiled in.\n"), stderr);
